/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include "gtkfiles.h"
#include "um.h"
#include "processdialog.h"
#include "chunk.h"
#include "inifile.h"
#include "main.h"
#include "ringbuf.h"
#include "rawdialog.h"


static GtkObjectClass *parent_class;

static GList *chunks = NULL;


static Datasource *_chunk_filter(Chunk *chunk, chunk_filter_proc proc, 
				 chunk_filter_proc eof_proc, 
				 gboolean allchannels, gboolean convert);


guint chunk_alive_count(void)
{
     return g_list_length ( chunks );
}

static void chunk_destroy(GtkObject *object)
{
     Chunk *c;
     GList *l;
     DataPart *dp;
     c=CHUNK(object);
     g_assert(c->opencount == 0);
     if (c->parts) {
	  for (l=c->parts; l!=NULL; l=l->next) {
	       dp = (DataPart *)l->data;
	       gtk_object_unref(GTK_OBJECT(dp->ds));
	       g_free(dp);
	  }
	  g_list_free(c->parts);
     }
     c->parts = NULL;
     parent_class->destroy(object);
     chunks = g_list_remove ( chunks, object );
}

static void chunk_class_init(GtkObjectClass *klass)
{
     parent_class = gtk_type_class(gtk_object_get_type());
     klass->destroy = chunk_destroy;
}

static void chunk_init(Chunk *object)
{
	object->format.samplerate=44100;
	object->format.samplesize=2;
	object->format.channels=2;
	object->format.sign=1;
	object->format.samplebytes=4;
	object->length=0;
	object->size=0;
	object->parts = 0;
	chunks = g_list_append( chunks, object );
}

guint chunk_get_type(void)
{
	static guint id=0;
	if (!id) {
		GtkTypeInfo info = {
		     "Chunk",
		     sizeof(Chunk),
		     sizeof(ChunkClass),
		     (GtkClassInitFunc) chunk_class_init,
		     (GtkObjectInitFunc) chunk_init 
		};
		id=gtk_type_unique(gtk_object_get_type(),&info);
		}
	return id;
}

static Chunk *chunk_new(void)
{
	return gtk_type_new(chunk_get_type());
}

Chunk *chunk_new_silent(Dataformat *format, gfloat seconds)
{
     return chunk_new_from_datasource(
	  datasource_new_silent(format, seconds * format->samplerate));
}

static void chunk_calc_length(Chunk *c)
{
     GList *l;
     DataPart *dp;
     guint32 len=0;
     for(l=c->parts; l!=NULL; l=l->next) {
	  dp = (DataPart *)l->data;
	  len += dp->length;
     }
     c->length = len;
     c->size = len*c->format.samplebytes;
}

gboolean chunk_format_equal(Chunk *c1, Chunk *c2)
{
	return (c1->format.samplerate==c2->format.samplerate && c1->format.samplesize==c2->format.samplesize && c1->format.channels==c2->format.channels && c1->format.sign==c2->format.sign);
}

static Dataformat target_format;

static gint target_samplerate;
static gint samplecount;

gboolean chunk_convert_samplerate_proc(void *in, guint samplesize, 
				       chunk_writeout_func out_func, WriteoutID id,
				       Chunk *chunk)
{
     while (samplecount<target_samplerate) {
	  if (out_func(id,in,samplesize)) return TRUE;
	  samplecount += chunk->format.samplerate;
     }
     samplecount -= target_samplerate;
     return FALSE;
}

static sample_t last_samples[16];
static gfloat step_count;
static gfloat step_inc;

gboolean chunk_convert_samplerate_linear_proc(void *in, guint samplesize, 
					      chunk_writeout_func out_func, WriteoutID id,
					      Chunk *chunk)
{
     sample_t *f = (sample_t *)in, inc[16];
     gint i,steps;
     if (!samplecount) { 
	  samplecount++; 
	  memcpy(last_samples,f,samplesize);
	  return out_func(id,in,samplesize);
     }
     step_count += step_inc;
     steps = (gint)step_count;
     step_count -= (gfloat)steps;
     for (i=0; i<chunk->format.channels; i++) inc[i] = (f[i] - last_samples[i]) / steps;
     for (; steps>0; steps--) {
	  for (i=0; i<chunk->format.channels; i++)
	       last_samples[i] += inc[i];
	  if (out_func(id,last_samples,samplesize)) return TRUE;
     }
     return FALSE;
}

Chunk *chunk_convert_samplerate(Chunk *chunk, gint samplerate, 
				ChunkInterpolation interpolation)
{
     Datasource *ds;
     target_samplerate = samplerate;
     samplecount = 0;
     step_count = 0.0;
     /* Fixme: The samplerate conversion code sucks. */
     step_inc = (gfloat)samplerate / (gfloat)chunk->format.samplerate;
     if (interpolation == INTERPOLATION_LINEAR && step_inc > 1.0)
	  ds = _chunk_filter( chunk, chunk_convert_samplerate_linear_proc, NULL,
			    TRUE, TRUE);
     else 
	  ds = _chunk_filter(chunk,chunk_convert_samplerate_proc,NULL,TRUE,FALSE);
     if (ds) {
	  /* Somewhat ugly */
	  ds->format.samplerate = samplerate;
	  return chunk_new_from_datasource(ds);
     }
     return NULL;
}

Chunk *chunk_convert_speed(Chunk *chunk, gfloat speed_factor)
{
     target_samplerate = ((float)chunk->format.samplerate) / speed_factor;
     samplecount = 0;
     step_count = 0.0;
     step_inc = (gfloat)(chunk->format.samplerate) / (gfloat)chunk->format.samplerate;
     if (step_inc > 1.0) 
	  return chunk_filter(chunk,chunk_convert_samplerate_linear_proc,NULL,
			      TRUE,TRUE);
     else
	  return chunk_filter(chunk,chunk_convert_samplerate_proc,NULL,TRUE,
			      FALSE);	       
}

static gboolean chunk_convert_samplesize_proc(void *in, guint samplesize, 
					      chunk_writeout_func out_func, WriteoutID id,
					      Chunk *chunk)
{
     char out[16];
     convert_float_pcm(convert_pcm_float(in,&(chunk->format)),
		       out,&target_format);
     return out_func(id,out,target_format.samplesize);
}

/* fixme: smarter algo that doesn't convert data from datasources with 
   fake_pcm==TRUE */
Chunk *chunk_convert_samplesize(Chunk *chunk, guint samplesize, gboolean sign)
{
     Datasource *ds;

     target_format . samplesize = samplesize;
     target_format . sign = sign;
          
     ds = _chunk_filter(chunk,chunk_convert_samplesize_proc,NULL,FALSE,FALSE);
     if (ds) {
	  ds->format.samplesize = samplesize;
	  ds->format.sign = sign;
	  ds->format.samplebytes = ds->format.samplesize * ds->format.channels;
	  ds->length = ds->size / ds->format.samplebytes;
     }
     return chunk_new_from_datasource(ds);
}

static gint channel_to_remove, current_channel;

static gboolean chunk_remove_channel_proc(void *in, guint samplesize, 
					  chunk_writeout_func out_func, WriteoutID id,
					  Chunk *chunk)
{
     guint c = current_channel;
     current_channel++;
     if (current_channel == chunk->format.channels) current_channel = 0;
     if (c != channel_to_remove) return out_func(id,in,samplesize);
     else return FALSE;
}

Chunk *chunk_remove_channel(Chunk *chunk, gint channel)
{
     Datasource *ds;
     channel_to_remove = channel;
     current_channel = 0;
     ds = _chunk_filter(chunk,chunk_remove_channel_proc,NULL,FALSE,FALSE);
     if (ds) {
	  ds->format.channels--;
	  ds->format.samplebytes = ds->format.samplesize * ds->format.channels;
	  ds->length = ds->size / ds->format.samplebytes;
     }
     return chunk_new_from_datasource(ds);
}

gboolean chunk_dump(Chunk *chunk, EFILE *file)
{
     GList *l;
     DataPart *dp;
     l = chunk->parts;
     while (l) {
	  dp = (DataPart *)l->data;
	  if (datasource_dump(dp->ds,dp->position,dp->length,file))
	       return TRUE;
	  l = l->next;
     }
     return FALSE;
}

static Chunk *chunk_filter_chunk;

static gboolean chunk_filter_sample_t_write(WriteoutID id, gpointer data, guint length)
{
     static sample_t buf[512];
     static guint bufpos=0;
     sample_t *d = (sample_t *)data;
     gchar buf2[2048];
     guint i, count;
     if (!data) {
	  if (bufpos == 0) return FALSE;
	  i = bufpos;
	  bufpos = 0;
	  chunk_sample_put_array(buf,buf2,i,chunk_filter_chunk);
	  return chunk_create_temp_write(id,buf2,i*chunk_filter_chunk->format.samplesize);
     }
     count = length / sizeof(sample_t);
     while (count > 0) {
	  i = ARRAY_LENGTH(buf)-bufpos;
	  if (i>count) i=count;
	  memcpy(buf+bufpos,d,i*sizeof(sample_t));
	  bufpos += i;
	  if (bufpos == ARRAY_LENGTH(buf)) {
	       chunk_sample_put_array(buf,buf2,bufpos,chunk_filter_chunk);
	       if (chunk_create_temp_write(id,buf2,
					   bufpos*chunk_filter_chunk->format.samplesize))
		    return TRUE;
	       bufpos = 0;
	  }
	  d += i;
	  count -= i;
     }
     return FALSE;
}

/* Allmn funktion fr stegvis bearbetning av en Chunk.
 * Funktionen lser steg fr steg av chunken och anropar proc.  
 * Efter allt data skickats in kommer eof_proc (om != NULL) att anropas med 
 * samplesize=0.
 */

static Datasource *_chunk_filter(Chunk *chunk, chunk_filter_proc proc, 
				 chunk_filter_proc eof_proc, 
				 gboolean allchannels, gboolean convert)
{
     guint proc_size,single_size,full_size;
     gchar buf[8192];
     guint32 samplepos=0, samplesleft=chunk->length, u, x;
     ChunkHandle *ch;
     ChunkTemp tmp;

     chunk_filter_chunk = chunk;

     if (convert) single_size = sizeof(sample_t);
     else single_size = chunk->format.samplesize;
     
     full_size = single_size * chunk->format.channels;

     if (allchannels) proc_size = full_size;
     else proc_size = single_size;

     ch = chunk_open(chunk);
     if (ch == NULL) return NULL;
     tmp = chunk_create_temp_init();
     processdialog_setup ( samplesleft );

     while (samplesleft > 0) {
          /* printf("samplepos = %d, samplesleft =
%d\n",samplepos,samplesleft); */
	  if (convert)
	       u = chunk_read_array_fp(ch,samplepos,
				       MIN(sizeof(buf)/full_size,samplesleft),
				       (sample_t *)buf) * full_size;
	  else
	       u = chunk_read_array(ch,samplepos,
				    MIN(sizeof(buf),samplesleft*full_size),buf);
          /* printf("u = %d, u/full_size = %d\n",u,u/full_size); */
	  if (!u) {
	       chunk_close(ch);
	       chunk_create_temp_abort(tmp);
	       processdialog_finished();
	       return NULL;
	  }
	  for (x=0; x<u; x+=proc_size) {
	       if (proc(buf+x,proc_size,convert?chunk_filter_sample_t_write:tempfile_write,tmp,
			chunk)) {
			 chunk_close(ch);
			 chunk_create_temp_abort(tmp);
			 processdialog_finished();
			 return NULL;
	       }
	  }
	  samplesleft -= u / full_size;
	  samplepos += u / full_size;
	  if (processdialog_progress(u / full_size)) {
	       chunk_close(ch);
	       chunk_create_temp_abort(tmp);
	       processdialog_finished();
	       return NULL;
	  }
     }
     if (eof_proc && eof_proc(NULL,0,convert?chunk_filter_sample_t_write:
			      tempfile_write,tmp,chunk)) {	  
	  chunk_close(ch);
	  chunk_create_temp_abort(tmp);
	  processdialog_finished();
	  return NULL;
     }
     if (convert && chunk_filter_sample_t_write(tmp,NULL,0)) {
	  chunk_close(ch);
	  chunk_create_temp_abort(tmp);
	  processdialog_finished();
	  return NULL;
     }
     chunk_close(ch);
     processdialog_finished ();
     return tempfile_finished(tmp,&(chunk->format));
}

Chunk *chunk_filter(Chunk *chunk, chunk_filter_proc proc, 
		    chunk_filter_proc eof_proc, 
		    gboolean allchannels, gboolean convert)
{
     return chunk_new_from_datasource(_chunk_filter(chunk,proc,eof_proc,
						    allchannels,convert));
}


/* Allmn funktion fr stegvis avlsning av en Chunk. proc kommer att anropas 
 * en gng fr varje sampling i chunken */

gboolean chunk_parse(Chunk *chunk, chunk_parse_proc proc, gboolean allchannels,
		     gboolean convert)
{

     guint single_size,full_size,proc_size;
     gchar buf[4096],*c=NULL,*d;
     guint32 samplepos=0, samplesleft=chunk->length, u, x;
     ChunkHandle *ch;

     if (convert) single_size = sizeof(sample_t);
     else single_size = chunk->format.samplesize;
     
     full_size = single_size * chunk->format.channels;

     if (allchannels) proc_size = full_size;
     else proc_size = single_size;


     ch = chunk_open(chunk);
     if (ch == NULL) { g_free(c); return TRUE; }
     processdialog_setup ( chunk->length );

     while (samplesleft > 0) {
	  /* printf("samplesleft = %d, samplepos = %d\n",samplesleft,samplepos); */
	  if (convert)	       
	       u = chunk_read_array_fp(ch,samplepos,sizeof(buf)/full_size,
				       (sample_t *)buf) * full_size;
	  else
	       u = chunk_read_array(ch,samplepos,sizeof(buf),buf);
	  /* printf("bytes returned: %d, samples returned: %d\n",u,u/full_size); */
	  if (!u) {
	       chunk_close(ch);
	       processdialog_finished();
	       g_free(c);
	       return TRUE;
	  }
	  d = buf;
	  for (x=0; x<u; x+=proc_size) {
	       if (proc(d+x,proc_size,chunk)) {
		    chunk_close(ch);
		    processdialog_finished();
		    g_free(c);
		    return TRUE;
	       }
	  }
	  samplesleft -= u / full_size;
	  samplepos += u / full_size;
	  if (processdialog_progress(u / full_size)) {
	       chunk_close(ch);
	       g_free(c);
	       return TRUE;
	  }
     }
     chunk_close(ch);
     processdialog_finished ();
     g_free(c);
     return FALSE;
}

static gboolean chunk_onechannel_proc(void *in, guint sample_size, 
				      chunk_writeout_func out_func, 
				      WriteoutID id, Chunk *chunk)
{
     static int channel_count = 0;
     static sample_t mixup = 0.0;
     sample_t s = *((sample_t *)in);
     mixup += s;
     channel_count++;
     if (channel_count == chunk->format.channels) {
	  s = mixup / (sample_t)channel_count;
	  channel_count = 0;
	  mixup = 0.0;
	  return out_func(id,&s,sizeof(s));
     } else 
	  return FALSE;
}

Chunk *chunk_onechannel(Chunk *chunk)
{
     Datasource *ds;
     ds = _chunk_filter(chunk,chunk_onechannel_proc,NULL,FALSE,TRUE);
     if (ds){
	  ds->format.channels = 1;
	  ds->format.samplebytes = ds->format.samplesize;
	  ds->length = ds->size / ds->format.samplebytes;
     }
     return chunk_new_from_datasource(ds);
}

static gint channel_to_copy;

static gboolean chunk_copy_channel_proc(void *in, guint sample_size, 
					chunk_writeout_func out_func, 
					WriteoutID id,
					Chunk *chunk)
{
     static gint channel_count=0;
     gint i;
     i = channel_count;
     channel_count ++;
     if (channel_count == chunk->format.channels) channel_count = 0;
     if (out_func(id,in,sample_size)) return TRUE;
     if (channel_to_copy==i) return out_func(id,in,sample_size);
     else return FALSE;
}

static gboolean chunk_copy_channel_mono_proc(void *in, guint sample_size,
					     chunk_writeout_func out_func, WriteoutID id,
					     Chunk *chunk)
{
     return (out_func(id,in,sample_size) || out_func(id,in,sample_size));
}

Chunk *chunk_copy_channel(Chunk *chunk, gint channel)
{
     Datasource *ds;
     channel_to_copy = channel;
     if (chunk->format.channels == 1) 
	  ds=_chunk_filter(chunk,chunk_copy_channel_mono_proc,NULL,FALSE,FALSE);
     else 
	  ds = _chunk_filter(chunk,chunk_copy_channel_proc,NULL,FALSE,FALSE);
     if (ds) {
	  ds->format.channels++;
	  ds->format.samplebytes = ds->format.samplesize * ds->format.channels;
	  ds->length = ds->size / ds->format.samplebytes;
     }
     return chunk_new_from_datasource(ds);
}

guint get_time_mode = 3;

gchar *get_time(guint32 samplerate, guint32 samples, gchar *timebuf)
{
     static gchar static_buf[16];
     gfloat secs;
     guint32 mins, msecs;
     if (get_time_mode > 2) {
	  get_time_mode = inifile_get_guint32(INI_SETTING_TIME_DISPLAY,0);
	  if (get_time_mode > 2) get_time_mode = 0;
     }
     if (!timebuf) return get_time(samplerate,samples,static_buf);
     if (get_time_mode == 2) 
	  g_snprintf(timebuf,16,"%d",samples);
     else {
	  secs = (gfloat) samples / (gfloat) samplerate;
	  mins = (guint) (secs / 60.0);
	  msecs = ((guint) (secs  * 1000.0));
	  msecs %= 60000;
	  if (get_time_mode == 0) 
	       g_snprintf(timebuf,16,"%02d:%02d.%d",mins,msecs/1000,
			  (msecs%1000)/100);
	  else 
	       g_snprintf(timebuf,16,"%02d:%02d.%03d",mins,msecs/1000,
			  msecs%1000);
     }
     return timebuf;
}


Chunk *chunk_mix(Chunk *c1, Chunk *c2)
{
     #define BUFLEN 256
     guint32 u,l;
     guint i,chn,x1,x2,x;
     sample_t buf1[BUFLEN],buf2[BUFLEN],buf3[BUFLEN];
     sample_t s;
     gpointer outbuf;
     Chunk *m,*c,*d;
     ChunkHandle *ch1, *ch2;
     ChunkTemp tmp;
     g_assert(chunk_format_equal(c1,c2));
     chn = c1->format.channels;
     outbuf = g_malloc(BUFLEN * c1->format.samplesize);
     ch1 = chunk_open(c1);
     if ( ch1 == NULL ) return NULL;
     ch2 = chunk_open(c2);
     if ( ch2 == NULL ) { chunk_close(ch1); return NULL; }
     tmp = chunk_create_temp_init();
     l = MIN(c1->length, c2->length);
     processdialog_setup(l);
     for (u=0; u<l; u+=x) {
	  x1 = chunk_read_array_fp(ch1,u,BUFLEN/chn,buf1);
	  x2 = chunk_read_array_fp(ch2,u,BUFLEN/chn,buf2);
	  if (x1 == 0 || x2 == 0) {
	       chunk_create_temp_abort(tmp);
	       chunk_close(ch1);
	       chunk_close(ch2);
	       processdialog_finished();
	       g_free(outbuf);
	       return NULL;
	  }
	  x = MIN(x1,x2) * chn;
	  for (i=0; i<x; i++) {
	       s = buf1[i] + buf2[i];
	       buf3[i] = CLAMP(s,-1.0,1.0);
	  }
	  convert_float_pcm_array(buf3, outbuf, &(c1->format), x);
	  if (chunk_create_temp_write(tmp,outbuf,x*c1->format.samplebytes) 
	      || processdialog_progress(x)) {
	       chunk_create_temp_abort(tmp);
	       chunk_close(ch1);
	       chunk_close(ch2);
	       processdialog_finished();
	       g_free(outbuf);
	       return NULL;
	  }
     }
     chunk_close(ch1);
     chunk_close(ch2);
     processdialog_finished();
     m = chunk_create_temp_finished(tmp,c1);
     g_free(outbuf);
     if (!m) return NULL;
     if (c1->length == c2->length) return m;
     if (c1->length > l)
	  c = chunk_get_part(c1,l,c1->length-l);
     else 
	  c = chunk_get_part(c2,l,c2->length-l);
     d = chunk_append(m,c);
     gtk_object_sink(GTK_OBJECT(m));
     gtk_object_sink(GTK_OBJECT(c));
     return d;
}


void chunk_foreach(GFunc func, gpointer user_data)
{
     g_list_foreach(chunks,func,user_data);
}

Chunk *chunk_new_from_datasource(Datasource *ds)
{
     Chunk *c;
     DataPart *dp;
     if (ds==NULL) return NULL;
     dp = g_malloc(sizeof(*dp));
     dp->ds = ds;
     dp->position = 0;
     dp->length = ds->length;
     gtk_object_ref(GTK_OBJECT(ds));
     gtk_object_sink(GTK_OBJECT(ds));
     c = chunk_new();
     memcpy(&(c->format),&(ds->format),sizeof(Dataformat));
     c->length = ds->length;
     c->size = c->length * c->format.samplebytes;
     c->parts = g_list_append(NULL,dp);
     return c;
}

ChunkHandle *chunk_open(Chunk *chunk)
{
     GList *l;
     DataPart *dp;
     for (l=chunk->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  if (datasource_open(dp->ds)) {
	       for (;l!=NULL;l=l->prev) datasource_close(dp->ds);
	       return NULL;
	  }
     }
     chunk->opencount ++;
     return chunk;     
}

void chunk_close(ChunkHandle *handle)
{
     GList *l;
     DataPart *dp;
     g_assert(handle->opencount > 0);
     for (l=handle->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  datasource_close(dp->ds);
     }
     handle->opencount --;     	  
}

gboolean chunk_read(ChunkHandle *handle, guint32 sampleno, void *buffer)
{
     GList *l;     
     DataPart *dp;
     for (l=handle->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  if (dp->length > sampleno) 
	       return datasource_read(dp->ds,sampleno,buffer);
	  else
	       sampleno -= dp->length;
     }
     g_assert_not_reached();
     return TRUE;
}

gboolean chunk_read_fp(ChunkHandle *handle, guint32 sampleno, sample_t *buffer)
{
     GList *l;     
     DataPart *dp;
     for (l=handle->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  if (dp->length > sampleno) 
	       return datasource_read_fp(dp->ds,sampleno,buffer);
	  else
	       sampleno -= dp->length;
     }
     g_assert_not_reached();
     return TRUE;
}

guint chunk_read_array(ChunkHandle *handle, guint32 sampleno, 
		       guint size, void *buffer)
{
     GList *l;     
     DataPart *dp;
     guint r=0,i;
     for (l=handle->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  if (dp->length > sampleno) {
	       i = datasource_read_array(dp->ds,
					 dp->position + sampleno,
					 MIN(size,(dp->length - sampleno)*
					     handle->format.samplebytes),
					 buffer);
	       if (i == 0) return 0;
	       r+=i;
	       sampleno = 0;
	       size -= i;
	       if (size==0 || l->next==NULL) return r;
	  } else
	       sampleno -= dp->length;
     }
     return r;
}

guint chunk_read_array_fp(ChunkHandle *handle, guint32 sampleno, 
			  guint samples, sample_t *buffer)
{
     GList *l;     
     DataPart *dp;
     guint r=0,i;
     g_assert(sampleno < handle->length);
     for (l=handle->parts;l!=NULL;l=l->next) {
	  dp = (DataPart *)l->data;
	  if (dp->length > sampleno) {
	       i = datasource_read_array_fp(dp->ds,
					    dp->position+sampleno,
					    MIN(samples,dp->length-sampleno),
					    buffer);
	       g_assert(i <= samples);
	       g_assert(i <= dp->length - sampleno);
	       if (i==0) return 0;
	       r += i;
	       samples -= i;
	       sampleno = 0;
	       if (samples == 0 || l->next==NULL) return r;
	  } else
	       sampleno -= dp->length;
     }
     return r;
}

static DataPart *datapart_list_copy(GList *src, GList **dest)
{
     DataPart *old,*new;
     old = src->data;
     new = g_malloc(sizeof(DataPart));
     new->ds = old->ds;
     new->position = old->position;
     new->length = old->length;
     gtk_object_ref(GTK_OBJECT(old->ds));
     (*dest) = g_list_append((*dest),new);
     return new;
}

Chunk *chunk_append(Chunk *first, Chunk *second)
{
     Chunk *c;
     GList *nl=NULL,*l;
     guint32 len=0;
     DataPart *dp;
     g_assert(chunk_format_equal(first,second));
     for (l=first->parts; l!=NULL; l=l->next) {
	  dp = datapart_list_copy(l,&nl);
	  len += dp->length;
     }
     for (l=second->parts; l!=NULL; l=l->next) {
	  dp = datapart_list_copy(l,&nl);
	  len += dp->length;
     }
     c = chunk_new();
     memcpy(&(c->format),&(first->format),sizeof(Dataformat));
     c->parts = nl;
     c->length = len;
     c->size = len * c->format.samplebytes;
     return c;
}

Chunk *chunk_insert(Chunk *chunk, Chunk *part, guint32 sampleno)
{
     GList *nl=NULL,*l,*m;     
     DataPart *dp;
     Chunk *c;

     g_assert(sampleno <= chunk->length);
     g_assert(chunk_format_equal(chunk,part));
     if (sampleno == chunk->length) return chunk_append(chunk,part);

     /* Get beginning */
     for (l=chunk->parts; sampleno>0; l=l->next) {
	  dp = datapart_list_copy(l,&nl);
	  if (sampleno < dp->length) {
	       dp->length = sampleno;
	       break;
	  }
	  sampleno -= dp->length;
     }

     /* Add the inserted part */
     for (m=part->parts; m!=NULL; m=m->next) 
	  datapart_list_copy(m,&nl);

     /* Add the rest */
     for (; l!=NULL; l=l->next) {
	  dp = datapart_list_copy(l,&nl);
	  dp->position += sampleno;
	  dp->length -= sampleno;
	  sampleno = 0;
     }

     /* Create the chunk */
     c = chunk_new();
     memcpy(&(c->format),&(chunk->format),sizeof(Dataformat));
     c->parts = nl;
     c->length = chunk->length + part->length;
     c->size = chunk->size + part->size;
     return c;
}

Chunk *chunk_get_part(Chunk *chunk, guint32 start, guint32 length)
{
     GList *l,*nl=NULL;
     DataPart *dp;
     Chunk *c;
     l=chunk->parts;
     /* Skip beginning parts */
     while (1) {
	  dp = (DataPart *)l->data;
	  if (start < dp->length) break;
	  start -= dp->length;
	  l=l->next;
     }
     while (length>0 && l!=NULL) {
	  dp = datapart_list_copy(l,&nl);
	  dp->position += start;
	  dp->length -= start;
	  start = 0;
	  if (dp->length > length) dp->length=length;
	  length -= dp->length;
	  l=l->next;
     }

     c = chunk_new();
     memcpy(&(c->format),&(chunk->format),sizeof(Dataformat));
     c->parts = nl;
     chunk_calc_length(c);
     return c;
}

Chunk *chunk_remove_part(Chunk *chunk, guint32 start, guint32 length)
{
     GList *nl=NULL,*l;
     DataPart *dp;
     Chunk *c;
     
     /* Copy beginning parts */
     l=chunk->parts;
     while (start>0) {
	  dp = datapart_list_copy(l,&nl);
	  if (start < dp->length) {
	       dp->length = start;
	       break;
	  }
	  start -= dp->length;
	  l=l->next;
     }
     
     /* Skip parts */
     length += start;
     while (length>0 && l!=NULL) {
	  dp = (DataPart *)l->data;
	  if (dp->length > length) break;
	  length -= dp->length;
	  l=l->next;
     }

     /* Copy remaining parts */
     for (;l!=NULL;l=l->next) {
	  dp = datapart_list_copy(l,&nl);
	  dp->position += length;
	  dp->length -= length;	  
     }

     c = chunk_new();
     memcpy(&(c->format),&(chunk->format),sizeof(Dataformat));
     c->parts = nl;
     chunk_calc_length(c);
     return c;
}

Chunk *chunk_replace_part(Chunk *chunk, guint32 start,
			  guint32 length, Chunk *new)
{
     Chunk *c;
     GList *nl=NULL,*l;
     DataPart *dp;
     guint i;
     for (l=new->parts; l!=NULL; l=l->next)
	  datapart_list_copy(l,&nl);
     c = chunk_remove_part(chunk,start,length);
     for (l=c->parts,i=0; start>0; l=l->next,i++) {
	  dp = (DataPart *)l->data;
	  start -= dp->length;
     }
     for (l=nl; l!=NULL; l=l->next)
	  c->parts = g_list_insert(c->parts, l->data, i);
     g_list_free(nl);
     c->length += new->length;
     c->size = c->length * c->format.samplebytes;
     return c;
}

Chunk *chunk_clone_df(Chunk *chunk, Dataformat *format)
{
     Chunk *c;
     GList *l,*nl=NULL;
     DataPart *dp,*ndp;

     for (l=chunk->parts; l!=NULL; l=l->next) {
	  dp = (DataPart *)l->data;
	  ndp = g_malloc(sizeof(*ndp));
	  ndp->ds = datasource_clone_df(dp->ds, format);
	  gtk_object_ref(GTK_OBJECT(ndp->ds));
	  gtk_object_sink(GTK_OBJECT(ndp->ds));
	  ndp->position = dp->position;
	  ndp->length = dp->length;
	  nl = g_list_append(nl,ndp);
     }

     c = chunk_new();
     memcpy(&(c->format),format,sizeof(Dataformat));
     c->parts = nl;
     chunk_calc_length(c);
     return c;
}

