/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */



/* Chunk objects represent a segment of (PCM) audio data. The Chunks gets its
 * data from one or many Datasource objects. 
 *
 * The Chunks are never modified after they are created, instead all functions 
 * that modify a chunk return a new chunk with the changes, while the original
 * chunk and its data remains intact. 
 */



#ifndef CHUNK_H_INCLUDED
#define CHUNK_H_INCLUDED

#include <stdio.h>
#include <gtk/gtk.h>

#ifdef HAVE_LIBSNDFILE
#include <sndfile.h>
#endif

#include "datasource.h"

#include "gtkfiles.h"

#define CHUNK(obj) GTK_CHECK_CAST(obj,chunk_get_type(),Chunk)
#define CHUNK_CLASS(class) GTK_CHECK_CLASS_CAST(class,chunk_get_type(),ChunkClass)
#define IS_CHUNK(obj) GTK_CHECK_TYPE(obj,chunk_get_type())


typedef struct _Chunk {
     GtkObject object;

     Dataformat format;

     GList *parts;               /* All items should be of type DataPart */

     /* These should match the sum of the parts */
     guint32 length;             /*Number of samples */
     guint32 size;               /* Number of bytes (should always be 
				  * length*samplesize*channels) */

     guint opencount;            /* For detecting missing chunk_close calls */

} Chunk;

typedef struct _ChunkClass {
	GtkObjectClass parent;
} ChunkClass;


typedef struct {
     Datasource *ds;
     guint32 position;
     guint32 length;
} DataPart;


/* Data structure used for chunk_open,chunk_read etc. */
typedef Chunk ChunkHandle; 

/* Different interpolation types that can be used in chunk_convert_samplerate */
typedef enum { INTERPOLATION_NONE, INTERPOLATION_LINEAR } ChunkInterpolation;




typedef gpointer WriteoutID;

/* Function for outputting processed data 
 *
 * data - The data to output
 * length - Length of the data
 *
 * return value - TRUE on failure, FALSE on success
 */

typedef gboolean (*chunk_writeout_func)(WriteoutID id, gpointer data, guint length);


/* Type for use as callback function for chunk_filter.
 *
 * NOTE: These callback functions should not call any other chunk functions 
 * except the chunk_sample_get/put-functions...
 *
 *  in - Buffer of input data
 *  sample_size - The size of the in buffer (0 means EOF) 
 *  out - Function to call for outputting data. If a out function call fails,
 *    this function should fail as well. 
 *  id - Parameter to send with every call to out.
 *  source - The chunk being processed.
 *
 *  return value - TRUE on failure, FALSE on success
 */

typedef gboolean (*chunk_filter_proc)(void *in, guint sample_size, 
				      chunk_writeout_func out_func, WriteoutID id,
				      Chunk *source);




/* Type for use as callback function for chunk_parse.

    sample - Buffer of input data
    sample_size - The size of the in buffer (0 means EOF) 
    chunk - The chunk being processed.

    return value - Return FALSE to continue parsing, otherwise parsing stops.
 */

typedef gboolean (*chunk_parse_proc)(void *sample,gint sample_size,
				     Chunk *chunk);




/* ------------------------------ FUNCTIONS --------------------------------- */


/* Note: All functions that return gboolean values return FALSE on success. */

/* If any function fails, it will display an error message box before it 
 * returns.*/



/*  ------ Creation functions ------  */


guint chunk_get_type(void);
Chunk *chunk_new_from_datasource(Datasource *ds);
Chunk *chunk_clone_df(Chunk *chunk, Dataformat *format);


/* Creates a new Chunk from a file. */

#define chunk_new_from_file(filename) chunk_load(filename)



/* Creates a silent chunk
     format - The data format of the chunk.
     num_samples - The length of the chunk.
*/

#define chunk_new_empty(format,num_samples) chunk_new_from_datasource(datasource_new_silent(format,num_samples))



/* Same as chunk_new_empty but you specify length as number of seconds instead 
 * of samples.
 */

Chunk *chunk_new_silent(Dataformat *format, gfloat seconds);



/* The number of chunk objects currently in existence */

guint chunk_alive_count(void);






/*  ------ Saving functions ------ */


/* Save a chunk to a file. 

     chunk - The chunk to save. 
     filename - The filename to save to. The extension of the filename 
       determines what type to save the data in. If the type cannot be
       detected this way, a dialog box will ask for the format.

     return value - A chunk with the newly saved data, or NULL on failure.
*/

#define chunk_save_to_file(chunk,filename) chunk_save(chunk,filename)



/* Save a chunk's data into a file.

     chunk - The chunk to save.
     file - An opened file handle to save the data into.
     filename - The name of the file being saved into (for a possible error 
       message)
*/

gboolean chunk_dump(Chunk *chunk, EFILE *file);




/*  ------ Chunk reading functions. ------ */


/* Open a chunk for reading. Returns a ChunkHandle that can be used with
 * chunk_read and chunk_read_array. The ChunkHandle must be released using
 * chunk_close. 
 */

ChunkHandle *chunk_open(Chunk *chunk);



/* Read a raw multi-channel sample from the chunk. 
     handle - ChunkHandle returned by chunk_open.
     sampleno - The sample to read.
     buffer - The buffer to read into. Must be at least chunk->samplebytes 
       bytes long. */

gboolean chunk_read(ChunkHandle *handle, guint32 sampleno, void *buffer);



/* Read a floating-point multi-channel sample from the chunk
     handle - ChunkHandle returned by chunk_open.
     sampleno - The sample to read.
     buffer - The buffer to read into. Must be at least channels * sizeof(sample_t) 
       bytes long. */

gboolean chunk_read_fp(ChunkHandle *handle, guint32 sampleno, sample_t *buffer);



/* Read an entire array of samples from the chunk.

     handle - ChunkHandle returned by chunk_open
     sampleno - The first sample to read.
     size - The number of bytes to read. Will be rounded down to whole sample.
     buffer - The buffer to read into

     return value - The number of bytes read (0 for error)
*/

guint chunk_read_array(ChunkHandle *handle, guint32 sampleno, guint size, 
		       void *buffer);



/* Read an entire array of floating-point samples from the chunk.

     handle - ChunkHandle returned by chunk_open
     sampleno - The first sample to read.
     samples - The number of multi-channel samples to read
     buffer - The buffer to read into

     return value - The number of multi-channel samples read (0 for error)
*/

guint chunk_read_array_fp(ChunkHandle *handle, guint32 sampleno, guint samples,
			  sample_t *buffer);


/* Frees the resources used by a ChunkHandle. All handles should be freed when
 * they're no longer in use.
 */

void chunk_close(ChunkHandle *handle);



/* Convert a single-channel PCM sample into a floating point number.
     samplebuf - Pointer to the sample to convert.
     format - A chunk used to determine the data format.
     returns a floating point number between -1.0 and 1.0.
*/

#define chunk_sample_get(samplebuf,fmt) convert_pcm_float(samplebuf,&((fmt)->format))



/* Converts an array of PCM samples into floating point numbers.
     samplebuf - Pointer to the samples to convert.
     format - A chunk used to determine the data format.
     buffer - Pointer to where to store the numbers
     count - The number of one-channel samples to convert
*/

#define chunk_sample_get_array(samplebuf,fmt,buffer,count) convert_pcm_float_array(samplebuf,&((fmt)->format),buffer,count)



/* Converts a floating point number into a PCM sample.
     in - The number to convert.
     out - Where to store the sample.
     format - A chunk used to determine the data format.
*/

#define chunk_sample_put(in,out,fmt) convert_float_pcm(in,out,&((fmt)->format))


/* Converts an array of floating point number into PCM samples.
     inbuf - The numbers to convert.
     outbuf - Where to store the PCM samples.
     count - The number of samples to convert.
     format - A chunk used to determine the data format.
*/

#define chunk_sample_put_array(inbuf,outbuf,count,fmt) convert_float_pcm_array(inbuf,outbuf,&((fmt)->format),count)



/* ------ Chunk effects and utility functions ------- */

/* All these functions return a new (temporary) chunk on success and NULL on 
 * failure */



/* Creates a new chunk with the original chunks data converted into a different
 * sample rate possibly using interpolation. */

Chunk *chunk_convert_samplerate(Chunk *chunk, gint samplerate, 
				ChunkInterpolation interpolation);

Chunk *chunk_convert_speed(Chunk *chunk, gfloat speed_factor);


/* Creates a new chunk with the original chunk's data converted into a different
 * sample size and signedness */

Chunk *chunk_convert_samplesize(Chunk *chunk, guint samplesize, gboolean sign);



/* Creates a new chunk with the original chunk's data but with one channel 
 * removed. */

Chunk *chunk_remove_channel(Chunk *chunk, gint channel);



/* Creates a new chunk with the original chunk's data but with an extra channel
 * inserted containing a copy of another channel's data. */

Chunk *chunk_copy_channel(Chunk *chunk, gint channel);



/* Creates a new chunk with the original chunk's data mixed into one channel. */

Chunk *chunk_onechannel(Chunk *chunk);



/* Creates a new chunk containing a mix of two chunk's data. The chunks must 
 * have the same format but need not be of the same length. */

Chunk *chunk_mix(Chunk *c1, Chunk *c2);



/* Create a new chunk consisting of one chunk's data followed by another
 * chunk's data. Always succeeds.
 */

Chunk *chunk_append(Chunk *first, Chunk *second);



/* Creates a new chunk with the original chunk's data and another chunk's data 
 * inserted at a certain offset (offset must be at an even sample) */

Chunk *chunk_insert(Chunk *chunk, Chunk *part, guint32 position);

/* Returns a chunk containing a part of the original chunk's data */

Chunk *chunk_get_part(Chunk *chunk, guint32 start, guint32 length);

/* These functions return a new chunk where a part of the original chunk's 
 * data has been removed or replaced by the data of another chunk.
 */

Chunk *chunk_remove_part(Chunk *chunk, guint32 start, guint32 length);
Chunk *chunk_replace_part(Chunk *chunk, guint32 start, guint32 length, Chunk *new);



/* ------ Processing functions ------ */


/* Pipe a chunk's contents through a callback function and create a new chunk 
 * with the output.
 *
 *   chunk - The chunk to process
 *   proc - The callback function (see declaration of chunk_filter_proc above)
 *   eof_proc - If this argument is non-NULL, when all data has been filtered
 *     through proc, this function will be called with sample_size==0 to produce
 *     the last data. This argument is usually either NULL or the same as proc.
 *   allchannels - If TRUE the samples will be sent with all channels' data at
 *     the same time, if FALSE one channel at a time will be sent.
 *   convert - If TRUE the in and out buffer will be in sample_t format instead
 *     of the chunk's native sample format.
 *
 *   returns - The output or NULL on failure.
 */

Chunk *chunk_filter(Chunk *chunk, chunk_filter_proc proc, 
		    chunk_filter_proc eof_proc, gboolean allchannels, 
		    gboolean convert);



/* Send a chunk's contents one sample at a time to a callback function.
 *
 *   chunk - The chunk to process.
 *   proc - The callback function (see definition of chunk_parse_proc above).
 *   allchannels - If TRUE the samples will be sent with all channels' data at
 *     the same time, if FALSE one channel at a time will be sent.
 *   convert - If TRUE the samples will be sent as sample_t (floating point) 
 *     values, otherwise they will be sent in raw format.
 *
 *   returns - FALSE on success, TRUE if proc returned TRUE or a read error 
 *     occured.
 */

gboolean chunk_parse(Chunk *chunk, chunk_parse_proc proc, gboolean allchannels,
		     gboolean convert);




/*  ------ Temporary file chunk creation functions ------ */

/* This type identifies the temporary file during creation */
typedef TempFile ChunkTemp;

/* Start creating a temporary file. Returns a handle to the temporary file or
 * NULL on failure. The handle must later be destroyed using either 
 * chunk_create_temp_finished or chunk_create_temp_abort.
 */

#define chunk_create_temp_init tempfile_init



/* Write data into the temporary file. 
 *
 * temp - Handle to the temporary file.
 * data - The data to output
 * length - Length of the data
 *
 * return value - TRUE on failure, FALSE on success
 */

#define chunk_create_temp_write(t,d,l) tempfile_write(t,d,l)



/* Abort creating the temporary file. 
 *   temp - Handle to the temporary file.
 */

#define chunk_create_temp_abort(temp) tempfile_abort(temp)



/* End the temporary file.
 *   temp - Handle to the temporary file.
 *   format - The format the written data was in (only the samplesize,channels,
 *     sign members are read)
 *   return value - A Chunk containing the temporary data or NULL on failure. 
 */

#define chunk_create_temp_finished(temp,fmt) chunk_new_from_datasource(tempfile_finished(temp,&((fmt)->format)))



/* ------ Miscellaneous chunk functions ------ */



/* Determines whether the format parameters of two chunks are the same.
 * To return TRUE, the chunks must have the same samplerate, samplesize, 
 * channels and sign. */

gboolean chunk_format_equal(Chunk *c1, Chunk *c2);



/* Determines how long time a certain number of samples represents, and returns
   it as a text string.

     chunk - The chunk for which we should determine the time (only the 
       samplerate member is used.)
     samples - The number of samples to measure.
     timebuf - The buffer to store the text string into (should be at least 16 
       bytes)

     return value - timebuf
*/

extern guint get_time_mode;
gchar *get_time(guint32 samplerate, guint32 samples, gchar *timebuf);
#define chunk_get_time(c,s,t) get_time((c)->format.samplerate,s,t)



/* Calls func for each existing chunk. */
void chunk_foreach(GFunc func, gpointer user_data);



/* Backwards compatibility macros. */
#define chunk_samples_to_bytes(chunk,samples) ((samples)*(chunk)->format.samplebytes)
#define chunk_bytes_to_samples(chunk,bytes) ((bytes)/(chunk)->format.samplebytes)



#endif
