/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include "chunkview.h"
#include "main.h"
#include "inifile.h"
#include "player.h"

static GtkObjectClass *parent_class;
static guint font_height=0;

enum { VIEW_CHANGED_SIGNAL, SELECTION_CHANGED_SIGNAL, CURSOR_CHANGED_SIGNAL, 
       CHUNK_CHANGED_SIGNAL, LAST_SIGNAL };
static gint chunk_view_signals[LAST_SIGNAL] = { 0 };

static void do_chunk_view_set_cursor(ChunkView *view, guint32 cursor, gboolean 
				     user);

static void chunk_view_changed(ChunkView *cv)
{
     // puts("chunk_view_changed");
     if (cv->image) {
	  gdk_pixmap_unref(cv->image);
	  cv->image = NULL;
     }
     gtk_widget_queue_draw(GTK_WIDGET(cv));
}

static void chunk_view_destroy (GtkObject *object)
{
     ChunkView *cv = CHUNKVIEW(object);
     /* puts("chunk_view_shutdown!"); */
     if (cv->chunk) gtk_object_unref(GTK_OBJECT(cv->chunk));
     cv->chunk = NULL;
     if (cv->marks) g_hash_table_destroy(cv->marks);
     cv->marks = NULL;
     parent_class->destroy(object);
     if (cv->cache) view_cache_free(cv->cache);
     cv->cache = NULL;
}

static gint calc_x(ChunkView *cv, guint32 sample, guint32 width)
{
     gfloat f,g;
     if (sample < cv->viewstart)
	  return -1;
     if (sample > cv->viewend )
	  return width;
     if (cv->viewend - cv->viewstart < width)
	  return sample - cv->viewstart;
     else {
	  f = sample - cv->viewstart;
	  g = cv->viewend - cv->viewstart;
	  return width * (f/g);
     }
}

static void chunk_view_update_image(ChunkView *cv)
{
     GtkWidget *wid = GTK_WIDGET(cv);
     int i,j,y;
     gint w,h;
     /* gint cx,v1,v2,pv1,pv2; */
     /* Skapa pixmap och gc */
     w = wid->allocation.width;
     h = wid->allocation.height;
     if (cv->timescale) h-=font_height;
     cv->image = gdk_pixmap_new(wid->window, w, h, -1);
     cv->image_width = w;
     cv->image_height = h;
     /* Svart bakgrund */
     gdk_draw_rectangle( cv->image, gc_black(wid), TRUE, 0, 0, w, h );
     /* Rita ut markeringen med speciell frg */
     if (cv->selstart != cv->selend) {
	  i = calc_x( cv, cv->selstart, w );
	  j = calc_x( cv, cv->selend, w ) - i;
	  if (j==0) j=1;
	  // printf("i = %d, j = %d\n",i,j);
	  gdk_draw_rectangle( cv->image, gc_selection(wid), TRUE, i, 0, j, h );
     }
     /* Om det inte finns ngon sampling laddad, stanna hr. */
     if (!cv->chunk) return;
     /* Rita gra streck */
     j = h/cv->chunk->format.channels;
     for (i=0; i<cv->chunk->format.channels; i++) {
	  y = j/2 + i*j;
	  gdk_draw_line(cv->image,gc_grey(wid),0,y,w-1,y);
     }
     player_work();
     view_cache_draw(cv->cache, cv->image, h, GTK_WIDGET(cv));
}

struct draw_mark_data {
     ChunkView *view;
     GdkEventExpose *event;
};

static void draw_mark(gpointer key, gpointer value, gpointer user_data)
{
     struct draw_mark_data *dd = (struct draw_mark_data *)user_data;
     ChunkView *cv = dd->view;
     GdkEventExpose *event = dd->event;
     GtkWidget *widget = GTK_WIDGET(cv);
     guint32 o = *((guint32 *)value);
     guint i;

     if ( o >= cv->viewstart && o <= cv->viewend ) {
	  i = calc_x( cv, o, widget->allocation.width );
	  if (event->area.x <= i+10 && event->area.x+event->area.width > i) {
	       gdk_draw_line( widget->window, gc_mark(widget), i, 
			      event->area.y, i, 
			      event->area.y+event->area.height-1);
	       /* Render the text */
	       gdk_gc_set_foreground(widget->style->fg_gc[0],color_mark());
	       gtk_draw_string( widget->style, widget->window, 
				GTK_STATE_NORMAL, i+2, font_height+2, (gchar *)key );
	       gdk_gc_set_foreground(widget->style->fg_gc[0],color_black());
/*	       gdk_draw_text( widget->window, widget->style->font, 
			      gc_mark(widget), i+2, 12, (gchar *)key, 
			      strlen(key) ); */
	  }
     }
}

/* FIXME: Handle very small chunks */
/* Returns: 0 - No bars drawn (too short distance)
 * 1 - Large bars drawn but no text.
 * 2 - Large bars drawn with text or small bars drawn */
static guint draw_time_bars(ChunkView *view, GdkEventExpose *event, 
			    gfloat pixelspersec, gfloat secs, 
			    gboolean wantsmall, gboolean text)
{
     guint32 x;
     char buf[32];
     gboolean big;
     gfloat f,g,h,q;
     gint i;
#if GTK_MAJOR_VERSION == 2
     PangoLayout *pl;
#endif
     GtkWidget *widget = GTK_WIDGET(view);
     g = pixelspersec*secs;
     /* printf("draw_time_bars (pixelspersec=%f,secs=%f,g=%f)\n",pixelspersec,secs,g); */
     if (g<6.0) return 0;
     big = (!wantsmall && g>35.0);
     /* printf("draw_time_bars: text=%d\n",(int)text); */
     f = (gfloat)view->viewstart / (gfloat)view->chunk->format.samplerate;
     h = 0.0;
     /* Avrunda f nert till nrmaste multipel av secs */
     x = (guint32)(f/secs);
     q = ((gfloat)x)*secs;
     h = (q-f)*pixelspersec;
     f = q;
     /* Rita */
     do {
	  /* printf("draw_time_bars: f=%f, h=%f\n",f,h); */
	  gdk_draw_line( widget->window, widget->style->white_gc, h, 
			 widget->allocation.height-font_height-(big?7:3), 
			 h, widget->allocation.height-font_height);
	  if (big && text) {
	       g_snprintf(buf,sizeof(buf),"%ld",(long int)f);
#if GTK_MAJOR_VERSION == 1
	       i = gdk_string_width( widget->style->font, buf ) / 2;
	       gtk_draw_string( widget->style, widget->window, 
				GTK_STATE_NORMAL, h-i, 
				widget->allocation.height-1, buf );
#else
	       pl = gtk_widget_create_pango_layout( widget, buf );
	       pango_layout_get_pixel_size(pl, &i, NULL);
	       gdk_draw_layout( widget->window, widget->style->black_gc,
				h-i/2, widget->allocation.height-font_height,
				pl);
	       g_object_unref(G_OBJECT(pl));
#endif
	  }
	  h += g;
	  f += secs;
     } while (h<event->area.x+event->area.width);
     /* puts("draw_time_bars_finished"); */
     return (big)?2:1;
}

static void draw_timescale(ChunkView *view, GdkEventExpose *event, gboolean text)
{
     GtkWidget *widget = GTK_WIDGET(view);
     guint i;
     gfloat f;
     gfloat bigsizes[] = { 1.0, 2.0, 5.0, 10.0, 20.0, 50.0, 100.0, 200.0, 500.0, 
			   1000.0, 2000.0, 5000.0 };
     gboolean bigskip[] = { FALSE, TRUE, FALSE, FALSE, TRUE, FALSE, 
			    FALSE, TRUE, FALSE, FALSE, TRUE, FALSE };
     gfloat smallsizes[] = { 0.1, 0.2, 0.5 };
     /* puts("draw_timescale"); */
     /* gdk_gc_set_foreground(widget->style->fg_gc[0],color_white()); */
     if (text) gdk_window_clear_area(widget->window, event->area.x, 
				     view->image_height,event->area.width,
				     font_height);
     /* Draw the horizontal line */
     i = widget->allocation.height-font_height;
     gdk_draw_line( widget->window, widget->style->white_gc, event->area.x, 
		    i, event->area.x+event->area.width-1, i);
     /* pixels/sec */
     f = (gfloat)(widget->allocation.width * view->chunk->format.samplerate) / 
	  (gfloat)(view->viewend - view->viewstart); 
     if (f > view->chunk->format.samplerate) f=view->chunk->format.samplerate;
     /* Draw the big bars */
     
     for (i=0; i<ARRAY_LENGTH(bigsizes); i++)
	  switch(draw_time_bars(view,event,f,bigsizes[i],FALSE,text)) {
	  case 0: break;
	  case 1: while (bigskip[i]) i++; break;
	  case 2: goto outer;
	  }
outer:
     for (i=0; i<ARRAY_LENGTH(smallsizes); i++)
	  if (draw_time_bars(view,event,f,smallsizes[i],TRUE,text)) break;

     
     /* gdk_gc_set_foreground(widget->style->fg_gc[0],color_black()); */
     /* puts("draw_timescale finished."); */
}

static gint chunk_view_expose(GtkWidget *widget, GdkEventExpose *event)
{
     ChunkView *cv = CHUNKVIEW(widget);
     guint i;
     struct draw_mark_data dd;
     gboolean expose_timescale=FALSE, expose_text=FALSE;
     if ( cv->image == NULL ) 
	  chunk_view_update_image(cv);
     if (cv->timescale) 
	  if (event->area.y+event->area.height > cv->image_height-7) {
	       expose_timescale=TRUE;
	       if (event->area.y+event->area.height > cv->image_height) {
		    expose_text = TRUE;
		    event->area.height = cv->image_height-event->area.y;
	       }
	  }
     gdk_draw_pixmap ( widget->window, gc_black(widget), cv->image,
		       event->area.x, event->area.y, event->area.x, event->area.y,
		       event->area.width, event->area.height );
     /* Draw the marks */
     dd.view = cv;
     dd.event = event;
     g_hash_table_foreach( cv->marks, draw_mark, &dd );
     /* Draw the cursor */
     if ( cv->cursorpos >= cv->viewstart && 
	  cv->cursorpos <= cv->viewend ) {
	  i = calc_x( cv, cv->cursorpos, widget->allocation.width );
	  if (event->area.x <= i && event->area.x+event->area.width > i) {
	       gdk_draw_line( widget->window, gc_grey(widget), i, 
			      event->area.y, i, 
			      event->area.y+event->area.height-1 );
	  }
     }
     /* Draw the time scale */
     if (cv->chunk && cv->timescale && expose_timescale) 
	  draw_timescale( cv, event, expose_text );
     return FALSE;
}


static void chunk_view_size_request(GtkWidget *widget, GtkRequisition *req)
{
     req->width = 300;
     req->height = 150;
}

static void chunk_view_size_allocate(GtkWidget *widget, 
				     GtkAllocation *allocation)
{
     ChunkView *cv = CHUNKVIEW(widget);
     GtkWidgetClass *wc = GTK_WIDGET_CLASS(parent_class);
     if (wc->size_allocate)
	  wc->size_allocate(widget,allocation);
     if (cv->image != NULL && (
	      widget->allocation.width != cv->image_width ||
	      widget->allocation.height-(cv->timescale?font_height:0) != cv->image_height ) )
	  chunk_view_changed ( cv );
}

static guint32 calc_sample(ChunkView *cv, gfloat x, gfloat xmax)
{
     if (cv->viewend - cv->viewstart < xmax)
	  return MIN(cv->viewstart + x, cv->chunk->length);
     else
	  return cv->viewstart + (x/xmax) * (cv->viewend-cv->viewstart);
}

static guint32 new_selstart;
static guint32 new_selend;

static gint chunk_view_button_press(GtkWidget *widget, GdkEventButton *event)
{
     ChunkView *cv = CHUNKVIEW(widget);
     if (!cv->chunk) return FALSE;
     new_selend = new_selstart = calc_sample( cv, event->x, 
					      (gfloat)widget->allocation.width);
     if (event->button==1) {
	  chunk_view_set_selection( cv, new_selstart, new_selend );
     } else {
	  do_chunk_view_set_cursor(cv,calc_sample(cv,event->x,(gfloat)widget->
						  allocation.width),TRUE);
     }
     return FALSE;
}

static gint chunk_view_motion_notify(GtkWidget *widget, GdkEventMotion *event)
{
     ChunkView *cv = CHUNKVIEW(widget);
     if (!cv->chunk) return FALSE;
     new_selend = calc_sample ( cv, (event->x > 0.0) ? event->x : 0.0, 
				(gfloat)widget->allocation.width );
     chunk_view_set_selection ( cv, new_selstart, new_selend );
     return FALSE;
}

static void chunk_view_class_init(GtkObjectClass *klass)
{
     GtkWidgetClass *wc = GTK_WIDGET_CLASS(klass);
     ChunkViewClass *cvc = CHUNKVIEW_CLASS(klass);
     parent_class = gtk_type_class( gtk_drawing_area_get_type() );

     klass->destroy = chunk_view_destroy;
     wc->expose_event = chunk_view_expose;
     wc->size_request = chunk_view_size_request;
     wc->size_allocate = chunk_view_size_allocate;
     wc->button_press_event = chunk_view_button_press;
     wc->motion_notify_event = chunk_view_motion_notify;
     cvc->view_changed = chunk_view_changed;
     cvc->selection_changed = NULL;
     cvc->cursor_changed = NULL;
     cvc->chunk_changed = NULL;

     chunk_view_signals[VIEW_CHANGED_SIGNAL] = 
	  gtk_signal_new( "view-changed", GTK_RUN_FIRST, GTK_CLASS_TYPE(klass), 
			  GTK_SIGNAL_OFFSET(ChunkViewClass,view_changed),
			  gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0 );
     chunk_view_signals[SELECTION_CHANGED_SIGNAL] = 
	  gtk_signal_new( "selection-changed", GTK_RUN_FIRST, GTK_CLASS_TYPE(klass),
			  GTK_SIGNAL_OFFSET(ChunkViewClass,selection_changed),
			  gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0 );
     chunk_view_signals[CURSOR_CHANGED_SIGNAL] = 
	  gtk_signal_new( "cursor-changed", GTK_RUN_FIRST, GTK_CLASS_TYPE(klass),
			  GTK_SIGNAL_OFFSET(ChunkViewClass,cursor_changed),
			  gtk_marshal_NONE__BOOL, GTK_TYPE_NONE, 1, 
			  GTK_TYPE_BOOL );
     chunk_view_signals[CHUNK_CHANGED_SIGNAL] = 
         gtk_signal_new("chunk-changed", GTK_RUN_FIRST, GTK_CLASS_TYPE(klass),
                        GTK_SIGNAL_OFFSET(ChunkViewClass,chunk_changed),
                        gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

     gtk_object_class_add_signals(klass,chunk_view_signals,LAST_SIGNAL);
}

static void chunk_view_init(GtkObject *obj)
{
     ChunkView *cv;
     cv = CHUNKVIEW(obj);
     cv->chunk = NULL;
     cv->image = NULL;
     cv->viewstart = cv->viewend = cv->selstart = cv->selend = 0;
     cv->cursorpos = 0;
     cv->linemode = cv->timescale = TRUE;
     cv->marks = g_hash_table_new(g_str_hash, g_str_equal);
     cv->cache = view_cache_new();
     gtk_widget_add_events( GTK_WIDGET(obj), GDK_BUTTON_MOTION_MASK | 
			    GDK_BUTTON_PRESS_MASK );
     if (!font_height) {
#if GTK_MAJOR_VERSION == 1
	  font_height = gdk_string_height( GTK_WIDGET(obj)->style->font, 
					   "0123456789")+3;
#else
	  PangoLayout *pl;
	  pl = gtk_widget_create_pango_layout( GTK_WIDGET(obj), "0123456789" );
	  pango_layout_get_pixel_size(pl, NULL, &font_height);
	  g_object_unref(pl);
#endif
     }
}

GtkType chunk_view_get_type(void)
{
     static GtkType id = 0;
     if (!id) {
	  GtkTypeInfo info = {
	       "ChunkView",
	       sizeof(ChunkView),
	       sizeof(ChunkViewClass),
	       (GtkClassInitFunc) chunk_view_class_init,
	       (GtkObjectInitFunc) chunk_view_init 
	  };
	  id = gtk_type_unique( gtk_drawing_area_get_type(), &info );
     }
     return id;
}

GtkWidget *chunk_view_new(void)
{
     return GTK_WIDGET(gtk_type_new(chunk_view_get_type()));
}

void chunk_view_set_chunk(ChunkView *cv, Chunk *c)
{
     if (cv->chunk == c) return;
     if (cv->chunk) gtk_object_unref(GTK_OBJECT(cv->chunk));
     cv->chunk = c;
     if (c) { gtk_object_ref(GTK_OBJECT(c)); gtk_object_sink(GTK_OBJECT(c)); }
     chunk_view_set_selection(cv,0,0);
     chunk_view_set_view(cv, 0, c ? c->length : 0);
     if (c && cv->cursorpos > c->length) 
	  chunk_view_set_cursor(cv,c->length);
     else if (!c)
	  cv->cursorpos = 0;
     chunk_view_changed(cv);
     gtk_signal_emit(GTK_OBJECT(cv),chunk_view_signals[CHUNK_CHANGED_SIGNAL]);
}

void chunk_view_set_chunk_keep_view(ChunkView *cv, Chunk *c)
{
     if (cv->chunk) gtk_object_unref(GTK_OBJECT(cv->chunk));
     cv->chunk = c;
     if (c) { gtk_object_ref(GTK_OBJECT(c)); gtk_object_sink(GTK_OBJECT(c)); }
     chunk_view_set_selection(cv,0,0);
     chunk_view_set_view(cv, cv->viewstart, cv->viewend);
     if (cv->cursorpos>c->length) 
	  chunk_view_set_cursor(cv,c->length);
     gtk_signal_emit( GTK_OBJECT(cv), chunk_view_signals[VIEW_CHANGED_SIGNAL] );
     gtk_signal_emit(GTK_OBJECT(cv),chunk_view_signals[CHUNK_CHANGED_SIGNAL]);
}

static void chunk_view_redraw_samples(ChunkView *cv, guint32 start, guint32 end)
{
     guint startpix,endpix;

     startpix = calc_x(cv,start,GTK_WIDGET(cv)->allocation.width);
     endpix = calc_x(cv,end,GTK_WIDGET(cv)->allocation.width);

     gtk_widget_queue_draw_area(GTK_WIDGET(cv),startpix,0,1+endpix-startpix,
				GTK_WIDGET(cv)->allocation.height-
				(cv->timescale?font_height:0));
     if (cv->image) {
	  gdk_pixmap_unref(cv->image);
	  cv->image = NULL;
     }
}

void chunk_view_set_selection(ChunkView *cv, guint32 selstart, guint32 selend)
{
     guint32 u;
     if (!cv->chunk) {
	  cv->selstart = cv->selend = 0;
	  gtk_signal_emit ( GTK_OBJECT(cv), 
			    chunk_view_signals[SELECTION_CHANGED_SIGNAL] );
	  return;
     }

     if (selstart>selend) {
	  u = selstart;
	  selstart = selend;
	  selend = u;
     }
     if (selend > cv->chunk->length)
	  selend = cv->chunk->length;
     if (selstart >= cv->chunk->length) 
	  selstart = (cv->chunk->length) ? cv->chunk->length-1 : 0;
     if (cv->selstart == selstart && cv->selend == selend) return;

     /* Fixme: Smart algoritm som rknar ut minsta mjliga sampleyta */

     if (!(selend <= cv->viewstart || selstart >= cv->viewend))
	  chunk_view_redraw_samples(cv,MAX(selstart,cv->viewstart),
				    MIN(selend,cv->viewend));
     if (!(cv->selend <= cv->viewstart || cv->selstart >= cv->viewend))
	  chunk_view_redraw_samples(cv,MAX(cv->selstart,cv->viewstart),
				    MIN(cv->selend,cv->viewend));

     cv->selstart = selstart;
     cv->selend = selend;

     gtk_signal_emit ( GTK_OBJECT(cv), 
		       chunk_view_signals[SELECTION_CHANGED_SIGNAL] );
}

void chunk_view_set_view(ChunkView *cv, guint32 viewstart, guint32 viewend)
{
     guint32 u;
     if (!cv->chunk) return; 
     if (viewstart>viewend) {
	  u = viewstart;
	  viewstart = viewend;
	  viewend = u;
     }
     if (viewstart >= cv->chunk->length && viewstart > 0)
	  viewstart = cv->chunk->length-1;
     if (viewend - viewstart < GTK_WIDGET(cv)->allocation.width)
	  viewend = viewstart + GTK_WIDGET(cv)->allocation.width;
     if (viewend > cv->chunk->length)
	  viewend = cv->chunk->length;
     if (cv->viewstart == viewstart && cv->viewend == viewend) return;
     cv->viewstart = viewstart;
     cv->viewend = viewend;
     view_cache_update(cv->cache, cv->chunk, cv->viewstart, cv->viewend,
		       cv->image_width);
     gtk_signal_emit( GTK_OBJECT(cv), chunk_view_signals[VIEW_CHANGED_SIGNAL] );
}

static void do_chunk_view_set_cursor(ChunkView *cv, guint32 cursor, 
				     gboolean user)
{
     gboolean old_in_view,new_in_view;
     guint32 oldpos;
     gint curpix,oldpix;
     g_assert(cv->chunk && cursor<=cv->chunk->length);
     oldpos = cv->cursorpos;
     if (oldpos == cursor) return;
     cv->cursorpos = cursor;
     old_in_view = (oldpos>=cv->viewstart && oldpos < cv->viewend);
     new_in_view = (cursor>=cv->viewstart && cursor < cv->viewend);
     curpix = calc_x(cv,cursor,GTK_WIDGET(cv)->allocation.width);
     oldpix = calc_x(cv,oldpos,GTK_WIDGET(cv)->allocation.width);
     if (oldpix != curpix) {
	  gtk_widget_queue_draw_area(GTK_WIDGET(cv),oldpix,0,1,
				     cv->image_height);
	  gtk_widget_queue_draw_area(GTK_WIDGET(cv),curpix,0,1,
				     cv->image_height);
     }

     gtk_signal_emit(GTK_OBJECT(cv), chunk_view_signals[CURSOR_CHANGED_SIGNAL], 
		     user);
}

static void chunk_view_move_to_cursor(ChunkView *cv)
{
     guint32 w;
     w = (cv->viewend - cv->viewstart) / 2;
     if (cv->chunk->length - cv->cursorpos < w) 
	  chunk_view_set_view(cv, cv->chunk->length - 2*w, cv->chunk->length);
     else if (cv->cursorpos < w)
	  chunk_view_set_view(cv, 0, 2*w);
     else
	  chunk_view_set_view(cv, cv->cursorpos - w, cv->cursorpos + w);
}

void chunk_view_set_cursor(ChunkView *cv, guint32 cursor)
{
     guint32 oc = cv->cursorpos;
     gint i,j;
     do_chunk_view_set_cursor(cv,cursor,FALSE);
     i = calc_x(cv, oc, cv->image_width);
     j = calc_x(cv, cv->cursorpos, cv->image_width);
     if (cv->followmode && (i!=j || i<0 || i>=cv->image_width))
	  chunk_view_move_to_cursor(cv);
}

void chunk_view_set_linemode(ChunkView *cv, gboolean mode)
{
     if (cv->linemode == mode) return;
     cv->linemode = mode;
     chunk_view_changed(cv);
}

void chunk_view_set_timescale(ChunkView *cv, gboolean scale_visible)
{
     if (cv->timescale == scale_visible) return;
     cv->timescale = scale_visible;
     chunk_view_changed(cv);
}

void chunk_view_zoom(ChunkView *cv, gfloat zoom)
{
     guint32 dist,newdist;
     guint32 newcentre;
     dist = cv->viewend - cv->viewstart;
     newdist = (guint32) ((gfloat) dist / zoom);
     if (newdist >= cv->chunk->length) {
	  chunk_view_set_view(cv,0,cv->chunk->length);
	  return;
     }
     newdist = MAX(newdist,GTK_WIDGET(cv)->allocation.width);
     if (cv->cursorpos >= cv->viewstart && cv->cursorpos < cv->viewend)
	  newcentre = cv->cursorpos;
     else
	  newcentre = cv->viewstart + dist/2;
     /*printf("chunk_view_zoom: cursorpos = %d, newcentre = %d\n",cv->cursorpos,
       newcentre);*/
     if (newcentre < newdist/2) 
	  chunk_view_set_view(cv,0,newdist);
     else if (newcentre > cv->chunk->length - newdist/2)
	  chunk_view_set_view(cv,cv->chunk->length-newdist,cv->chunk->length);
     else 
	  chunk_view_set_view(cv, newcentre - newdist/2, newcentre + newdist/2);
}

void chunk_view_set_mark(ChunkView *cv, gchar *label, guint32 position)
{
     guint32 *up;
     gchar *orig_key;
     gboolean b;
     if (!cv->chunk || cv->chunk->length == 0) return;
     b = g_hash_table_lookup_extended(cv->marks,label,(gpointer *)(&orig_key),
				      (gpointer *)(&up));
     if (position == CHUNK_VIEW_BAD_MARK) {
	  if (b) {
	       g_hash_table_remove(cv->marks, label);
	       g_free(orig_key);
	       g_free(up);
	  }
     } else {
	  if (b) {
	       *up = position;
	  } else {
	       up = g_malloc(sizeof(*up));
	       *up = position;
	       g_hash_table_insert(cv->marks, g_strdup(label), up);
	  }
     }
     gtk_widget_queue_draw(GTK_WIDGET(cv));
}

guint32 chunk_view_get_mark(ChunkView *cv, gchar *label)
{
     guint32 *up;
     up = g_hash_table_lookup(cv->marks, label);
     if (up) return *up;
     else return CHUNK_VIEW_BAD_MARK;
}

void chunk_view_clear_marks(ChunkView *cv)
{
     g_hash_table_foreach_remove(cv->marks, free2, NULL);
     gtk_widget_queue_draw(GTK_WIDGET(cv));
}

struct chunk_view_update_marks_data {
     guint32 offset;
     gint32 motion;
};

static gboolean chunk_view_update_marks_r(gpointer key, gpointer value, 
					    gpointer user_data)
{
     struct chunk_view_update_marks_data *data = user_data;
     guint32 *v = value;
     gint32 i;
     i = ((gint32) *v);
     if (i >= data->offset && i+data->motion < data->offset) {
	  g_free(key);
	  g_free(value);
	  return TRUE;
     }
     return FALSE;
}

static void chunk_view_update_marks_f(gpointer key, gpointer value, 
					gpointer user_data)
{
     struct chunk_view_update_marks_data *data = user_data;
     guint32 *v = value;
     gint32 i;
     i = ((gint32) *v);
     if (i >= data->offset)
	  *v += data->motion;
}

void chunk_view_update_marks(ChunkView *cv, guint32 offset, gint32 motion)
{
     struct chunk_view_update_marks_data data;
     data.offset = offset;
     data.motion = motion;
     if (motion < 0)
	  g_hash_table_foreach_remove(cv->marks,chunk_view_update_marks_r,&data);
     g_hash_table_foreach(cv->marks,chunk_view_update_marks_f,&data);
}

gboolean chunk_view_update_cache(ChunkView *view)
{
     gboolean b;
     b = view_cache_update(view->cache, view->chunk, view->viewstart,
			   view->viewend, view->image_width);
     if (b && (view->last_redraw_time != time(0) ||
	  view_cache_uptodate(view->cache))) {
	  chunk_view_changed(view);
	  view->last_redraw_time = time(0);
     }
     return b;
}

void chunk_view_set_followmode(ChunkView *view, gboolean fm)
{
     view->followmode = fm;     
}
