/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include <sys/stat.h>
#include <unistd.h>
#include "gtkfiles.h"
#include "ringbuf.h"
#include "um.h"
#include "main.h"
#include "inifile.h"

#include "datasource.h"

#define MAX_REAL_SIZE inifile_get_guint32(INI_SETTING_REALMAX,INI_SETTING_REALMAX_DEFAULT)

static GtkObjectClass *parent_class;
static GList *datasource_list = NULL;

static int tempfile_count = 0;

static gchar *get_temp_filename(void);

guint datasource_count(void)
{
     return g_list_length(datasource_list);
}

static void datasource_init(Datasource *obj)
{
     datasource_list = g_list_append(datasource_list, obj);
     obj->type = DATASOURCE_SILENCE;
     obj->format.samplerate=44100;
     obj->format.samplesize=2;
     obj->format.channels=2;
     obj->format.sign=1;
     obj->format.samplebytes=4;
     obj->length = 0;
     obj->size = 0;
     obj->fake_pcm = FALSE;
     obj->opencount = 0;
     obj->tag = 0;
}

static void datasource_clear(Datasource *ds)
{
     g_assert(ds->opencount == 0);

     switch (ds->type) {

     case DATASOURCE_TEMPFILE:
	  if (ds->data.virtual.filename)
	       xunlink(ds->data.virtual.filename);	  
	  /* Fall through */

     case DATASOURCE_VIRTUAL:
	  if (ds->data.virtual.filename) {
	       g_free(ds->data.virtual.filename);
	       ds->data.virtual.filename = NULL;
	  }
	  break;
	  
     case DATASOURCE_REAL:
	  if (ds->data.real) {
	       g_free(ds->data.real);
	       ds->data.real = NULL;
	  }
	  break;

     case DATASOURCE_SNDFILE_TEMPORARY:
	  if (ds->data.sndfile.filename)
	       xunlink(ds->data.sndfile.filename);
	  /* Fall through */

     case DATASOURCE_SNDFILE:
	  if (ds->data.sndfile.filename) {
	       g_free(ds->data.sndfile.filename);
	       ds->data.sndfile.filename = NULL;
	  }
	  break;

     case DATASOURCE_CLONE:
	  if (ds->data.clone) 
	       gtk_object_unref(GTK_OBJECT(ds->data.clone));
	  ds->data.clone = NULL;
     }
     
     ds->type = DATASOURCE_SILENCE;
}

static void datasource_destroy(GtkObject *obj)
{
     Datasource *ds = DATASOURCE(obj);
     datasource_clear(ds);
     datasource_list = g_list_remove(datasource_list, obj);
     parent_class->destroy(obj);
}

static void datasource_class_init(GtkObjectClass *klass)
{
     parent_class = gtk_type_class(gtk_object_get_type());
     klass->destroy = datasource_destroy;
}

GtkType datasource_get_type(void)
{
     static GtkType id = 0;
     if (!id) {
		GtkTypeInfo info = {
		     "Datasource",
		     sizeof(Datasource),
		     sizeof(DatasourceClass),
		     (GtkClassInitFunc) datasource_class_init,
		     (GtkObjectInitFunc) datasource_init 
		};
		id=gtk_type_unique(gtk_object_get_type(),&info);	  
     }
     return id;
}

Datasource *datasource_clone(Datasource *orig)
{
     Datasource *ds;
     ds = gtk_type_new(datasource_get_type());
     ds->type = orig->type;
     memcpy(&(ds->format),&(orig->format),sizeof(Dataformat));
     ds->length = orig->length;
     ds->size = orig->size;
     ds->fake_pcm = orig->fake_pcm;
     /* We don't want to fail so if the cloning would require creating a
      * symbolic link create a DATASOURCE_CLONE instead */
     if (orig->type == DATASOURCE_TEMPFILE || 
	 orig->type == DATASOURCE_SNDFILE_TEMPORARY) {
	  ds->type = DATASOURCE_CLONE;
	  ds->data.clone = orig;
	  gtk_object_ref(GTK_OBJECT(orig));
	  return ds;
     }
     switch (orig->type) {
     case DATASOURCE_TEMPFILE:
     case DATASOURCE_SNDFILE_TEMPORARY:
	  g_assert_not_reached();
	  break;
     case DATASOURCE_VIRTUAL:
	  ds->data.virtual.filename = g_strdup(orig->data.virtual.filename);
	  ds->data.virtual.offset = orig->data.virtual.offset;
	  break;
     case DATASOURCE_REAL:
	  ds->data.real = g_malloc(orig->size);
	  memcpy(ds->data.real, orig->data.real, orig->size);
	  break;
     case DATASOURCE_SNDFILE:
	  ds->data.sndfile.filename = g_strdup(orig->data.sndfile.filename);
	  ds->data.sndfile.raw_readable = orig->data.sndfile.raw_readable;
     case DATASOURCE_CLONE:
	  ds->data.clone = orig->data.clone;
	  gtk_object_ref(GTK_OBJECT(ds->data.clone));
     }
     return ds;
}

Datasource *datasource_clone_df(Datasource *source, Dataformat *format)
{
     Datasource *df;

     /* If we have changed the sample format, we need to make DATASOURCE_SNDFILE
      * sources into DATASOURCE_CLONE sources, since libsndfile will still know
      * (and assume) the original format. 
      */
     if (source->type == DATASOURCE_SNDFILE && 
	 (source->format.samplesize != format->samplesize ||
	 source->format.sign != format->sign ||
	 source->format.channels != format->channels)) {
	  df = gtk_type_new(datasource_get_type());
	  df->type = DATASOURCE_CLONE;
	  memcpy(&(df->format),format,sizeof(Dataformat));
	  df->size = source->size;
	  df->length = source->size / format->samplebytes;
	  df->data.clone = source;
	  gtk_object_ref(GTK_OBJECT(source));
	  return df;
     }

     /* The regular case */
     df = datasource_clone(source);
     memcpy(&(df->format),format,sizeof(Dataformat));     
     return df;     
}

Datasource *datasource_new_silent(Dataformat *format, guint32 samples)
{
     Datasource *ds;
     ds = gtk_type_new(datasource_get_type());
     memcpy(&(ds->format), format, sizeof(Dataformat));
     ds->type = DATASOURCE_SILENCE;
     ds->length = samples;
     ds->size = samples * ds->format.samplebytes;
     return ds;
}

Datasource *datasource_new_from_data(void *data, Dataformat *format, 
				     guint32 size)
{
     Datasource *ds;
     ds = (Datasource *)gtk_type_new(datasource_get_type());
     ds->type = DATASOURCE_REAL;
     memcpy(&(ds->format),format,sizeof(Dataformat));
     ds->length = size / format->samplebytes;
     ds->size = size;
     ds->data.real = data;
     return ds;
}

gboolean datasource_dump(Datasource *ds, guint32 position, 
			 guint32 length, EFILE *file)
{
     gchar buf[4096];
     guint32 i;
     guint u;
     if (datasource_open(ds)) return TRUE;
     while (length > 0) {
	  i = MIN(length*ds->format.samplebytes,sizeof(buf));
	  u = datasource_read_array(ds,position,i,buf);
	  if (u==0 || e_fwrite(buf,u,file)) { 
	       datasource_close(ds); 
	       return TRUE; 
	  }
	  length -= u/ds->format.samplebytes;
	  position += u/ds->format.samplebytes;
     }
     datasource_close(ds);
     return FALSE;
}

gboolean datasource_realize(Datasource *ds)
{
     gchar *c;
     guint32 sz=ds->size;

     if (datasource_open(ds)) return TRUE;

     c = g_malloc(sz);
     if (datasource_read_array(ds,0,sz,c)) {
	  g_free(c);
	  datasource_close(ds);
	  return TRUE;
     }
     datasource_close(ds);
     
     datasource_clear(ds);
     ds->type = DATASOURCE_REAL;
     ds->data.real = c;     

     return FALSE;
}

static gboolean datasource_open_sndfile(Datasource *ds)
{
#if defined(HAVE_LIBSNDFILE)
     SF_INFO i;
     SNDFILE *s;
     char *d;
     s = sf_open(ds->data.sndfile.filename,SFM_READ,&i);
     if (s == NULL) {
	  d = g_strdup_printf("Couldn't open %s",
			      ds->data.sndfile.filename);
	  user_message(d,UM_OK);
	  g_free(d);
	  return TRUE;
     }

     ds->data.sndfile.handle = s;
     ds->data.sndfile.pos = 0;
     return FALSE;
#else
     g_assert_not_reached();
     return NULL;
#endif     
}

static void datasource_close_sndfile(Datasource *ds)
{
#if defined(HAVE_LIBSNDFILE)
     sf_close(ds->data.sndfile.handle);
#else
     g_assert_not_reached();
#endif
}

gboolean datasource_open(Datasource *ds)
{
     if (ds->opencount == 0) 
	  switch (ds->type) {
	  case DATASOURCE_VIRTUAL:
	  case DATASOURCE_TEMPFILE:
	       ds->data.virtual.handle = e_fopen(ds->data.virtual.filename,
						 EFILE_READ);
	       if (ds->data.virtual.handle == NULL) return TRUE;
	       ds->data.virtual.pos = 0;
	       break;
	  case DATASOURCE_SNDFILE:
	  case DATASOURCE_SNDFILE_TEMPORARY:
	       if (datasource_open_sndfile(ds)) return TRUE;
	       break;
	  case DATASOURCE_CLONE:
	       if (datasource_open(ds->data.clone)) return TRUE;
	  }     
     ds->opencount ++;
     return FALSE;
}

void datasource_close(Datasource *ds)
{
     g_assert(ds->opencount != 0);
     ds->opencount --;
     if (ds->opencount > 0) return;
     switch (ds->type) {
     case DATASOURCE_VIRTUAL:
     case DATASOURCE_TEMPFILE:
	  e_fclose(ds->data.virtual.handle);
	  break;
     case DATASOURCE_SNDFILE:
     case DATASOURCE_SNDFILE_TEMPORARY:
	  datasource_close_sndfile(ds);
	  break;
     case DATASOURCE_CLONE:
	  datasource_close(ds->data.clone);
	  break;
     }     
}

static guint datasource_clone_read_array(Datasource *source, guint32 sampleno,
					 guint size, gpointer buffer)
{
     /* This is not optimized but it's only used in rare cases so the important
      * thing is that it works.*/
     guint32 orig_offset;
     guint32 orig_sampleno;
     guint32 orig_size;
     guint32 orig_adjust;
     guint x;
     gchar *p;
     /* Offset to be sent to the original chunk */
     orig_offset = sampleno * source->format.samplebytes;
     /* orig_offset converted to sample number */
     orig_sampleno = orig_offset / source->data.clone->format.samplebytes;
     /* How bany bytes too early will we read? */
     orig_adjust = orig_offset % source->data.clone->format.samplebytes;
     /* How much data should we read (should be able to fill buffer) */
     orig_size = size + orig_adjust + source->data.clone->format.samplebytes - 1;
     p = g_malloc(orig_size);
     x = datasource_read_array(source->data.clone, orig_sampleno, orig_size, p);
     if (x != 0) {
	  g_assert(x-orig_adjust >= size);
	  memcpy(buffer, p+orig_adjust, size);
	  x = size;
     }
     g_free(p);
     return x;			       
}

#if defined(HAVE_LIBSNDFILE)
static void sndfile_read_error(sf_count_t x, Datasource *source)
{
     gchar c[256],*d;
     if (x<0) sf_error_str(source->data.sndfile.handle,c,sizeof(c));
     else strcpy(c,"Unexpected end of file");
     d = g_strdup_printf("Error reading %s: %s",
			 source->data.sndfile.filename, c);
     user_message(d,UM_OK);
     g_free(d);
}
#endif

static guint datasource_sndfile_read_array(Datasource *source, 
					   guint32 sampleno, guint size,
					   gpointer buffer)
{
#if defined(HAVE_LIBSNDFILE)
     gchar c[256],*d,*b;
     guint u;
     sf_count_t x;
     if (source->data.sndfile.pos != sampleno) {
	  if (sf_seek(source->data.sndfile.handle,sampleno,SEEK_SET) == -1) {
	       sf_error_str(source->data.sndfile.handle,c,sizeof(c));
	       d = g_strdup_printf("Error seeking in %s: %s",
				   source->data.sndfile.filename, c);
	       user_message(d,UM_OK);
	       g_free(d);
	       return 0;
	  }
	  source->data.sndfile.pos = sampleno;
     }
     if (source->data.sndfile.raw_readable) {
	  /* Read raw */
	  x = sf_read_raw(source->data.sndfile.handle,buffer,size);	  
	  if (x>0) source->data.sndfile.pos += x/source->format.samplebytes;
	  if (x<size) {
	       sndfile_read_error(x,source);
	       return 0;
	  }
	  return x;
     } else if (!(source->fake_pcm) && source->format.samplesize==2 &&
	   source->format.sign) {
	  /* Read as short */
	  x = sf_readf_short(source->data.sndfile.handle,buffer,
			     size / source->format.samplebytes);
	  if (x>0) source->data.sndfile.pos += x;
	  if (x < size / source->format.samplebytes) {
	       sndfile_read_error(x,source);
	       return 0;
	  }
	  return x * source->format.samplebytes;
     } else {
	  /* Read as float, convert to pcm */
	  u = size / source->format.samplebytes;
	  b = g_malloc(u * source->format.channels * sizeof(sample_t));
	  x = sf_readf_sample_t(source->data.sndfile.handle,(sample_t *)b,u);
	  if (x>0) source->data.sndfile.pos += x;
	  if (x<u) {
	       sndfile_read_error(x,source);
	       g_free(b);
	       return 0;		    
	  }
	  convert_float_pcm_array((sample_t *)b,buffer,&(source->format),
				  x*source->format.channels);
	  g_free(b);
	  return x*source->format.samplebytes;
     } 
#else
     g_assert_not_reached();
     return 0;
#endif
}


guint datasource_read_array(Datasource *source, guint32 sampleno, guint size,
			    gpointer buffer)
{
     guint32 x;
     g_assert(source->opencount > 0);
     /* Check sampleno */
     g_assert(sampleno <= source->length);     
     /* Check size */
     /* Round down to even samples */     
     size = size - size % source->format.samplebytes; 
     x = (source->length - sampleno) * (guint32)(source->format.samplebytes);
     if (size > x) size = x; /* Round down to available data */
     /* Do it */
     switch (source->type) {
     case DATASOURCE_REAL:
	  memcpy(buffer,source->data.real+sampleno*source->format.samplebytes,
		 size);
	  return size;
     case DATASOURCE_VIRTUAL: 
     case DATASOURCE_TEMPFILE:
	  /* Calculate offset in file */
	  x = source->data.virtual.offset + 
	       sampleno*source->format.samplebytes;
	  if (x != source->data.virtual.pos && 
	      e_fseek(source->data.virtual.handle, x, SEEK_SET))
	       return 0;
	  source->data.virtual.pos = x;
	  if (e_fread(buffer,size,source->data.virtual.handle)) return 0;
	  source->data.virtual.pos += size;
	  return size;
     case DATASOURCE_SILENCE:
	  memset(buffer,0,size);
	  return size;
     case DATASOURCE_SNDFILE:
     case DATASOURCE_SNDFILE_TEMPORARY:
	  return datasource_sndfile_read_array(source,sampleno,size,buffer);
     case DATASOURCE_CLONE:
	  return datasource_clone_read_array(source,sampleno,size,buffer);
     default:
	  g_assert_not_reached();
	  return 0;
     }
}

gboolean datasource_read(Datasource *source, guint32 sampleno, gpointer buffer)
{
     return (datasource_read_array(source,sampleno,source->format.samplebytes,
				   buffer) != source->format.samplebytes);
}

static guint datasource_sndfile_read_array_fp(Datasource *source, 
					      guint32 sampleno, guint samples,
					      sample_t *buffer)
{
#if defined(HAVE_LIBSNDFILE)
     gchar c[256],*d;
     sf_count_t x;

     if (source->data.sndfile.pos != sampleno) {
	  if (sf_seek(source->data.sndfile.handle,sampleno,SEEK_SET) == -1) {
	       sf_error_str(source->data.sndfile.handle,c,sizeof(c));
	       d = g_strdup_printf("Error seeking in %s: %s",
				   source->data.sndfile.filename, c);
	       user_message(d,UM_OK);
	       g_free(d);
	       return 0;
	  }
	  source->data.sndfile.pos = sampleno;
     }

     x = sf_readf_sample_t(source->data.sndfile.handle,buffer,samples);
     if (x>0) source->data.sndfile.pos += x;
     if (x<samples) {
	  sndfile_read_error(x,source);
	  return 0;		    
     }
     return x;

#else
     g_assert_not_reached();
     return 0;
#endif
}

guint datasource_read_array_fp(Datasource *source, guint32 sampleno,
			       guint samples, sample_t *buffer)
{
     gchar *p;
     guint x,s;
     
     g_assert(sampleno <= source->length);
     if (samples > source->length-sampleno) 
	  samples = (guint)(source->length-sampleno);

     switch (source->type) {
     case DATASOURCE_SILENCE:
	  memset(buffer,0,samples*source->format.channels*sizeof(sample_t));
	  return samples;
     case DATASOURCE_SNDFILE:
     case DATASOURCE_SNDFILE_TEMPORARY:
	  return 
	       datasource_sndfile_read_array_fp(source,sampleno,samples,buffer);
     default:
	  s = samples * source->format.samplebytes;
	  p = g_malloc(s);
	  x = datasource_read_array(source,sampleno,s,p);
	  g_assert(x==s || x==0);
	  if (x==s) {
	       convert_pcm_float_array(p,&(source->format),buffer,
				       samples*source->format.channels);
	       return samples;
	  }
	  return 0;
     }
}

gboolean datasource_read_fp(Datasource *ds, guint32 sampleno, sample_t *buffer)
{
     return (datasource_read_array_fp(ds,sampleno,1,buffer)!=1);
}

static gboolean datasource_uses_file(Datasource *ds, gchar *filename)
{
     return ((ds->type==DATASOURCE_VIRTUAL && 
	      is_same_file(ds->data.virtual.filename,filename)) ||
	     ((ds->type==DATASOURCE_SNDFILE &&
	       is_same_file(ds->data.sndfile.filename,filename))));
}

gboolean datasource_backup_unlink(gchar *filename)
{
     GList *l;
     Datasource *ds;
     gchar *lastname=filename,*t;
     for (l=datasource_list; l!=NULL; l=l->next) {
	  ds = (Datasource *)l->data;
	  if (!datasource_uses_file(ds,filename)) continue;
	  t = get_temp_filename();
	  if (xlink(lastname,t)) {
	       g_free(t);
	       return TRUE;
	  }
	  switch (ds->type) {
	  case DATASOURCE_VIRTUAL:
	       g_free(ds->data.virtual.filename);
	       ds->data.virtual.filename = t;
	       ds->type = DATASOURCE_TEMPFILE;
	       break;
	  case DATASOURCE_SNDFILE:
	       g_free(ds->data.sndfile.filename);
	       ds->data.sndfile.filename = t;
	       ds->type = DATASOURCE_SNDFILE_TEMPORARY;
	       break;
	  default: 
	       g_assert_not_reached(); 
	       break;
	  }	  
	  lastname = t;
     }
     return xunlink(filename);
}


/* TEMPFILE */

static gchar *try_tempdir(gchar *dir)
{
     gchar *c;
     FILE *f;
     if (!dir) return 0;
     c = g_strjoin("/",dir,".mhwaveedit_temp_test",NULL);
     f = fopen(c,"w");
     if (f) {
	  fclose(f);
	  unlink(c);
     }
     g_free(c);
     return f?dir:NULL;
}

static gchar *get_temp_directory(void)
{
     static gchar *tempdir = NULL;
     gchar *c;
     if (!tempdir) 
	  tempdir = try_tempdir(getenv("TEMP"));
     if (!tempdir)
	  tempdir = try_tempdir(getenv("TMP"));
     if (!tempdir) {
	  c = g_strjoin("/",get_home_directory(),".mhwaveedit",NULL);
	  mkdir(c,0700);
	  tempdir = try_tempdir(c);
     }
     if (!tempdir)
	  tempdir = try_tempdir("/tmp");
     if (!tempdir)
	  tempdir = try_tempdir(get_home_directory());
     if (!tempdir)
	  tempdir = ".";
     return tempdir;
}

G_LOCK_DEFINE_STATIC(tempfile);

static gchar *get_temp_filename(void)
{
     gchar *c;
     G_LOCK(tempfile);
     c =  g_strdup_printf("%s/mhwaveedit-temp-%d-%04d",get_temp_directory(),
			  getpid(),++tempfile_count);
     G_UNLOCK(tempfile);
     return c;
}

struct temp {
     EFILE *handle;
     Ringbuf *start_buf;
};
 
TempFile tempfile_init(void)
{
     struct temp *t;
     t = g_malloc(sizeof(*t));
     t->handle = NULL;
     t->start_buf = ringbuf_new(MAX_REAL_SIZE);
     g_assert(ringbuf_available(t->start_buf) == 0);
     return t;
}

gboolean tempfile_write(TempFile handle, gpointer data, guint length)
{
     struct temp *t = (struct temp *)handle;
     gchar *buf,*c;
     guint32 u;
     gboolean b;
     /* If we haven't started writing to a real file, fill up the start buffer 
      */
     if (t->handle == NULL) {
	  if (ringbuf_freespace(t->start_buf) < length){
	       c = get_temp_filename();
	       t->handle = e_fopen(get_temp_filename(),EFILE_WRITE);
	       g_free(c);
	       if (t->handle == NULL ||
		   write_wav_header(t->handle,1,1,11025,0x7FFFFFFF)) {
		    return TRUE;
	       }
	       /* Fall through to file-writing part */
	  } else {
	       ringbuf_enqueue(t->start_buf,data,length);
	       return FALSE;
	  }
     }
     /* Do we still have data left in the start buffer to write to file?
      * In that case, write out length*2 bytes and put in the new data in the
      * start buffer 
      */
     if (ringbuf_available(t->start_buf)>0) {
	  buf = g_malloc(length*2);
	  u = ringbuf_dequeue(t->start_buf,buf,length*2);
	  /* Do we have room for the new data as well? */
	  if (u<=length) {
	       memcpy(buf+u,data,length);
	       u+=length;
	  } else
	       ringbuf_enqueue(t->start_buf,data,length);
	  b = e_fwrite(buf, u, t->handle);
	  g_free(buf);
	  return b;
     } else {
	  return e_fwrite(data,length,t->handle);
     }
}

void tempfile_abort(TempFile handle)
{
     struct temp *t = (struct temp *)handle;
     if (t->handle)
	  e_fclose_remove(t->handle);
     ringbuf_free(t->start_buf);
     g_free(t);
}

Datasource *tempfile_finished(TempFile handle, Dataformat *format)
{
     struct temp *t = (struct temp *)handle;
     gchar buf[4096];
     guint32 u;
     Datasource *ds;
     long int l;
     if (t->handle == NULL) {
	  ds = gtk_type_new(datasource_get_type());
	  ds->type = DATASOURCE_REAL;
	  memcpy(&(ds->format),format,sizeof(Dataformat));
	  u = ringbuf_available(t->start_buf);
	  ds->size = u;
	  ds->length = u / format->samplebytes;
	  ds->data.real = g_malloc(u);
	  ringbuf_dequeue(t->start_buf,ds->data.real,u);
	  ringbuf_free(t->start_buf);
	  g_free(t);
	  return ds;
     }
     while (ringbuf_available(t->start_buf)>0) {
	  u = ringbuf_dequeue(t->start_buf,buf,sizeof(buf));
	  if (e_fwrite(buf,u,t->handle)) {
	       tempfile_abort(handle);
	       return NULL;
	  }
     }

     l = e_ftell(t->handle);
     if (l == -1) {
	  tempfile_abort(t);
	  return NULL;
     }
     
     if (e_fseek(t->handle,0,SEEK_SET) ||
	 write_wav_header(t->handle, format->channels, 
			  format->samplesize, format->samplerate, l - 44)) {
	  tempfile_abort(t);
	  return NULL;
     }

     g_assert(e_ftell(t->handle) == 44);

     ds = gtk_type_new(datasource_get_type());
     ds->type = DATASOURCE_TEMPFILE;
     memcpy(&(ds->format),format,sizeof(Dataformat));
     ds->size = l-44;
     ds->length = (l-44)/format->samplebytes;
     ds->data.virtual.filename = g_strdup(t->handle->filename);
     ds->data.virtual.offset = 44;

     e_fclose(t->handle);
     ringbuf_free(t->start_buf);
     g_free(t);
     return ds;     
}

static guint8 *copyle16(guint8 *buf, guint16 n)
{
     buf[0] = n&0xFF;
     n >>= 8;
     buf[1] = n&0xFF;
     return buf+2;
}

static guint8 *copyle32(guint8 *buf, guint32 n)
{
     buf[0] = n&0xFF;
     n >>= 8;
     buf[1] = n&0xFF;
     n >>= 8;
     buf[2] = n&0xFF;
     n >>= 8;
     buf[3] = n&0xFF;
     return buf+4;
}

void get_wav_header(guint16 channels, guint16 samplesize, guint32 samplerate, 
		    guint32 datasize, void *buffer)
{
     guint32 l, Bps=samplerate*channels*samplesize;
     guint16 Bpsmp=channels*samplesize, samplebits=samplesize<<3;
     guint8 *c = buffer;
     /* 'WAVE' + 'fmt ' + lngd + fmt(16) + 'data' + lngd + data*/
     l = 4     +  4     +  4    +  16     +  4     +  4    + datasize;
     memcpy(c,"RIFF",4);
     c += 4;
     c = copyle32(c,l);
     memcpy(c,"WAVEfmt \20\0\0\0\1\0",14);
     c += 14;
     c = copyle16(c,channels);
     c = copyle32(c,samplerate);
     c = copyle32(c,Bps);
     c = copyle16(c,Bpsmp);
     c = copyle16(c,samplebits);
     memcpy(c,"data",4);
     c += 4;
     c = copyle32(c,datasize);
     g_assert((c-(guint8 *)buffer) == 44);
}

gboolean write_wav_header(EFILE *f, guint16 channels, 
			  guint16 samplesize, guint32 samplerate, 
			  guint32 datasize)
{
     gchar c[44];
     get_wav_header(channels,samplesize,samplerate,datasize,c);
     return e_fwrite(c,44,f);
}



/* PCM<->FLOAT SAMPLE CONVERSION ROUTINES */
/* These routines could really use some optimizing if anyone feels like it. */

#if G_BYTE_ORDER == G_LITTLE_ENDIAN

sample_t convert_pcm_float(void *samplebuf, Dataformat *format)
{
     sample_t s;
     gint32 i32;
     guint32 u32;
     if (format->sign)
	  switch (format->samplesize) {
	  case 1: 
	       s=(sample_t)(*((gchar *)samplebuf));
	       if (s<0.0) s/=128.0; else s/=127.0;
	       return s;
	  case 2: 
	       s=(sample_t)(*((gint16 *)samplebuf));
	       if (s<0.0) s/=32768.0; else s/=32767.0;
	       return s;
	  case 3: 
	       i32 = (*((gint32 *)samplebuf));
	       i32 &= 0xFFFFFF;
	       if (i32 & 0x800000) i32 -= 16777216;
	       s = (sample_t) i32;
	       if (s<0.0) s/=8388608.0; else s/=8388607.0;
	       return s;
	  case 4: 
	       s=(sample_t)(*((gint32 *)samplebuf));
	       if (s<0.0) s/=2147483648.0; else s/=2147483647.0;
	       return s;
	  }
     else 
	  switch (format->samplesize) {
	  case 1: 
	       s=(sample_t)(*((guchar *)samplebuf));
	       s-=127.5;
	       s/=127.5;
	       return s;
	  case 2: 
	       s=(sample_t)(*((guint16 *)samplebuf));
	       s-=32767.5;
	       s/=32767.5;
	       return s;
	  case 3: 
	       u32 = (*((guint32 *)samplebuf));
	       u32 &= 0xFFFFFF;
	       s = (sample_t) u32;
	       s-=8388607.5;
	       s/=8388607.5;
	       return s;
	  case 4: 
	       s=(sample_t)(*((guint32 *)samplebuf));
	       s-=2147483647.5;
	       s/=2147483647.5;
	       return s;
	  }
     g_assert_not_reached();
     return 0.0;
}

void convert_pcm_float_array(void *samplebuf, Dataformat *format, 
			     sample_t *buffer, guint count)
{
     sample_t s;
     gchar *c = samplebuf;
     gint32 i32;
     guint32 u32;
     if (format->sign)
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
		    s=(sample_t)(*((gchar *)c));
		    if (s<0.0) s/=128.0; else s/=127.0;
		    *buffer = s;
		    c += 1;
		    buffer ++;
	       }
	       break;
	  case 2: 
	       for (; count>0; count--) {
		    s=(sample_t)(*((gint16 *)c));
		    if (s<0.0) s/=32768.0; else s/=32767.0;
		    *buffer = s;
		    c += 2;
		    buffer ++;
	       }
	       break;
	  case 3: 
	       for (; count>0; count--) {
		    i32 = (*((gint32 *)c));
		    i32 &= 0xFFFFFF;
		    if (i32 & 0x800000) i32 -= 16777216;
		    s = (sample_t) i32;
		    if (s<0.0) s/=8388608.0; else s/=8388607.0;
		    *buffer = s;
		    c += 3;
		    buffer ++;
	       }
	       break;
	  case 4: 
	       for (; count>0; count--) {
		    s=(sample_t)(*((gint32 *)c));
		    if (s<0.0) s/=2147483648.0; else s/=2147483647.0;
		    *buffer = s;
		    c += 4;
		    buffer ++;
	       }
	       break;
	  }
     else 
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
		    s=(sample_t)(*((guchar *)c));
		    s-=127.5;
		    s/=127.5;
		    *buffer = s;
		    c += 1;
		    buffer ++;
	       }
	       break;
	  case 2: 
	       for (; count>0; count--) {
		    s=(sample_t)(*((guint16 *)c));
		    s-=32767.5;
		    s/=32767.5;
		    *buffer = s;
		    c += 2;
		    buffer ++;
	       }
	       break;
	  case 3: 
	       for (; count>0; count--) {
		    u32 = (*((guint32 *)c));
		    u32 &= 0xFFFFFF;
		    s = (sample_t) u32;
		    s-=8388607.5;
		    s/=8388607.5;
		    *buffer = s;
		    c += 3;
		    buffer ++;
	       }
	       break;
	  case 4: 
	       for (; count>0; count--) {
		    s=(sample_t)(*((guint32 *)c));
		    s-=2147483647.5;
		    s/=2147483647.5;
		    *buffer = s;
		    c += 4;
		    buffer ++;
	       }
	       break;
	  }
}

void convert_float_pcm(sample_t d, void *samplebuf, Dataformat *format)
{
     gint32 i32;
     guint32 u32;
     guint8 u80,u81,u82;
     if (d > 1.0) d = 1.0;
     else if (d < -1.0) d = -1.0;
     if (format->sign)
	  switch (format->samplesize) {
	  case 1: 
	       if (d<0.0) d*=128.0; else d*=127.0;
	       (*((gchar *)samplebuf)) = (gchar)d;
	       return;
	  case 2: 
	       if (d<0.0) d*=32768.0; else d*=32767.0;
	       (*((gint16 *)samplebuf)) = (gint16)d;
	       return;
	  case 3: 
	       if (d<0.0) d*=8388608.0; else d*=8388607.0;
	       i32 = (gint32) d;
	       u32 = (guint32) i32;
	       u80 = u32;
	       u81 = u32 >> 8;
	       u82 = u32 >> 16;
	       ((gint8 *)samplebuf)[0] = u80;
	       ((gint8 *)samplebuf)[1] = u81;
	       ((gint8 *)samplebuf)[2] = u82;
	       return;
	  case 4: 
	       if (d<0.0) d*=2147483648.0; else d*=2147483647.0;
	       (*((gint32 *)samplebuf)) = (gint32)d;
	       return;
	  }
     else 
	  switch (format->samplesize) {
	  case 1: 
	       d*=127.5;
	       d+=127.5;
	       (*((guchar *)samplebuf)) = (guchar)d;
	       return;
	  case 2: 
	       d*=32767.5;
	       d+=32767.5;
	       (*((guint16 *)samplebuf)) = (guint16)d;
	       return;
	  case 3: 
	       d*=8388607.5;	       
	       d+=8388607.5;	       
	       u32 = (guint32) d;
	       u80 = u32;
	       u81 = u32 >> 8;
	       u82 = u32 >> 16;
	       ((gint8 *)samplebuf)[0] = u80;
	       ((gint8 *)samplebuf)[1] = u81;
	       ((gint8 *)samplebuf)[2] = u82;
	       return;
	  case 4: 
	       d*=2147483647.5;
	       d+=2147483647.5;
	       (*((guint32 *)samplebuf)) = (guint32)d;
	       return ;
	  }
}

void convert_float_pcm_array(sample_t *inbuf, void *outbuf, Dataformat *format, 
			     guint count)
{
     gint32 i32;
     guint32 u32;
     guint8 u80,u81,u82;
     sample_t d;
     sample_t *in = inbuf;
     gchar *out = outbuf;
     if (format->sign)
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    if (d<0.0) d*=128.0; else d*=127.0;
		    (*((gchar *)out)) = (gchar)d;
		    in++;
		    out+=1;
	       }
	       return;
	  case 2: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;

		    if (d<0.0) d*=32768.0; else d*=32767.0;
		    (*((gint16 *)out)) = (gint16)d;
		    in ++;
		    out += 2;
	       }
	       return;
	  case 3: 
	       if (count == 0) return;
	       count--;
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    if (d<0.0) d*=8388608.0; else d*=8388607.0;
		    i32 = (gint32) d;
		    (*((gint32 *)out)) = i32;
		    in ++;
		    out += 3;
	       }
	       d = *in;
	       if (d > 1.0) d = 1.0;
	       else if (d < -1.0) d = -1.0;
	       if (d<0.0) d*=8388608.0; else d*=8388607.0;
	       i32 = (gint32) d;
	       u32 = (guint32) i32;
	       u80 = u32;
	       u81 = u32 >> 8;
	       u82 = u32 >> 16;
	       out[0] = u80;
	       out[1] = u81;
	       out[2] = u82;
	       return;
	  case 4: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    if (d<0.0) d*=2147483648.0; else d*=2147483647.0;
		    (*((gint32 *)out)) = (gint32)d;
		    in ++;
		    out += 4;
	       }
	       return;
	  }
     else 
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    d*=127.5;
		    d+=127.5;
		    (*((guchar *)out)) = (guchar)d;
		    in ++;
		    out += 1;
	       }
	       return;
	  case 2: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    d*=32767.5;
		    d+=32767.5;
		    (*((guint16 *)out)) = (guint16)d;
		    in ++;
		    out += 2;
	       }
	       return;
	  case 3: 
	       if (count == 0) return;
	       count --;
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    d*=8388607.5;	       
		    d+=8388607.5;	       
		    u32 = (guint32) d;
		    (*((guint32 *)out)) = u32;
		    in ++;
		    out += 3;
	       }
	       d = *in;
	       if (d > 1.0) d = 1.0;
	       else if (d < -1.0) d = -1.0;
	       d*=8388607.5;	       
	       d+=8388607.5;	       
	       u32 = (guint32) d;
	       u80 = u32;
	       u81 = u32 >> 8;
	       u82 = u32 >> 16;
	       out[0] = u80;
	       out[1] = u81;
	       out[2] = u82;
	       return;
	  case 4: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    d*=2147483647.5;
		    d+=2147483647.5;
		    (*((guint32 *)out)) = (guint32)d;
		    in ++;
		    out += 4;
	       }
	       return ;
	  }
}

#else

#warning "Untested code."

sample_t convert_pcm_float(void *samplebuf, Dataformat *format)
{
     sample_t s;
     gint32 i32;
     guint32 u32;
     guint16 u16;
     gint16 i16;
     if (format->sign)
	  switch (format->samplesize) {
	  case 1: 
	       s=(sample_t)(*((gchar *)samplebuf));
	       if (s<0.0) s/=128.0; else s/=127.0;
	       return s;
	  case 2: 
	       u16 = (*((guint16 *)samplebuf));
	       u16 = u16<<8 | u16>>8;
	       i16 = (gint16)u16;
	       s=(sample_t)i16;
	       if (s<0.0) s/=32768.0; else s/=32767.0;
	       return s;
	  case 3: 
	       u32 = (*((guint32 *)samplebuf));
	       /* abcd => 0cba */
	       u32 = ((u32 << 8) & 0xFF0000) | ((u32 >> 8) & 0xFF00) |
		    ((u32 >> 24) & 0xFF);
	       i32 = (gint32)u32;
	       if (i32 & 0x800000) i32 = - (i32 & 0x7FFFFF);
	       s = (sample_t) i32;
	       if (s<0.0) s/=8388608.0; else s/=8388607.0;
	       return s;
	  case 4: 
	       u32 = *((guint32 *)samplebuf);
	       /* abcd => dcba */
	       u32 = (u32 << 24) | ((u32 << 8) & 0xFF0000) |
		    ((u32 >> 8) & 0xFF00) | ((u32 >> 24) & 0xFF);
			 
	       i32 = (gint32)u32;
	       s=(sample_t)i32;
	       if (s<0.0) s/=2147483648.0; else s/=2147483647.0;
	       return s;
	  }
     else 
	  switch (format->samplesize) {
	  case 1: 
	       s=(sample_t)(*((guchar *)samplebuf));
	       s-=127.5;
	       s/=127.5;
	       /* g_assert(s<=1.0 && s>=-1.0); */
	       return s;
	  case 2: 
	       u16 = (*((guint16 *)samplebuf));
	       u16 = u16<<8 | u16>>8;
	       s=(sample_t)(*((guint16 *)u16));
	       s-=32767.5;
	       s/=32767.5;
	       return s;
	  case 3: 
	       u32 = (*((guint32 *)samplebuf));
	       /* abcd => 0cba */
	       u32 = ((u32 << 8) & 0xFF0000) | ((u32 >> 8) & 0xFF00) |
		    ((u32 >> 24) & 0xFF);
	       s = (sample_t) u32;
	       s-=8388607.5;
	       s/=8388607.5;
	       return s;
	  case 4: 
	       u32 = *((guint32 *)samplebuf);
	       /* abcd => dcba */
	       u32 = (u32 << 24) | ((u32 << 8) & 0xFF0000) |
		    ((u32 >> 8) & 0xFF00) | ((u32 >> 24) & 0xFF);
	       s=(sample_t)u32;
	       s-=2147483647.5;
	       s/=2147483647.5;
	       return s;
	  }
     g_assert_not_reached();
     return 0.0;
}

void convert_pcm_float_array(void *samplebuf, Dataformat *format, 
			     sample_t *buffer, guint count)
{
     sample_t s;
     gchar *c = samplebuf;
     gint32 i32;
     guint32 u32;
     gint16 i16;
     guint16 u16;
     if (format->sign)
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
		    s=(sample_t)(*((gchar *)c));
		    if (s<0.0) s/=128.0; else s/=127.0;
		    *buffer = s;
		    c += 1;
		    buffer ++;
	       }
	       break;
	  case 2: 
	       for (; count>0; count--) {
		    u16 = (*((gint16 *)c));
		    u16 = (u16 << 8) | (u16 >> 8);
		    i16 = (gint16)u16;
		    s=(sample_t)i16;
		    if (s<0.0) s/=32768.0; else s/=32767.0;
		    *buffer = s;
		    c += 2;
		    buffer ++;
	       }
	       break;
	  case 3: 
	       for (; count>0; count--) {
		    u32 = (*((guint32 *)c));
		    /* abcd => 0cba */
		    u32 = ((u32 << 8) & 0xFF0000) | ((u32 >> 8) & 0xFF00) |
			 ((u32 >> 24) & 0xFF);
		    i32 = (gint32)u32;
		    if (i32 & 0x800000) i32 = - (i32 & 0x7FFFFF);
		    s = (sample_t) i32;
		    if (s<0.0) s/=8388608.0; else s/=8388607.0;
		    *buffer = s;
		    c += 3;
		    buffer ++;
	       }
	       break;
	  case 4: 
	       for (; count>0; count--) {
		    u32 = (*((guint32 *)c));
		    u32 = (u32 << 24) | ((u32 << 8) & 0xFF0000) |
			 ((u32 >> 8) & 0xFF00) | ((u32 >> 24) & 0xFF);
		    i32 = (gint32)u32;
		    s=(sample_t)i32;
		    if (s<0.0) s/=2147483648.0; else s/=2147483647.0;
		    *buffer = s;
		    c += 4;
		    buffer ++;
	       }
	       break;
	  }
     else 
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
		    s=(sample_t)(*((guchar *)c));
		    s-=127.5;
		    s/=127.5;
		    *buffer = s;
		    c += 1;
		    buffer ++;
	       }
	       break;
	  case 2: 
	       for (; count>0; count--) {
		    u16 = (*((gint16 *)c));
		    u16 = (u16 << 8) | (u16 >> 8);
		    s=(sample_t)u16;
		    s-=32767.5;
		    s/=32767.5;
		    *buffer = s;
		    c += 2;
		    buffer ++;
	       }
	       break;
	  case 3: 
	       for (; count>0; count--) {
		    u32 = (*((guint32 *)c));
		    /* abcd => 0cba */
		    u32 = ((u32 << 8) & 0xFF0000) | ((u32 >> 8) & 0xFF00) |
			 ((u32 >> 24) & 0xFF);
		    s = (sample_t) u32;
		    s-=8388607.5;
		    s/=8388607.5;
		    *buffer = s;
		    c += 3;
		    buffer ++;
	       }
	       break;
	  case 4: 
	       for (; count>0; count--) {
		    u32 = (*((guint32 *)c));
		    u32 = (u32 << 24) | ((u32 << 8) & 0xFF0000) |
			 ((u32 >> 8) & 0xFF00) | ((u32 >> 24) & 0xFF);
		    s=(sample_t)u32;
		    s-=2147483647.5;
		    s/=2147483647.5;
		    *buffer = s;
		    c += 4;
		    buffer ++;
	       }
	       break;
	  }
}

void convert_float_pcm(sample_t d, void *samplebuf, Dataformat *format)
{
     gint32 i32;
     guint32 u32;
     gint16 i16;
     guint16 u16;
     guint8 u80,u81,u82;
     if (d > 1.0) d = 1.0;
     else if (d < -1.0) d = -1.0;
     if (format->sign)
	  switch (format->samplesize) {
	  case 1: 
	       if (d<0.0) d*=128.0; else d*=127.0;
	       (*((gchar *)samplebuf)) = (gchar)d;
	       return;
	  case 2: 
	       if (d<0.0) d*=32768.0; else d*=32767.0;
	       i16 = (gint16)d;
	       i16 = (i16 >> 8) | (i16 << 8);
	       (*((gint16 *)samplebuf)) = i16;
	       return;
	  case 3: 
	       if (d<0.0) d*=8388608.0; else d*=8388607.0;
	       i32 = (gint32) d;
	       u80 = (guint8)i32;
	       u81 = (guint8)(i32 >> 8);
	       u82 = (guint8)(i32 >> 16);
	       *(((guint8 *)samplebuf)+1) = u80;
	       *(((guint8 *)samplebuf)+1) = u80;
	       *(((guint8 *)samplebuf)+1) = u80;
	       return;
	  case 4: 
	       if (d<0.0) d*=2147483648.0; else d*=2147483647.0;
	       i32 = (gint32) d;
	       i32 = (i32 << 24) | ((i32 << 8) & 0xFF0000) |
		    ((i32 >> 8) & 0xFF00) | ((i32 >> 24) & 0xFF);	       
	       (*((gint32 *)samplebuf)) = (gint32)d;
	       return;
	  }
     else 
	  switch (format->samplesize) {
	  case 1: 
	       d*=127.5;
	       d+=127.5;
	       (*((guchar *)samplebuf)) = (guchar)d;
	       return;
	  case 2: 
	       d*=32767.5;
	       d+=32767.5;
	       u16 = (guint16)d;
	       u16 = (u16 >> 8) | (u16 << 8);
	       *((guint16 *)samplebuf) = u16;
	       return;
	  case 3: 
	       d*=8388607.5;	       
	       d+=8388607.5;	       
	       u32 = (guint32) d;
	       u80 = (guint8) u32;
	       u81 = (guint8) (u32 >> 8);
	       u82 = (guint8) (u32 >> 16);
	       *(((guint8 *)samplebuf)+0) = u80;
	       *(((guint8 *)samplebuf)+1) = u81;
	       *(((guint8 *)samplebuf)+2) = u82;
	       return;
	  case 4: 
	       d*=2147483647.5;
	       d+=2147483647.5;
	       u32 = (guint32) d;
	       u32 = (u32 << 24) | ((u32 << 8) & 0xFF0000) |
		    ((u32 >> 8) & 0xFF00) | ((u32 >> 24) & 0xFF);
	       (*((guint32 *)samplebuf)) = (guint32)d;
	       return ;
	  }
}

void convert_float_pcm_array(sample_t *inbuf, void *outbuf, Dataformat *format, 
			     guint count)
{
     gint32 i32;
     guint32 u32;
     gint16 i16;
     guint16 u16;
     sample_t d;
     sample_t *in = inbuf;
     gchar *out = outbuf;
     if (format->sign)
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    if (d<0.0) d*=128.0; else d*=127.0;
		    (*((gchar *)out)) = (gchar)d;
		    in++;
		    out+=1;
	       }
	       return;
	  case 2: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    if (d<0.0) d*=32768.0; else d*=32767.0;
		    i16 = (gint16)d;
		    out[0] = i16;
		    out[1] = i16>>8;
		    in ++;
		    out += 2;
	       }
	       return;
	  case 3: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    if (d<0.0) d*=8388608.0; else d*=8388607.0;
		    i32 = (gint32) d;
		    out[0] = i32;
		    out[1] = i32>>8;
		    out[2] = i32>>16;
		    in ++;
		    out += 3;
	       }
	       return;
	  case 4: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    if (d<0.0) d*=2147483648.0; else d*=2147483647.0;
		    i32 = (guint32) d;
		    out[0] = i32;
		    out[1] = i32 >> 8;
		    out[2] = i32 >> 16;
		    out[3] = i32 >> 24;
		    in ++;
		    out += 4;
	       }
	       return;
	  }
     else 
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    d*=127.5;
		    d+=127.5;
		    (*((guchar *)out)) = (guchar)d;
		    in ++;
		    out += 1;
	       }
	       return;
	  case 2: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    d*=32767.5;
		    d+=32767.5;
		    u16 = (guint16)d;
		    out[0] = u16;
		    out[1] = u16 >> 8;
		    in ++;
		    out += 2;
	       }
	       return;
	  case 3: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    d*=8388607.5;		    
		    d+=8388607.5;		    
		    u32 = (guint32) d;
		    out[0] = u32;
		    out[1] = u32 >> 8;
		    out[2] = u32 >> 16;
		    (*((guint32 *)out)) = u32;
		    in ++;
		    out += 3;
	       }
	       return;
	  case 4: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    d*=2147483647.5;
		    d+=2147483647.5;
		    u32 = (guint32)d;
		    out[0] = u32;
		    out[1] = u32>>8;
		    out[2] = u32>>16;
		    out[3] = u32>>24;
		    in ++;
		    out += 4;
	       }
	       return ;
	  }
}

#endif

void datasource_conversion_selftest(void)
{
     guint samplesizes[] = { 1, 1, 2, 2, 3, 3, 4, 4 };
     gboolean signs[] = { FALSE, TRUE, FALSE, TRUE, FALSE, TRUE, FALSE, TRUE };
     guint8 pcm_buf[128],pcm_buf2[128];
     sample_t sbuf[32],sbuf2[32],s;
     guint i,j;
     Dataformat fmt;

     /* Generate random data in pcm buffer */
     srand(time(0));     
     for (j=0; j<sizeof(pcm_buf); j++) 
	  pcm_buf[j] = rand();


     /* Perform tests */
     for (i=0; i<8; i++) {
	  fmt.samplesize = samplesizes[i];
	  fmt.sign = signs[i];
	  /* Test array funcs. */
	  convert_pcm_float_array(pcm_buf,&fmt,sbuf,32);
	  convert_float_pcm_array(sbuf,pcm_buf2,&fmt,32);
	  if (memcmp(pcm_buf,pcm_buf2,32*fmt.samplesize)) {
	       printf("Self-test failed for samplesize=%d, %s <array>\n",
		      samplesizes[i],signs[i]?"signed":"unsigned");
	  }
	  /* Test individual funcs. */
	  memset(pcm_buf2,0,sizeof(pcm_buf2));
	  for (j=0; j<32; j++) {
	       s = convert_pcm_float(pcm_buf+j*fmt.samplesize,&fmt);
	       convert_float_pcm(s,pcm_buf2+j*fmt.samplesize,&fmt);
	  }
	  if (memcmp(pcm_buf,pcm_buf2,32*fmt.samplesize)) {
	       printf("Self-test failed for samplesize=%d, %s\n",
		      samplesizes[i],signs[i]?"signed":"unsigned");
	  }

	  /* Test extreme values */
	  sbuf[0] = -1.0;
	  sbuf[1] = 1.0;
	  convert_float_pcm_array(sbuf,pcm_buf,&fmt,2);
	  convert_pcm_float_array(pcm_buf,&fmt,sbuf2,2);
	  if (memcmp(sbuf,sbuf2,2*sizeof(sample_t))) {
	       printf("Self-test failed for samplesize=%d, %s <array extremes>\n",
		      samplesizes[i],signs[i]?"signed":"unsigned");
	       printf("    (%f,%f) => (%f,%f)\n",sbuf[0],sbuf[1],sbuf2[0],sbuf2[1]);
	  }
	  for (j=0; j<2; j++) {
	       convert_float_pcm(sbuf[j],pcm_buf2,&fmt);
	       sbuf2[j] = convert_pcm_float(pcm_buf2,&fmt);
	  }
	  if (memcmp(sbuf,sbuf2,2*sizeof(sample_t))) {
	       printf("Self-test failed for samplesize=%d, %s <extremes>\n",
		      samplesizes[i],signs[i]?"signed":"unsigned");
	       printf("    (%f,%f) => (%f,%f)\n",sbuf[0],sbuf[1],sbuf2[0],sbuf2[1]);
	  }

     }
}

gboolean dataformat_equal(Dataformat *f1, Dataformat *f2)
{
	return (f1->samplerate==f2->samplerate && f1->samplesize==f2->samplesize && f1->channels==f2->channels && f1->sign==f2->sign);
}
