/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */

/* Datasource object represent a source of PCM data. */

/* This is (and should be) mainly used by chunk.c */

#ifndef DATASOURCE_H_INCLUDED
#define DATASOURCE_H_INCLUDED

#include <gtk/gtk.h>
#include "gtkfiles.h"

#define DATASOURCE(obj) GTK_CHECK_CAST(obj,datasource_get_type(),Datasource)
#define DATASOURCE_CLASS(klass) GTK_CHECK_CLASS_CAST(klass,datasource_get_type(),DatasourceClass)
#define IS_DATASOURCE(obj) GTK_CHECK_TYPE(obj,datasource_get_type())

#ifndef HAVE_LIBSNDFILE
#define SNDFILE void
#else
#include <sndfile.h>
#endif

/* struct describing the sample format. */

typedef struct {

     guint32 samplerate;         /* Samples/sec */
     guint samplesize;           /* bytes (1,2,3 or 4)*/
     gint channels;              /* Number of channels (1-16) */
     gboolean sign;              /* TRUE=signed, FALSE=unsigned */

     gint samplebytes;           /* should always be samplesize*channels */
     
} Dataformat;

/* Values for the type member of the Datasource struct. Determines how to 
 * actually get the data. */

#define DATASOURCE_REAL              0 /* Data stored directly in memory */
#define DATASOURCE_VIRTUAL           1 /* The data is stored in a (read-only)
					* file. */ 
#define DATASOURCE_TEMPFILE          2 /* Data stored in a file that will be
			                * removed when the chunk is
			                * destroyed. */ 
#define DATASOURCE_SILENCE           3 /* Silence (duh) */
#define DATASOURCE_SNDFILE           4 /* The data is read using libSndfile */
#define DATASOURCE_SNDFILE_TEMPORARY 5 /* Same as DATASOURCE_SNDFILE but will be
                                        * removed when the chunk is
                                        * destroyed. */ 
#define DATASOURCE_CLONE             6 /* Clone of another Datasource */

struct _Datasource;

struct _Datasource {

     GtkObject object;

     gint type;

     Dataformat format;

     guint32 length;             /*Number of samples */
     guint32 size;               /* Number of PCM bytes in file. */

     gboolean fake_pcm;          /* TRUE if the data is not real pcm. */
     
     guint opencount;            /* To keep track of nested open/close calls */

     gint tag;                   /* Extra field... */

     /* Type-specific data */
     union {

	  /* --- DATASOURCE_TEMPFILE and DATASOURCE_VIRTUAL --- */
	  struct {
	       char *filename; 
	       guint32 offset;  /* Offset of data in file (bytes) */
	       EFILE *handle;
	       guint32 pos;     /* Current position in file (bytes) */
	  } virtual;

	  /* --- DATASOURCE_REAL ---*/
	  char *real;  /* Pointer into the memory buffer with the sound data. */
	  

	  /* --- DATASOURCE_SNDFILE and DATASOURCE_SNDFILE_TEMPORARY --- */
	  struct {
	       char *filename;
	       gboolean raw_readable;
	       SNDFILE *handle;
	       guint32 pos; /* Current position in file (samples) */
	  } sndfile;

	  /* DATASOURCE_CLONE */
	  struct _Datasource *clone;
     } data;
};

typedef struct _Datasource Datasource;

typedef struct {
     GtkObjectClass klass;
} DatasourceClass;


#ifdef USE_DOUBLE_SAMPLES

typedef gdouble sample_t;
#define sf_readf_sample_t sf_readf_double
#define sf_writef_sample_t sf_writef_double

#else

typedef gfloat sample_t;
#define sf_readf_sample_t sf_readf_float
#define sf_writef_sample_t sf_writef_float

#endif

GtkType datasource_get_type(void);

/* ------------ 
 * CONSTRUCTORS 
 * ------------ */


/* Create data source from another Datasource object. The format of the data 
 * will be that of the format argument, regardless of what the other source
 * specifies. This makes it possible to override the format when you know it is
 * wrong. 
 */
Datasource *datasource_clone_df(Datasource *source, Dataformat *format);


/* Create a data source from a memory buffer of data. The buffer will be 
 * g_free:d when the data source is destroyed and the data should not be 
 * modified by the caller any more after this call.
 */
Datasource *datasource_new_from_data(void *data, Dataformat *format, 
				     guint32 size);


/* Create a new datasource containing silence with a certain format and the 
 * specified number of samples  
 */
Datasource *datasource_new_silent(Dataformat *format, guint32 samples);



/* ---------------------
 * DATA ACCESS FUNCTIONS
 * --------------------- */



/* Open the data source. Must be called before any of the other data access 
 * functions. May be called more than one time for the same Datasource.
 */

gboolean datasource_open(Datasource *source);



/* Close the data source. Must be called once for each call to 
 * datasource_open. 
 */

void datasource_close(Datasource *source);



/* Read one/many PCM/float sample from the Data source into the buffer */
gboolean datasource_read(Datasource *source, guint32 sampleno, gpointer buffer);
gboolean datasource_read_fp(Datasource *source, guint32 sampleno, 
			    sample_t *buffer);
guint datasource_read_array(Datasource *source, guint32 sampleno, 
			    guint size, gpointer buffer);
guint datasource_read_array_fp(Datasource *source, guint32 sampleno,
			       guint samples, sample_t *buffer);






/* -------------------
 * CONVERSION ROUTINES 
 * ------------------- */


/* These convert one or many samples between pcm formats and sample_t 
 * NOTE: These convert SINGLE-CHANNEL samples so you have to multiply the
 * count with the number of channels if you want to read full samples. 
 */

sample_t convert_pcm_float(void *pcm, Dataformat *pcm_format);
void convert_pcm_float_array(void *pcm, Dataformat *pcm_format, 
			     sample_t *result, guint count);
void convert_float_pcm(sample_t f, void *result, Dataformat *result_format);
void convert_float_pcm_array(sample_t *f_buf, void *result, 
			     Dataformat *result_format, guint count);

void datasource_conversion_selftest(void);


/* -------
 * VARIOUS 
 * ------- */


/* Returns the number of active Datasource objects */
guint datasource_count(void);


/* Creates a wav file header for the specified sample format and size and 
 * puts it in the supplied buffer. Size of the header is always 44 bytes. */

void get_wav_header(guint16 channels, guint16 samplesize, guint32 samplerate, 
		    guint32 datasize, void *buffer);

/* This dumps the PCM contents of the datasource into a file. You don't need
 * to open the datasource when using this function.
 */

gboolean datasource_dump(Datasource *ds, guint32 position, guint32 length,
			 EFILE *file);


/* This reads all the data of the datasource into a memory buffer and converts
 * it into a memory datasource.
 */

gboolean datasource_realize(Datasource *ds);

/* This function will unlink a file. If any of the Datasources depend on this
 * file, the file will be moved or copied into a temporary file before the
 * original name is unlinked.
 */

gboolean datasource_backup_unlink(gchar *filename);


/* Returns TRUE if the two formats are equal */

gboolean dataformat_equal(Dataformat *f1, Dataformat *f2);



/* -------- */
/* TEMPFILE */
/* -------- */

void get_wav_header(guint16 channels, guint16 samplesize, guint32 samplerate,
		    guint32 datasize, void *buffer);

gboolean write_wav_header(EFILE *f, guint16 channels, 
			  guint16 samplesize, guint32 samplerate, 
			  guint32 datasize);


/* This type identifies the temporary file during creation */
typedef gpointer TempFile;

/* Start creating a temporary file. Returns a handle to the temporary file or
 * NULL on failure. The handle must later be destroyed using either 
 * chunk_create_temp_finished or chunk_create_temp_abort.
 */

TempFile tempfile_init(void);



/* Write data into the temporary file. 
 *
 * temp - Handle to the temporary file.
 * data - The data to output
 * length - Length of the data
 *
 * return value - TRUE on failure, FALSE on success
 */

gboolean tempfile_write(TempFile temp, gpointer data, guint length);



/* Abort creating the temporary file. 
 *   temp - Handle to the temporary file.
 */

void tempfile_abort(TempFile temp);



/* End the temporary file.
 *   temp - Handle to the temporary file.
 *   format - The format the written data was in.
 *   return value - A Datasource containing the temporary data or NULL on
 *                  failure. 
 */

Datasource *tempfile_finished(TempFile temp, Dataformat *format);

#endif
