/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


/* --------------------------------------------------------------------------
 * This is the base class for most of the dialogs invoked from the Effects
 * menu. The class should be considered "abstract" and the functions in this
 * file should only be used from subclasses. 
 *
 * Each EffectDialog is associated with a Mainwindow and when the associated
 * Mainwindow is closed, the dialog will automatically close as well. 
 *
 * The EffectDialog has an input area where subclasses can place widgets for
 * setting effect parameters etc. There is also a button area with three
 * buttons: The OK button will emit the "apply" signal, then close the dialog,
 * the Apply button will also emit the "apply" signal but keep the dialog
 * active and the Close button will just close the dialog. Subclasses can add
 * their own buttons. 
 * -------------------------------------------------------------------------- */


#ifndef EFFECT_DIALOG_H_INCLUDED
#define EFFECT_DIALOG_H_INCLUDED

#include <gtk/gtk.h>
#include "mainwindow.h"

#define EFFECT_DIALOG(obj) GTK_CHECK_CAST(obj,effect_dialog_get_type(),EffectDialog)
#define EFFECT_DIALOG_CLASS(klass) GTK_CHECK_CLASS_CAST(klass,effect_dialog_get_type(),EffectDialogClass)
#define IS_EFFECT_DIALOG(obj) GTK_CHECK_TYPE(obj,effect_dialog_get_type())

typedef struct _EffectDialog {

     GtkWindow window;

     /* Input area - subclasses can add their own widgets here. */
     GtkContainer *input_area;

     /* Associated mainwindow. Read-only, set this with the
      * effect_dialog_set_mainwindow function. */
     Mainwindow *mw;

     /* private */
     GtkContainer *button_area;

} EffectDialog;



typedef struct _EffectDialogClass {
     GtkWindowClass window_class;

     /* This signal is emitted when the OK or Apply button is pressed. Default
      * action: do nothing. */
     void (*apply)(EffectDialog *ed);

     /* This signal is called when the associated Mainwindow has been set for the
      * dialog. */
     void (*mainwindow_set)(EffectDialog *ed);

} EffectDialogClass;

GtkType effect_dialog_get_type(void);

/* This sets the associated Mainwindow for the dialog. This must be called once
 * and only once for a dialog.
 *
 * ed - An EffectDialog
 * mw - The Mainwindow to associate with the EffectDialog.
 */
void effect_dialog_set_mainwindow(EffectDialog *ed, Mainwindow *mw);

/* This function adds a button to the EffectDialog's button area and connects
 * it's "clicked" signal to a callback.
 *
 * ed - An EffectDialog
 * label - The text to be displayed on the new button.
 * callback - The function to be called whenever the new button is clicked.
 *
 * returns - The new button.
 */ 
GtkWidget *effect_dialog_add_button(EffectDialog *ed, gchar *label, 
				    GtkSignalFunc callback);

#endif
