/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include <string.h>
#include "um.h"
#include "inifile.h"
#include "rawdialog.h"
#include "processdialog.h"
#include "main.h"
#include "filetypes.h"

#define MAX_REAL_SIZE inifile_get_guint32(INI_SETTING_REALMAX,INI_SETTING_REALMAX_DEFAULT)

struct file_type;
struct file_type {
     gchar *name;
     gchar *extension;
     gboolean (*typecheck)(gchar *filename);
     Datasource *(*load)(gchar *filename);
     Datasource *(*save)(Chunk *chunk, gchar *filename, 
			 struct file_type *type, gboolean *fatal);
     gint extra_data;
};

static GList *file_types = NULL;

static gboolean wav_check(gchar *filename);
static Datasource *wav_load(gchar *filename);
static Datasource *wav_save(Chunk *chunk, gchar *filename, 
			    struct file_type *type,
			    gboolean *fatal);

#ifdef HAVE_LIBSNDFILE
static gboolean sndfile_check(gchar *filename);
static Datasource *sndfile_load(gchar *filename);
static Datasource *sndfile_save(Chunk *chunk, gchar *filename, 
				struct file_type *type,
				gboolean *fatal);
#endif

static Datasource *raw_load(gchar *filename);
static Datasource *raw_save(Chunk *chunk, gchar *filename, 
			    struct file_type *type,
			    gboolean *fatal);

static void register_file_type(gchar *name, gchar *ext, 
			       gboolean (*typecheck)(gchar *filename),
			       Datasource *(*load)(gchar *filename),
			       Datasource *(*save)(Chunk *chunk, 
						   gchar *filename, 
						   struct file_type *type,
						   gboolean *fatal),
			       int extra_data)
{
     struct file_type *t;
     t = g_malloc(sizeof(*t));
     t->name = g_strdup(name);
     t->extension = g_strdup(ext);
     t->typecheck = typecheck;
     t->load = load;
     t->save = save;
     t->extra_data = extra_data;
     file_types = g_list_append(file_types,t);
}

static void setup_types(void)
{
#if defined(HAVE_LIBSNDFILE)
     SF_FORMAT_INFO info;
     int major_count,i;
     gchar buf[32];
#endif
     if (file_types) return;
     register_file_type("Microsoft WAV format", ".wav", wav_check, 
			wav_load, wav_save, 0);
#if defined(HAVE_LIBSNDFILE)
     sf_command(NULL, SFC_GET_FORMAT_MAJOR_COUNT, &major_count, sizeof(int));
     for (i=0; i<major_count; i++) {
	  info.format = i;
	  if (sf_command(NULL, SFC_GET_FORMAT_MAJOR, &info, sizeof(info))) {
	       sf_perror(NULL);
	       continue;
	  }
	  if ((info.format&SF_FORMAT_TYPEMASK) == SF_FORMAT_RAW ||
	      (info.format&SF_FORMAT_TYPEMASK) == SF_FORMAT_WAV) continue;
	  snprintf(buf,sizeof(buf),".%s",info.extension);
	  register_file_type((gchar *)info.name, buf, sndfile_check, 
			     sndfile_load, sndfile_save, 
			     info.format&SF_FORMAT_TYPEMASK);
     }
#endif
     register_file_type("Raw PCM data", ".raw", NULL, raw_load, 
			raw_save, 0);
}

guint fileformat_count(void)
{
     setup_types();
     return g_list_length(file_types);
}

gchar *fileformat_name(guint fileformat)
{
     struct file_type *t;
     t = g_list_nth_data(file_types, fileformat);
     if (t) return t->name;
     return NULL;
}

gchar *fileformat_extension(guint fileformat)
{
     struct file_type *t;
     t = g_list_nth_data(file_types, fileformat);
     if (t) return t->extension;
     return NULL;
}

Chunk *chunk_load(gchar *filename)
{
     GList *l;
     struct file_type *ft;
     gchar *c;
     setup_types();
     /* First, see if the file exists */
     if (!file_exists(filename)) {
	  c = g_strdup_printf("The file %s does not exist!",filename);
	  user_message(c,UM_OK);
	  g_free(c);
	  return NULL;
     }
     if (!file_is_normal(filename)) {
	  c = g_strdup_printf("The file %s is not a regular file!",filename);
	  user_message(c,UM_OK);
	  g_free(c);
	  return NULL;	  
     }
     /* Search through the extensions */
     for (l=file_types; l!=NULL; l=l->next) {
	  ft = (struct file_type *)l->data;
	  if (!ft->extension) continue;
	  c = strchr(filename,0) - strlen(ft->extension);
	  if (c<filename || strcmp(c,ft->extension)!=0) continue;
	  if (ft->typecheck==NULL || ft->typecheck(filename)) 
	       return chunk_new_from_datasource(ft->load(filename));
     }
     /* Use the file checking functions */
     for (l=file_types; l!=NULL; l=l->next) {
	  ft = (struct file_type *)l->data;
	  if (ft->typecheck == NULL) continue;
	  if (ft->typecheck(filename))
	       return chunk_new_from_datasource(ft->load(filename));
     }
     /* Use the raw loader */
     return chunk_new_from_datasource(raw_load(filename));
}

static Chunk *chunk_save_main(Chunk *chunk, gchar *filename, gboolean *fatal)
{
     struct file_type *ft;
     GList *l;
     gchar *c, **z, **z2;
     guint32 i;
     gint x;
     setup_types();
     /* Search through the extensions */
     for (l=file_types; l!=NULL; l=l->next) {
	  ft = (struct file_type *)l->data;
	  if (!ft->extension) continue;
	  c = strchr(filename,0) - strlen(ft->extension);
	  if (c<filename || strcmp(c,ft->extension)!=0) continue;
	  if (ft->save != NULL)
	       return chunk_new_from_datasource(
		    ft->save(chunk,filename,ft,fatal));
     }
     /* Ask for file type */
     z = g_malloc((g_list_length(file_types)+1) * sizeof(char *));
     for (l=file_types,z2=z; l!=NULL; l=l->next,z2++) {
	  ft = (struct file_type *)l->data;
	  *z2 = ft->name;
     }
     *z2 = NULL;
     c = g_strdup_printf("The file name '%s' has an extension unknown to the "
			 "program. Please specify in which format this file "
			 "should be saved.",filename);
     i = inifile_get_guint32("lastSaveFileType",0);
     if (i>=g_list_length(file_types)) i=0;
     x = user_choice(z, (guint)i, "Unknown file type", c, TRUE);
     g_free(z);
     g_free(c);
     if (x==-1) return NULL;
     inifile_set_guint32("lastSaveFileType",(guint32)x);
     ft = g_list_nth_data( file_types, x );
     return chunk_new_from_datasource(ft->save(chunk,filename,ft,fatal));
}

Chunk *chunk_save(Chunk *chunk, gchar *filename)
{
     Chunk *res;
     gchar *d;
     gboolean b=FALSE;

     /* we may need to move or copy the file into a tempfile... */
     if (file_exists(filename) && datasource_backup_unlink(filename)) 
	  return NULL;

     res = chunk_save_main(chunk,filename,&b);

     if (!res && b) {
	  d = g_strdup_printf("Warning: The file %s may be destroyed since the "
			      "saving failed. Try to free up some disk space "
			      "and save again. If you exit now the "
			      "file's contents could be in a bad state. ",
			      filename);
	  user_message(d,UM_OK);
	  g_free(d);
     }

     return res;
}

Chunk *chunk_save_ff(Chunk *chunk, gchar *filename,
		     guint fileformat)
{
     struct file_type *ft;
     gboolean b;
     setup_types();
     ft = g_list_nth_data( file_types, fileformat );
     g_assert (ft != NULL);
     return chunk_new_from_datasource(ft->save(chunk, filename, ft, &b));
}


/* WAV */


static gboolean wav_check(gchar *filename)
{
     EFILE *f;
     char buf[12];
     f = e_fopen ( filename, EFILE_READ );
     if (!f) return FALSE;
     if (e_fread(buf,12,f)) { e_fclose(f); return FALSE; }
     e_fclose(f); 
     if (memcmp(buf,"RIFF",4) || memcmp(buf+8,"WAVE",4)) return FALSE; 
     return TRUE;     
}

static long int find_wav_chunk(EFILE *f, char *chunkname)
{
     char buf[4];
     unsigned long l;
     while (1) {
	  if (e_fread(buf,4,f) ||
	      e_fread_le(&l,4,f)) return -1;
	  if (!memcmp(buf,chunkname,4)) 
	       return (l>=0x80000000) ? 0x7FFFFFFFL : l;
	  if (e_fseek(f,l,SEEK_CUR)) return -1;
     }     
}

static Datasource *wav_load(char *filename)
{
     EFILE *f;
     char *c;
     char buf[12];
     long int m,e;
     unsigned short int s,channels,bits;
     unsigned long int samplerate,l;
     Datasource *ds;
     Dataformat fmt;
     f = e_fopen ( filename, EFILE_READ );
     if (!f) return NULL;
     if (e_fread(buf,12,f)) { e_fclose(f); return NULL; }
     if (memcmp(buf,"RIFF",4) || memcmp(buf+8,"WAVE",4)) {
	  c = g_strdup_printf ("%s is not a valid wav file",filename);
	  user_message(c,UM_OK);
	  g_free(c);
	  e_fclose(f);
	  return NULL;
     }
     l = find_wav_chunk(f,"fmt ");
     if ((l == -1) || e_fread_le(&s,2,f)) { e_fclose(f); return NULL; }
     if ( s != 0x0001 ) {
	  e_fclose(f);
#if defined(HAVE_LIBSNDFILE)
	  /* Fall back on Libsndfile for reading non-PCM wav files. */
	  return sndfile_load(filename);
#else
	  c = g_strdup_printf ("The file %s is a compressed wav file. This "
			       "program can only work with uncompressed wav "
			       "files.",filename);
	  user_message(c,UM_OK);
	  free(c);
	  return NULL;
#endif
     }
     if (e_fread_le(&channels,2,f) ||
	 e_fread_le(&samplerate,4,f) ||
	 e_fseek(f,6,SEEK_CUR) ||
	 e_fread_le(&bits,2,f) ||
	 (l>16 && e_fseek(f,l-16,SEEK_CUR))) { 
	  e_fclose(f); 
	  return NULL; 
     }
     l = find_wav_chunk(f,"data");
     m = e_ftell(f);
     if (l == -1 || m == -1) {
	  e_fclose(f);
	  return NULL;
     }
     /* Om filen r fr kort fr att rymma hela data-chunken, kapa av s hgt 
	som mjligt. */
     if (e_fseek(f,0,SEEK_END)) { e_fclose(f); return NULL; }
     e = e_ftell(f);
     if (e == -1) { e_fclose(f); return NULL; }
     if (e-m<l) l=e-m; /* Borde det vara l=e-m+1?? */
     e_fclose(f);

     fmt.samplerate = samplerate;
     fmt.samplesize = (bits+7) >> 3;
     fmt.channels = channels;
     fmt.sign = (bits>8);
     fmt.samplebytes = fmt.samplesize * fmt.channels;

     ds = gtk_type_new(datasource_get_type());

     memcpy(&(ds->format),&fmt,sizeof(fmt));
     ds->type = DATASOURCE_VIRTUAL;
     ds->size = l;
     ds->length = l / fmt.samplebytes;
     ds->data.virtual.filename = g_strdup(filename);
     ds->data.virtual.offset = m;

     if (ds->size <= MAX_REAL_SIZE)
	  datasource_realize(ds);

     return ds;
}

static Datasource *wav_save(Chunk *chunk, char *filename, 
			    struct file_type *type, gboolean *fatal)
{
     EFILE *f;
     Datasource *ds;
     DataPart *dp;
     Dataformat *fmt;
     fmt = &(chunk->format);
     /* Check if the format is OK */
     if (fmt->samplesize==1 && fmt->sign) {
	  user_message("8-bit wav-files must be in unsigned format!",UM_OK);
	  return NULL;
     }
     if (fmt->samplesize>1 && !fmt->sign) {
	  user_message("16/24/32-bit wav-files must be in signed format!",
		       UM_OK);
	  return NULL;
     }
     /* If we're saving an entire TEMPFILE and it's in WAV format, we can just 
      * create a link (this speeds up things like saving directly after 
      * recording. */
     dp = (DataPart *)chunk->parts->data;
     ds = dp->ds;
     if (chunk->parts->next == NULL &&
	 dp->position == 0 && dp->length == ds->length && 
	 (ds->type == DATASOURCE_TEMPFILE || ds->type == DATASOURCE_VIRTUAL) && 
	 ds->data.virtual.offset == 44 && 
	 ds->size == errdlg_filesize(ds->data.virtual.filename)-44 &&
	 wav_check(ds->data.virtual.filename)) {	  
	  if (xunlink(filename) || xlink(ds->data.virtual.filename,filename))
	       return NULL;
     } else {
	  /* Regular saving */
	  f = e_fopen(filename,EFILE_WRITE);
	  if (!f) return NULL;
	  *fatal = TRUE;
	  
	  if (write_wav_header(f,chunk->format.channels,
			       chunk->format.samplesize,
			       chunk->format.samplerate,chunk->size)) {
	       e_fclose(f);
	       return NULL;
	  }
	  if (chunk_dump(chunk,f)) {
	       e_fclose_remove(f);
	       return NULL;
	  }
	  e_fclose ( f );

     }    
     ds = gtk_type_new(datasource_get_type());
     memcpy(&(ds->format),fmt,sizeof(Dataformat));
     ds->type = DATASOURCE_VIRTUAL;
     ds->length = chunk->length;
     ds->size = ds->length * ds->format.samplebytes;
     ds->data.virtual.filename = g_strdup(filename);
     /* 'RIFF'+lngd+'WAVE'+'fmt '+lngd+fmt(16)+'data'+lngd == 44 bytes*/ 
     ds->data.virtual.offset = 44;
     if (ds->length <= MAX_REAL_SIZE) datasource_realize(ds);
     return ds;
}

/* RAW */

static Datasource *raw_load(gchar *filename)
{
     Datasource *ds;
     Dataformat *fmt;
     gint32 i;
     fmt = rawdialog_execute(filename);
     if (!fmt) return NULL;
     i = errdlg_filesize(filename);
     if (i==-1) return NULL;
     ds = (Datasource *)gtk_type_new(datasource_get_type());
     memcpy(&(ds->format),fmt,sizeof(Dataformat));
     ds->size = i;
     ds->length = i/fmt->samplebytes;
     ds->type = DATASOURCE_VIRTUAL;
     ds->data.virtual.filename = g_strdup(filename);
     ds->data.virtual.offset = 0;
     return ds;
}

static Datasource *raw_save(Chunk *chunk, gchar *filename, 
			    struct file_type *type, gboolean *fatal)
{
     EFILE *f;
     Datasource *ds;
     f = e_fopen(filename,EFILE_WRITE);
     if (!f) return NULL;
     if (chunk_dump(chunk,f)) {
          e_fclose(f);
	  *fatal = TRUE;
          return NULL;
     }
     e_fclose(f);
     ds = gtk_type_new(datasource_get_type());
     memcpy(&(ds->format),&(chunk->format),sizeof(Dataformat));
     ds->length = chunk->length;
     ds->size = ds->length * ds->format.samplebytes;
     ds->type = DATASOURCE_VIRTUAL;
     ds->data.virtual.filename = g_strdup(filename);
     ds->data.virtual.offset = 0;
     return ds;
}

/* SNDFILE */

#if defined(HAVE_LIBSNDFILE)

static gboolean sndfile_check(gchar *filename)
{
     SF_INFO info = {};
     SNDFILE *s;
     s = sf_open(filename, SFM_READ, &info);
     if (s) sf_close(s);
     return (s!=NULL && ((info.format&SF_FORMAT_SUBMASK) != SF_FORMAT_RAW));
}

static Datasource *sndfile_load(gchar *filename)
{
     SF_INFO info = {};
     SNDFILE *s;
     Datasource *ds;
     Dataformat f;
     gchar *ch,b[256];
     int rawbuf_size;
     sample_t *samplebuf = NULL;
     void *rawbuf;
     size_t st;
     gboolean raw_readable, fake_pcm=FALSE;
     ChunkTemp tmp;
     s = sf_open(filename, SFM_READ, &info);
     if (!s) {
          ch = g_strdup_printf("Failed to open '%s'!",filename);
          user_message(ch,UM_OK);
          g_free(ch);
          return NULL;
     }
     /* printf("info.seekable = %d, info.samples = %d\n",info.seekable,info.samples); */
     f.samplerate = info.samplerate;
     f.channels = info.channels;
     /* Fix samplesize parameter */
     /* Fixme: How can you tell whether a SNDFILE with PCM data has same 
      * endian-ness as the CPU? Apparently you can't right now. */
     /* printf("%d\n",info.format&SF_FORMAT_ENDMASK); */
     switch (info.format&SF_FORMAT_SUBMASK) {
     case SF_FORMAT_PCM_U8: f.sign=FALSE;f.samplesize=1;raw_readable=TRUE;break;
     case SF_FORMAT_PCM_S8: f.sign=TRUE; f.samplesize=1;raw_readable=TRUE;break;
     case SF_FORMAT_PCM_16: f.sign=TRUE; f.samplesize=2;raw_readable=FALSE;break;
     case SF_FORMAT_PCM_24: f.sign=TRUE; f.samplesize=3;raw_readable=FALSE;break;
     case SF_FORMAT_PCM_32: f.sign=TRUE; f.samplesize=4;raw_readable=FALSE;break;
	  /* Default to 16 bit. Fixme: Make this configurable */
     default: f.sign = TRUE; f.samplesize = 2; raw_readable = FALSE; 
	  fake_pcm = TRUE; break; 
     }
     f.samplebytes = f.channels * f.samplesize;
     /* Handle seekable chunks */
     if (info.seekable) {
          sf_close(s);
	  ds = gtk_type_new(datasource_get_type());
	  ds->type = DATASOURCE_SNDFILE;
	  memcpy(&(ds->format),&f,sizeof(f));
	  ds->length = info.frames;
	  ds->size = ds->length * f.samplebytes;
	  ds->fake_pcm = fake_pcm;
          ds->data.sndfile.filename = g_strdup(filename);
	  ds->data.sndfile.raw_readable = raw_readable;	  
	  return ds;
     }
     /* Chunk is not seekable, we must dump it into a tempfile. */
     /* puts("Non-native format!!"); */
     tmp = tempfile_init();
     processdialog_setup(info.frames);
     if (!raw_readable)
	  samplebuf = g_malloc(sizeof(sample_t) * f.channels * 1024);
     rawbuf_size = f.samplesize * f.channels * 1024;
     rawbuf = g_malloc(rawbuf_size);
     while (1) {

	  if (raw_readable)
	       st = sf_read_raw(s, rawbuf, rawbuf_size);
	  else
	       st = sf_readf_sample_t(s, samplebuf, 1024);

	  if (st == 0) break;
	  if (st == -1) {
	       sf_error_str(s,b,sizeof(b));
	       ch = g_strdup_printf("Error reading %s: %s",filename,b);
	       user_message(ch,UM_OK);
	       g_free(ch);
	       sf_close(s);
	       tempfile_abort(tmp);
	       g_free(samplebuf);
	       g_free(rawbuf);
	       return NULL;
	  }
	  
	  if (!raw_readable) 
	       convert_float_pcm_array(samplebuf,rawbuf,&f,st*f.channels);
	  else
	       st /= f.samplebytes; /* bytes -> frames */

	  if (tempfile_write(tmp,rawbuf,st*f.channels*f.samplesize)) {
		    sf_close(s);
		    tempfile_abort(tmp);
		    g_free(samplebuf);
		    g_free(rawbuf);
		    return NULL;
	  }
	  if (processdialog_progress(st)) {
	       sf_close(s);
	       tempfile_abort(tmp);
	       g_free(samplebuf);
	       g_free(rawbuf);
	       return NULL;
	  }
     }
     sf_close(s);
     g_free(samplebuf);
     g_free(rawbuf);
     return tempfile_finished(tmp,&f);
}


static gboolean find_nearest_sndfile_format(Dataformat *fmt, int format, 
					    SF_INFO *info, gchar *filename)
{
     int suggestions[] = { SF_FORMAT_PCM_U8, SF_FORMAT_PCM_S8, SF_FORMAT_PCM_16,
			   SF_FORMAT_PCM_24, SF_FORMAT_PCM_32, SF_FORMAT_FLOAT,
                           SF_FORMAT_DOUBLE };
     int i,j,k;
     SF_INFO sfi;
     info->samplerate = fmt->samplerate;
     info->channels = fmt->channels;
     /* Try suggestions first... */
     i = fmt->samplesize;
     if (i == 1) i--;
     for (; i<ARRAY_LENGTH(suggestions); i++) {
	  info->format = format | suggestions[i];
	  /* printf("Trying format %x samplerate %d channels %d...\n",info->format,
	     info->samplerate,info->channels); */
	  if (sf_format_check(info)) return FALSE;
     }
     /* Now try the formats that were not in the suggestion array... */
     sf_command(NULL, SFC_GET_FORMAT_SUBTYPE_COUNT, &i, sizeof(int));
     for (j=0; j<i; j++) {
	  sfi.format = j;
	  sf_command(NULL, SFC_GET_FORMAT_SUBTYPE, &sfi, sizeof(sfi));
	  for (k=0; k<ARRAY_LENGTH(suggestions); k++)
	       if ((sfi.format & SF_FORMAT_SUBMASK) == suggestions[k])
		    goto outer;
	  info->format = (sfi.format&SF_FORMAT_SUBMASK) | format;
	  /* printf("Trying format %x samplerate %d channels %d...\n",
	     info->format,
	     info->samplerate,info->channels); */
	  if (sf_format_check(info)) return FALSE;
     outer:
     }
     /* Try the left out suggestions finally... */
     i = fmt->samplesize;
     if (i > 1) {
	  i--;
	  for (; i>=0; i--) {
	       info->format = format | suggestions[i];
	       /* printf("Trying format %x samplerate %d channels %d...\n",
		  info->format,
		  info->samplerate,info->channels);
	       */
	       if (sf_format_check(info)) return FALSE;
	  }
     }

     user_message("Invalid sample format or number of channels for this file"
		  " format",UM_OK);
     return TRUE;
     
}


static Datasource *sndfile_save(Chunk *chunk, gchar *filename, 
				struct file_type *type, 
				gboolean *fatal)
{
     gchar *c;
     guint32 i;
     guint n;
     SNDFILE *s;
     SF_INFO info;
     sample_t *samplebuf;
     ChunkHandle *ch;

     if (find_nearest_sndfile_format(&(chunk->format),type->extra_data,&info,
				     filename)) 
	  return NULL;
     
     s = sf_open(filename, SFM_WRITE, &info);
     if (!s) {
          c = g_strdup_printf("Failed to open '%s'!",filename);
          user_message(c,UM_OK);
          g_free(c);
          return NULL;
     }

     ch = chunk_open(chunk);
     if (ch == NULL) {
	  sf_close(s);
	  return NULL;
     }

     /* We could speed this up a LOT esp. on output to PCM files. */
     samplebuf = g_malloc(sizeof(sample_t) * chunk->format.channels * 1024);

     for (i=0; i<chunk->length; i+=n) {
	  n = chunk_read_array_fp(ch,i,1024,samplebuf);
	  if (!n) {
	       chunk_close(ch);
	       sf_close(s);
	       g_free(samplebuf);
	       *fatal = TRUE;
	       return NULL;
	  }
	  if (sf_writef_sample_t(s,samplebuf,n) != n) {
	       c = g_strdup_printf("Failed to write to '%s'!",filename);
	       user_message(c,UM_OK);
	       g_free(c);
	       chunk_close(ch);
	       sf_close(s);
	       g_free(samplebuf);
	       *fatal = TRUE;
	       return NULL;  
	  }
     }
     chunk_close(ch);
     sf_close(s);
     g_free(samplebuf);

     return sndfile_load(filename);

}

#endif

