/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include "um.h"
#include "gtkfiles.h"
#include "main.h"

EFILE *e_fopen(char *filename, int mode)
{
     EFILE *e;
     FILE *f;
     char *c;
     switch (mode) {
     case EFILE_READ:   c = "r"; break;
     case EFILE_WRITE:  c = "w"; break;
     case EFILE_APPEND: c = "a"; break;	
     default: g_assert_not_reached(); return NULL;
     }
     f = fopen(filename,c);
     if (f == NULL) {
	  c = g_strdup_printf("Could not open %s: %s",filename,strerror(errno));
	  user_message(c,UM_OK);
	  g_free(c);
	  return NULL;
     }
     e = g_malloc(sizeof(*e));
     e->handle = f;
     e->filename = g_strdup(filename);
     return e;
}

gboolean e_fclose(EFILE *f)
{
     gchar *c;
     gboolean b=FALSE;
     if (fclose(f->handle) != 0) {
	  c = g_strdup_printf("Error closing %s: %s", f->filename, strerror(errno));
	  user_message(c,UM_OK);	  
	  g_free(c);
	  b = TRUE;
     }
     g_free(f->filename);
     g_free(f);
     return b;
}

gboolean e_fclose_remove(EFILE *f)
{
     gchar *c;
     gboolean b=FALSE;
     if (fclose(f->handle) != 0) {
	  c = g_strdup_printf("Error closing %s: %s", f->filename, strerror(errno));
	  user_message(c,UM_OK);	  
	  g_free(c);
	  b=TRUE;
     }
     if (xunlink(f->filename)) b=TRUE;
     g_free(f->filename);
     g_free(f);
     return b;
}

gboolean e_fseek(EFILE *stream, long int offset, int whence)
{
     char *c;
     if (fseek(stream->handle,offset,whence) != 0) {
	  c = g_strdup_printf("Could not seek in %s: %s",stream->filename,
			      strerror(errno));
	  user_message(c,UM_OK);
	  g_free(c);
	  return TRUE;
     }
     return FALSE;
}

gboolean e_fread(void *data, size_t size, EFILE *stream)
{
     char *c;
     if (fread(data,size,1,stream->handle) != 1) {
	  c = g_strdup_printf("Could not read from %s: %s",stream->filename,
			      strerror(errno));
	  user_message(c,UM_OK);
	  g_free(c);
	  return TRUE;
     }
     return FALSE;
}

gboolean e_fwrite(void *data, size_t size, EFILE *stream)
{
     char *c;
     if (!size) return FALSE;
     if (fwrite(data,size,1,stream->handle) != 1) {
	  c = g_strdup_printf("Could not write to %s: %s",stream->filename,
			      strerror(errno));
	  user_message(c,UM_OK);
	  g_free(c);
	  return TRUE;
     }
     return FALSE;
}

gboolean e_fwrite0(size_t size, EFILE *stream)
{
     char buf[4096];
     size_t s;
     memset(buf,0,sizeof(buf));
     while (size>0) {
	  s = MIN(size,sizeof(buf));
	  if (e_fwrite(buf,s,stream)) return TRUE;
	  size -= s;
     }
     return FALSE;     
}

gboolean e_fread_bswap(void *data, size_t size, EFILE *stream)
{
     size_t i;
     gchar c;
     if (e_fread(data,size,stream)) return TRUE;
     i=0;
     size--;
     while (i<size) {
	  c = ((gchar *)data)[i];
	  ((gchar *)data)[i] = ((gchar *)data)[size];
	  ((gchar *)data)[size] = c;
	  i++;
	  size--;	  
     }
     return FALSE;
}

gboolean e_fwrite_bswap(void *data, size_t size, EFILE *stream)
{
     static gchar *buf = NULL;
     static size_t bufsize = 0;
     size_t i;
     if (size>bufsize) { buf=g_realloc(buf,size); bufsize=size; }
     for (i=0; i<size; i++) buf[i]=((gchar *)data)[size-(1+i)];
     return e_fwrite(buf,size,stream);
}

long int e_ftell(EFILE *stream)
{
     long int l;
     char *c;
     l = ftell(stream->handle);
     if (l == -1) {
	  c = g_strdup_printf("Could not get file position in %s: %s",
			      stream->filename, strerror(errno));
	  user_message(c,UM_OK);
	  g_free(c);
     }
     return l;
}

gboolean is_same_file(char *filename1, char *filename2)
{
     struct stat s1, s2;
     if (!strcmp(filename1,filename2)) return TRUE;
     if (stat(filename1,&s1)) {
	  if (errno == ENOENT) return FALSE;
	  else return TRUE;
     }
     if (stat(filename2,&s2)) {
	  if (errno == ENOENT) return FALSE;
	  else return TRUE;
     }
     return (s1.st_dev==s2.st_dev && s1.st_ino==s2.st_ino);
}

static gchar *get_filename_result;
static gboolean get_filename_quitflag;

static void get_filename_callback(GtkButton *button, gpointer user_data)
{
if (button==GTK_BUTTON(GTK_FILE_SELECTION(user_data)->ok_button)) {
	if (get_filename_result) free(get_filename_result);
        get_filename_result = g_strdup(gtk_file_selection_get_filename(
					    GTK_FILE_SELECTION(user_data)));
	}
}

static void get_filename_destroy(void)
{
     get_filename_quitflag = TRUE;
}

gchar *get_filename(gboolean savemode, gchar *current_name, gchar *filemask )
{
     GtkFileSelection *f;
     get_filename_result=NULL;
     f=GTK_FILE_SELECTION(gtk_file_selection_new(savemode?"Save file":"Load file"));
     if (current_name) gtk_file_selection_set_filename(f,current_name);
     if (filemask) gtk_file_selection_complete(f,filemask);
     gtk_signal_connect(GTK_OBJECT(f),"destroy",
			GTK_SIGNAL_FUNC(get_filename_destroy),NULL);
     gtk_signal_connect(GTK_OBJECT(f->ok_button),"clicked",
			GTK_SIGNAL_FUNC(get_filename_callback),f);
     gtk_signal_connect(GTK_OBJECT(f->cancel_button),"clicked",
			GTK_SIGNAL_FUNC(get_filename_callback),f);
     gtk_signal_connect_object(GTK_OBJECT(f->ok_button),"clicked",
			       GTK_SIGNAL_FUNC(gtk_widget_destroy),GTK_OBJECT(f));
     gtk_signal_connect_object(GTK_OBJECT(f->cancel_button),"clicked",
			       GTK_SIGNAL_FUNC(gtk_widget_destroy),GTK_OBJECT(f));
     gtk_window_set_modal(GTK_WINDOW(f),TRUE);
     gtk_widget_show(GTK_WIDGET(f));
     get_filename_quitflag=FALSE;
     while (!get_filename_quitflag) mainloop();
     return get_filename_result;     
}

gboolean e_copydata(EFILE *from, EFILE *to, guint32 bytes)
{
     guint32 count;
     size_t rest;
     guint8 buf[4096];
     count = bytes / sizeof(buf);
     rest = bytes % sizeof(buf);
     for (; count>0; count--)
	  if (e_fread(buf,sizeof(buf),from) ||
	      e_fwrite(buf,sizeof(buf),to)) return TRUE;
     if (rest>0) 
	  return (e_fread(buf,rest,from) || 
		  e_fwrite(buf,rest,to));
     else
	  return FALSE;
}

gint32 errdlg_filesize(gchar *filename)
{
     EFILE *f;
     gint32 r;
     f = e_fopen(filename,EFILE_READ);
     if (f == NULL) return -1;
     if (e_fseek(f,0,SEEK_END)) { e_fclose(f); return -1; }
     r = e_ftell(f);
     e_fclose(f);
     return r;
}

gboolean errdlg_copyfile(gchar *from, gchar *to)
{
     guint32 l;
     EFILE *in, *out;
     gboolean b;
     l = errdlg_filesize ( from );
     if ( l == -1 ) return TRUE;
     in = e_fopen(from,EFILE_READ);
     if (!in) return TRUE;
     out = e_fopen(to,EFILE_WRITE);
     if (!out) { e_fclose(in); return TRUE; }
     b = e_copydata(in,out,l);
     e_fclose(in);
     if (b) e_fclose_remove(out);
     else e_fclose(out);
     return b;
}

gboolean file_exists(char *filename)
{
     FILE *f;
     f = fopen(filename,"r");
     if (!f) return (errno != ENOENT);
     fclose(f);
     return TRUE;
}

gboolean file_is_normal(char *filename)
{
     struct stat s;
     int i;
     i = stat(filename,&s);
     return (i==0 && S_ISREG(s.st_mode));
}

long int e_readline(gchar **line, size_t *size, EFILE *stream)
{
     size_t s = 0;
     int c;
     while (1) {
	  if (s == *size) {
	       *size = *size ? *size * 2 : 32;
	       *line = g_realloc(*line, *size);
	  }
	  c = fgetc(stream->handle);
	  if (c==EOF || (c=='\n' && s>0)) { (*line)[s]=0; return s; }
	  if (c=='\n') return e_readline(line,size,stream);
	  (*line)[s] = (gchar)c;
	  s++;
     }
}

gboolean xunlink(gchar *filename)
{
     gchar *c;
     if (unlink(filename) == -1 && errno != ENOENT) {
	  c = g_strdup_printf("Could not remove '%s': %s",filename,
			      strerror(errno));
	  user_message(c,UM_OK);
	  g_free(c);
	  return TRUE;
     }
     return FALSE;
}

gboolean xlink(gchar *oldname, gchar *newname)
{
     gchar *c;
     if (link(oldname,newname) != 0) {
	  if (errno == EXDEV) return errdlg_copyfile(oldname,newname);
	  c = g_strdup_printf("Error creating link to '%s': %s",oldname,
			      strerror(errno));
	  user_message(c,UM_OK);
	  g_free(c);
	  return TRUE;
     } 
     return FALSE;
}
