/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <fcntl.h>

#define GTK_ENABLE_BROKEN /* for GtkText under GTK 2.x */
#include <gtk/gtk.h>

#include "main.h"
#include "pipedialog.h"
#include "um.h"
#include "processdialog.h"

extern char **environ;

static Chunk *pipe_dialog_apply_proc(Chunk *chunk, Mainwindow *mw, 
				     gpointer user_data)
{
     return pipe_dialog_pipe_chunk(chunk,(gchar *)user_data);
}

static void pipe_dialog_apply(EffectDialog *ed)
{
     PipeDialog *pd = PIPE_DIALOG(ed);
     gchar *cmd;
     cmd = history_box_get_value(pd->cmd);
     mainwindow_effect_manual(ed->mw,pipe_dialog_apply_proc,TRUE,cmd);
     history_box_rotate_history(pd->cmd);
}

static void pipe_dialog_init(PipeDialog *pd)
{
     GtkWidget *a,*b;
     a = gtk_hbox_new(FALSE,0);
     gtk_container_add(EFFECT_DIALOG(pd)->input_area,a);
     gtk_widget_show(a);
     b = gtk_label_new("Command line: ");
     gtk_container_add(GTK_CONTAINER(a),b);
     gtk_widget_show(b);
     b = history_box_new("PipeDialog","");
     gtk_container_add(GTK_CONTAINER(a),b);
     gtk_widget_show(b);
     pd->cmd = HISTORY_BOX(b);

     gtk_window_set_title(GTK_WINDOW(pd),"Pipe through command");
}

static void pipe_dialog_class_init(EffectDialogClass *edc)
{
     edc->apply = pipe_dialog_apply;
}

GtkType pipe_dialog_get_type(void)
{
     static GtkType id = 0;
     if (!id) {
	  GtkTypeInfo info = {
	       "PipeDialog",
	       sizeof(PipeDialog),
	       sizeof(PipeDialogClass),
	       (GtkClassInitFunc) pipe_dialog_class_init,
	       (GtkObjectInitFunc) pipe_dialog_init
	  };
	  id = gtk_type_unique(effect_dialog_get_type(),&info);
     }
     return id;
}

GtkWidget *pipe_dialog_new(Mainwindow *mw)
{
     GtkWidget *w;
     w = gtk_type_new(pipe_dialog_get_type());
     effect_dialog_set_mainwindow(EFFECT_DIALOG(w),mw);
     return w;
}

static void add_error_text(GtkText *t,gchar *text)
{
     GtkWidget *w = GTK_WIDGET(t);
     gtk_text_insert(t,gtk_style_get_font(w->style),w->style->fg,
		     &(w->style->white),text,-1);     
}

static GtkText *create_error_window(gchar *command)
{
     GtkWidget *a,*b,*c;
     GtkText *t;
     gchar *q;

     a = gtk_window_new(GTK_WINDOW_DIALOG);
     gtk_window_set_title(GTK_WINDOW(a),"Process output");
     gtk_widget_set_usize(GTK_WIDGET(a),470,200);
     b = gtk_vbox_new(FALSE,0);
     gtk_container_add(GTK_CONTAINER(a),b);
     c = gtk_text_new(NULL,NULL);
     t = GTK_TEXT(c);
     gtk_container_add(GTK_CONTAINER(b),c);
     c = gtk_button_new_with_label("Close");
     gtk_signal_connect_object(GTK_OBJECT(c),"clicked",
			       GTK_SIGNAL_FUNC(gtk_widget_destroy),
			       GTK_OBJECT(a));
     gtk_box_pack_start(GTK_BOX(b),c,FALSE,FALSE,0);

     q = g_strdup_printf("Output from command '%s':\n\n",command);
     add_error_text(t,q);
     g_free(q);
     
     gtk_widget_show_all(a);
     
     return t;
}

#define BZ 65536

Chunk *pipe_dialog_pipe_chunk(Chunk *chunk, gchar *command)
{
     int cmd_in[2]={},cmd_out[2]={},cmd_err[2]={},bs,bp,i;
     gboolean writing=TRUE;
     guint32 ui,read_bytes=0;
     gchar *c=NULL;
     unsigned char syncbyte;
     char *argv[4];
     GtkText *errortext = NULL;
     ChunkHandle *ch=NULL;
     ChunkTemp ct=0;
     struct timeval notime={};
     pid_t p;
     fd_set rset,wset;
     /* Create pipes */
     if (pipe(cmd_in)==-1 || pipe(cmd_out)==-1 || pipe(cmd_err)==-1) {
	  c = g_strdup_printf("Could not create pipe: %s",strerror(errno));
	  user_message(c,UM_OK);
	  g_free(c);
	  if (!cmd_in[0]) close(cmd_in[0]);
	  if (!cmd_in[1]) close(cmd_in[1]);
	  if (!cmd_out[0]) close(cmd_out[0]);
	  if (!cmd_out[1]) close(cmd_out[1]);
	  if (!cmd_err[0]) close(cmd_err[0]);
	  return NULL;
     }
     /* Fork */
     p = fork();

     if (p == -1) {
	  /* Error (parent process) */
	  c = g_strdup_printf("Error: fork: %s",strerror(errno));
	  user_message(c,UM_OK);
	  g_free(c);
	  close(cmd_in[0]);
	  close(cmd_in[1]);
	  close(cmd_out[0]);
	  close(cmd_out[1]);
	  close(cmd_err[0]);
	  close(cmd_err[1]);
	  return NULL;
     } 

     if (p == 0) {
	  /* Child process */
	  /* Redirect streams */
	  puts("Child: redirecting streams");
	  i = dup(1);
	  dup2(cmd_in[0],0); 
	  dup2(cmd_out[1],1);
	  dup2(cmd_err[1],2);
	  /* Close the streams we don't use */
	  write(i,"Child: closing streams\n",23);
	  close(cmd_in[0]);
	  close(cmd_in[1]);
	  close(cmd_out[0]);
	  close(cmd_out[1]);
	  close(cmd_err[0]);
	  close(cmd_err[1]);
	  /* Synchronize */
	  read(0,&syncbyte,1);
	  g_assert(syncbyte == 0xAB);
	  syncbyte = 0xCD;
	  write(1,&syncbyte,1);
	  /* Execute subprocess */
	  argv[0] = "sh";
	  argv[1] = "-c";
	  argv[2] = command;
	  argv[3] = NULL;
	  write(i,"Child: execve!\n",15);
	  fflush(stdout);
	  if (execve("/bin/sh",argv,environ) == -1) {
	       perror("/bin/sh");
	       _exit(1);
	  }
     }	  

     /* Parent process */
     /* Close the streams we don't use. */
     close(cmd_in[0]);
     close(cmd_out[1]);
     close(cmd_err[1]);
     
     /* synchronize to child */
     syncbyte = 0xAB;
     write(cmd_in[1],&syncbyte,1);
     read(cmd_out[0],&syncbyte,1);
     g_assert(syncbyte == 0xCD);

     /* Open chunk and create buffer */
     ch = chunk_open(chunk);
     if (!ch) {
	  close(cmd_in[1]);
	  close(cmd_out[0]);
	  close(cmd_err[0]);
	  return NULL;
     }
     ct = chunk_create_temp_init();
     if (!ct) {
	  chunk_close(ch);
	  close(cmd_in[1]);
	  close(cmd_out[0]);
	  close(cmd_err[0]);
	  return NULL;
     }
     processdialog_setup(ch->length);
     c = g_malloc(BZ);
     ui = 0;
     bs = bp = 0;
     /* We want to handle broken pipes ourselves. */
     signal(SIGPIPE,SIG_IGN);
     /* Read/write data... */
     /* This loop will keep running until the child process closes it's standard
      * output (=our data input) */
     /* i = fcntl(cmd_in[1],F_GETFL);
	i |= O_NONBLOCK;
	g_assert(fcntl(cmd_in[1],F_SETFL,i) == 0);
     */
     while (1) {	       
	  /* Wait for IO */
	  FD_ZERO(&rset);
	  FD_SET(cmd_out[0],&rset);
	  FD_SET(cmd_err[0],&rset);
	  FD_ZERO(&wset);
	  if (writing) FD_SET(cmd_in[1],&wset);
	  /* printf("Waiting (writing=%d)\n",writing); */
	  i = select(FD_SETSIZE,&rset,&wset,NULL,NULL);
	  /* printf("Wait finished.\n"); */
	  if (i == -1) {
	       if (errno == EINTR) continue;
	       goto errno_error_exit;
	  } 
	  
	  /* Read data from standard output */
	  if (FD_ISSET(cmd_out[0],&rset)) {
	       /* printf("Reading from stdin..\n"); */
	       i = read(cmd_out[0],c,BZ);
	       /* printf("Read finished\n"); */
	       if (i == -1) { if (errno != EINTR) goto errno_error_exit; }
	       else if (i == 0) { break; }
	       else { if (chunk_create_temp_write(ct,c,i)) goto error_exit; }
	       read_bytes += i;
	       continue;
	  }

	  /* Read data from standard error... */
	  if (FD_ISSET(cmd_err[0],&rset)) {
	       /* puts("Reading from stderr"); */
	       i = read(cmd_err[0],c,BZ-1);
	       /* puts("Read finished"); */
	       if (i > 0) {
		    c[i]=0;
		    if (!errortext) errortext=create_error_window(command);
		    add_error_text(errortext,c);
	       }
	       continue;
	  }

	  /* Send data to process's stdin */
	  if (writing && FD_ISSET(cmd_in[1],&wset)) {
	       /* printf("FD_ISSET(cmd_in[1],&wset) == %d\n",FD_ISSET(cmd_in[1],&wset)); */
	       /* Make sure the buffer contains data */
	       if (bs == bp) {
		    if (ui == chunk->length) { 
			 writing = FALSE; 
			 close(cmd_in[1]); 
			 continue; 
		    }
		    i = chunk_read_array(ch,ui,BZ,c);
		    if (!i) goto error_exit;
		    bs = i;
		    bp = 0;
		    ui += i / chunk->format.samplebytes;
	       }
	       /* printf("Writing data...\n"); */
	       /* Write data */
	       i = write(cmd_in[1],c+bp,MIN(bs-bp,PIPE_BUF));
	       /* printf("Finished writing.\n"); */
	       if (i == -1) {
		    /* Broken pipe - stop writing */
		    if (errno == EPIPE) { 
			 /* printf("Writing: Broken pipe.\n"); */
			 writing = FALSE; 
			 continue; 
		    }
		    if (errno == EINTR) continue;
		    goto errno_error_exit;
	       } 
	       bp += i;
	       if (processdialog_progress(i / chunk->format.samplebytes)) 
		    goto error_exit;
	  }
     }
     
     /* Check for any remaining data sent to standard error */
     while (1) {
	  /* printf("Checking remaining error data.\n"); */
	  FD_ZERO(&rset);
	  FD_SET(cmd_err[0],&rset);
	  i = select(FD_SETSIZE,&rset,NULL,NULL,&notime);
	  if (i<=0) break;
	  i = read(cmd_err[0],c,BZ-1);
	  if (i<=0) break;
	  c[i]=0;
	  if (!errortext) errortext=create_error_window(command);
	  add_error_text(errortext,c);
     }
     
     /* Finish */
     g_free(c);
     processdialog_finished();
     chunk_close(ch);
     if (writing) close(cmd_in[1]);
     close(cmd_out[0]);
     close(cmd_err[0]);

     if (errortext) {
	  if (!read_bytes) {
	       add_error_text(errortext,"\n\n(Command failed without returning any data)");
	       chunk_create_temp_abort(ct);
	       return NULL;
	  } else {
	       c = g_strdup_printf("\n\n(Command returned %d bytes of data)",read_bytes);
	       add_error_text(errortext,c);
	       g_free(c);
	  }
     }

     return chunk_create_temp_finished(ct,chunk);

errno_error_exit:
     g_free(c);
     c = g_strdup_printf("Error: %s",strerror(errno));
     user_message(c,UM_OK);
error_exit:
     g_free(c);
     processdialog_finished();
     chunk_close(ch);
     chunk_create_temp_abort(ct);
     close(cmd_in[1]);
     close(cmd_out[0]);
     close(cmd_err[0]);
     return NULL;
}
