/*
 * Copyright (C) 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include <glib.h>
#include "sound.h"
#include "player.h"
#include "main.h"
#include "um.h"


/* Info about currently playing sound. */
static ChunkHandle *ch=NULL; /* NULL == stopped */
static guint32 playstart, playend, curpos, timerel;
static gboolean loop, small_loop;
static GTimeVal start_time;
static guint32 zeroes;

/* Output buffer */
static gchar player_buf[4096];
static gint player_bufsize,player_bufpos;


/* Check if the sound driver wants more data and send it.
 * Returns FALSE if it had nothing to do. */
gboolean player_work(void)
{
     guint32 i;
     if (!ch || !output_want_data()) return FALSE;
     if (player_bufpos == player_bufsize) {
	  if (!loop && curpos==playend) {
	       /* Emit half a second of silence before stopping output. */
	       if (zeroes < ch->format.samplerate*ch->format.samplebytes/2) {
		    memset(player_buf,0,sizeof(player_buf));
		    i = sizeof(player_buf);
		    i -= i%ch->format.samplebytes;
		    zeroes += output_play(player_buf,i);
	       }
	       if (output_play(NULL,0)==0 && player_get_real_pos()==playend) {
		    output_stop();
		    chunk_close(ch);
		    ch=NULL;
	       } 
	       return FALSE;
	  }
	  player_bufpos = 0;
	  if (!small_loop) {
	       i = MIN(sizeof(player_buf),(playend-curpos)*ch->format.samplebytes);
	       player_bufsize = chunk_read_array(ch,curpos,i,player_buf);
	       curpos += player_bufsize/ch->format.samplebytes;
	       if (curpos == playend) {
		    if (loop) curpos = playstart;
		    else zeroes=0;
	       }
	  }
     }
     i = output_play(player_buf+player_bufpos,player_bufsize-player_bufpos);
     player_bufpos += i;
     return (i>0);
}


static void check_small_loop(void)
{
     guint32 u,v;
     small_loop = FALSE;
     if (loop) {
	  u = (playend-playstart)*ch->format.samplebytes;
	  if (u <= sizeof(player_buf)) {
	       small_loop = TRUE;
	       chunk_read_array(ch,playstart,u,player_buf);
	       for (v=1; v < sizeof(player_buf)/u; v++) memcpy(player_buf+u*v,player_buf,u); 
	       /* printf("check_small_loop: u=%d, v=%d\n",u,v); */
	       player_bufsize = v*u;
	  }
     }
}


gboolean player_play(Chunk *chk, guint32 spos, guint32 epos, gboolean lp)
{
     player_stop();
     if (spos == epos) return TRUE;
     if (!output_supports_format(&(chk->format))) {
	  user_message("The sound format of this file is not supported for "
		       "playing.",UM_OK);
	  return TRUE;
     }
     output_select_format(&(chk->format));
     ch = chunk_open(chk);
     playstart = spos;
     playend = epos;
     loop = lp;
     player_bufsize = 0;
     player_bufpos = 0;
     check_small_loop();
     curpos = spos;
     timerel = curpos;
     g_get_current_time(&start_time);
     player_work();
     return FALSE;
}

guint32 player_get_buffer_pos(void)
{
     return curpos;
}

guint32 player_get_real_pos(void)
{
     gfloat t = output_play_delay_time();
     GTimeVal tv,tv2;
     guint32 i;
     if (t > 0.0) {
	  i = curpos - (guint32)(t*((gfloat)(ch->format.samplerate)));
	  if (curpos == playend && !loop) i += zeroes;
	  if (i>=curpos) i=curpos;
	  if (!loop && i<playstart) i=playstart;
	  while (i<playstart) i += (playend-playstart);
     } else {
	  g_get_current_time(&tv);
	  timeval_subtract(&tv2,&tv,&start_time);
	  t = tv2.tv_sec + 0.000001 * tv2.tv_usec;
	  i = timerel + (guint32)(t*((gfloat)(ch->format.samplerate)));
	  if (i>playend && !loop) i=playend;
	  while (i>playend) i -= (playend-playstart);
     }
     return i;
}

void player_get_range(guint32 *startpos, guint32 *endpos)
{
     *startpos = playstart;
     *endpos = playend;
}

void player_set_buffer_pos(guint32 pos)
{
     if (pos != curpos) {
	  curpos = pos;
	  if (curpos > playend) playend = ch->length;
	  if (curpos < playstart) playstart = curpos;
	  g_get_current_time(&start_time);
	  timerel = pos;
	  output_clear_buffers();
	  player_bufsize = player_bufpos = 0;
	  player_work();
     }
}

void player_replace(Chunk *chunk, guint32 start, guint32 end, guint32 pos)
{
     ChunkHandle *c;
     c = chunk_open(chunk);
     if (c == NULL) return;
     chunk_close(ch);
     ch = c;
     playstart = start;
     playend = end;
     curpos = pos;
     timerel = pos;
     g_get_current_time(&start_time);
}


gboolean player_playing(void)
{
     return (ch != NULL);
}



void player_stop(void)
{
     if (ch == NULL) return;
     chunk_close(ch);
     ch = NULL;
     output_clear_buffers();
     output_stop();
}


