/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */

#include <math.h>

#include "recorddialog.h"
#include "vu_meter.h"
#include "um.h"
#include "sound.h"
#include "inifile.h"
#include "main.h"

#define INI_LAST_SAMPLESIZE "recordDialogLastSampleSize"
#define INI_LAST_SIGNED "recordDialogLastSigned"
#define INI_LAST_SAMPLERATE "recordDialogLastSampleRate"
#define INI_LAST_CHANNELS "recordDialogLastChannels"

static GtkObjectClass *parent_class;
static gboolean record_dialog_ok_flag;
static gboolean record_dialog_shutdown_flag;

static struct {
     GtkWidget *wnd;
     VuMeter **meters;
     GtkLabel *status;
     Dataformat format;
     gchar vu_data[4096];
     gint vu_data_count;
     gint vu_data_max;
} rd_process = { 0 };

static void record_dialog_process_destroy(void)
{
     rd_process.wnd = NULL;
}

static void record_dialog_process_setup(gchar *title)
{
     GtkWidget *a,*b;
     gint i;
     rd_process.wnd = gtk_window_new(GTK_WINDOW_DIALOG);
     gtk_window_set_modal(GTK_WINDOW(rd_process.wnd),TRUE);
     gtk_window_set_title(GTK_WINDOW(rd_process.wnd),title);
     gtk_signal_connect(GTK_OBJECT(rd_process.wnd),"destroy",
			GTK_SIGNAL_FUNC(record_dialog_process_destroy),NULL);
     a = gtk_vbox_new(FALSE,5);
     gtk_container_add(GTK_CONTAINER(rd_process.wnd),a);
     gtk_widget_show(a);

     rd_process.meters = g_malloc(rd_process.format.channels*sizeof(VuMeter *));
     for (i = 0; i<rd_process.format.channels; i++) {
	  b = vu_meter_new(0.90, 0.85);
	  gtk_box_pack_start(GTK_BOX(a),b,FALSE,FALSE,0);
	  gtk_widget_show(b);
	  rd_process.meters[i] = VU_METER(b);
     }

     b = gtk_label_new("");
     gtk_box_pack_start(GTK_BOX(a),b,FALSE,FALSE,0);
     gtk_widget_show(b);
     rd_process.status = GTK_LABEL(b);
     
     b = gtk_button_new_with_label("Stop");
     gtk_signal_connect_object(
	  GTK_OBJECT(b),"clicked",GTK_SIGNAL_FUNC(gtk_widget_destroy),
	  GTK_OBJECT(rd_process.wnd));
     gtk_box_pack_start(GTK_BOX(a),b,FALSE,FALSE,0);
     gtk_widget_show(b);
     
     gtk_widget_show(rd_process.wnd);
     
     rd_process.vu_data_count = 0;
     rd_process.vu_data_max = sizeof(rd_process.vu_data_count) - 
	  sizeof(rd_process.vu_data_count) % 
	  (rd_process.format.samplesize*rd_process.format.channels);
}

static void record_dialog_process_add_vudata(gpointer data, guint32 len)
{
     int i = MIN(len, rd_process.vu_data_max - rd_process.vu_data_count);
     if (!i) return;
     memcpy(rd_process.vu_data+rd_process.vu_data_count, data, i);
     rd_process.vu_data_count += i;
}

static void record_dialog_process_vu_update(void)
{
     sample_t vu;
     gint i,skip,c,x;
     if (!rd_process.wnd) return;
     if (rd_process.vu_data_count < rd_process.vu_data_max) return;
     skip = rd_process.format.samplesize * rd_process.format.channels;
     c = rd_process.vu_data_max / skip;
     for (i=0; i<rd_process.format.channels; i++) {
	  vu = 0.0;
	  x = i * rd_process.format.samplesize;
	  for (; x<rd_process.vu_data_max; x+=skip) {
	       vu = MAX(vu,
			fabs(convert_pcm_float(
				  rd_process.vu_data+x, &(rd_process.format))));
	  }
	  /* printf("vu = %f\n",vu); */
	  vu_meter_set_value(rd_process.meters[i], vu);
     }
     rd_process.vu_data_count = 0;
}

static void record_dialog_destroy(GtkObject *obj)
{
     record_dialog_shutdown_flag = TRUE;
     if (parent_class->destroy) parent_class->destroy(obj);
}

static void record_dialog_class_init(GtkObjectClass *klass)
{
     parent_class = gtk_type_class(gtk_window_get_type());
     klass->destroy = record_dialog_destroy;
}

static gboolean record_dialog_getformat(RecordDialog *rd, Dataformat *format)
{
     if (intbox_check(rd->samplerate) || intbox_check(rd->channels)) 
	  return TRUE;
     if (rd->samplerate->val <= 0 || rd->channels->val <= 0) {
	  user_message("Please enter positive numbers only",UM_OK);
	  return TRUE;
     }
     format->samplerate = rd->samplerate->val;
     format->samplesize = rd->fs->samplesize;
     format->sign = rd->fs->sign;
     format->channels = rd->channels->val;
     format->samplebytes = format->channels * format->samplesize;
     return FALSE;
}


static void record_dialog_ok(GtkButton *button, gpointer user_data)
{
     if (record_dialog_getformat(RECORD_DIALOG(user_data),&(rd_process.format)))
	  return;

     if (!input_supports_format(&(rd_process.format))) {
	  user_message("This format is not supported by the sound driver!",
		       UM_OK);
	  return;
     }
     record_dialog_ok_flag = TRUE;

     inifile_set_guint32(INI_LAST_SAMPLESIZE,rd_process.format.samplesize);
     inifile_set_gboolean(INI_LAST_SIGNED,rd_process.format.sign);
     inifile_set_guint32(INI_LAST_SAMPLERATE,rd_process.format.samplerate);
     inifile_set_guint32(INI_LAST_CHANNELS,rd_process.format.channels);

     gtk_widget_destroy(GTK_WIDGET(user_data));
}

static void record_dialog_volume_test(GtkButton *button, gpointer user_data)
{
     gchar buf[1024]; 
     Ringbuf *r;
     guint32 i;
     gboolean stopflag=FALSE;
     guint32 bytecount = 0;

     if (record_dialog_getformat(RECORD_DIALOG(user_data),&(rd_process.format)))
	  return;

     if (!input_supports_format(&(rd_process.format))) {
	  user_message("This format is not supported by the sound driver!",
		       UM_OK);
	  return;
     }

     r = ringbuf_new(65536);
     
     record_dialog_process_setup("Volume test");
     
     input_select_format(&(rd_process.format));

     while (!stopflag) {
	  gtk_label_set_text(rd_process.status, 
			     get_time(rd_process.format.samplerate,
				      bytecount/rd_process.format.samplebytes,
				      NULL));
	  input_store(r);
	  while (ringbuf_isempty(r) && rd_process.wnd != NULL) {
	       while (gtk_events_pending()) 
		    gtk_main_iteration();
	       record_dialog_process_vu_update();
	       input_store(r);
	  }
	  if (rd_process.wnd == NULL) break;
	  bytecount += ringbuf_available(r);
	  while (!ringbuf_isempty(r)) {
	       i = ringbuf_dequeue(r,buf,sizeof(buf));
	       record_dialog_process_add_vudata(buf,i);
	       if (rd_process.wnd == NULL)
		    stopflag = TRUE;
	  }
     }

     input_stop();

     ringbuf_free(r);
     
     return;
     
}

static void record_dialog_init(RecordDialog *obj)
{
     GtkWidget *a,*b,*c;

     gtk_container_set_border_width(GTK_CONTAINER(obj),5);
     obj->fs = FORMAT_SELECTOR( format_selector_new() );
     format_selector_set(obj->fs, inifile_get_guint32(INI_LAST_SAMPLESIZE,1),
			 inifile_get_gboolean(INI_LAST_SIGNED,FALSE) );
     obj->samplerate = INTBOX ( intbox_new(inifile_get_guint32(
						INI_LAST_SAMPLERATE,22050)) );
     obj->channels = INTBOX ( intbox_new(inifile_get_guint32(INI_LAST_CHANNELS,
							     1)) );

     gtk_window_set_title(GTK_WINDOW(obj),"Record");

     a = gtk_vbox_new(FALSE,5);
     gtk_container_add(GTK_CONTAINER(obj),a);
     gtk_widget_show(a);
     b = GTK_WIDGET(obj->fs);
     gtk_box_pack_start(GTK_BOX(a),b,FALSE,FALSE,0);
     gtk_widget_show(b);
     b = gtk_hseparator_new();
     gtk_box_pack_start(GTK_BOX(a),b,FALSE,FALSE,0);
     gtk_widget_show(b);
     b = gtk_table_new(2,2,FALSE);
     gtk_box_pack_start(GTK_BOX(a),b,FALSE,FALSE,0);
     gtk_widget_show(b);
     c = gtk_label_new("Sample rate: ");
     gtk_table_attach(GTK_TABLE(b),c,0,1,0,1,0,0,0,0);
     gtk_widget_show(c);
     c = GTK_WIDGET(obj->samplerate);
     gtk_table_attach(GTK_TABLE(b),c,1,2,0,1,GTK_EXPAND|GTK_FILL,0,0,0);
     gtk_widget_show(c);
     c = gtk_label_new("Channels: ");
     gtk_table_attach(GTK_TABLE(b),c,0,1,1,2,0,0,0,0);
     gtk_widget_show(c);
     c = GTK_WIDGET(obj->channels);
     gtk_table_attach(GTK_TABLE(b),c,1,2,1,2,GTK_EXPAND|GTK_FILL,0,0,0);
     gtk_widget_show(c);
     b = gtk_hseparator_new();
     gtk_box_pack_start(GTK_BOX(a),b,FALSE,FALSE,0);
     gtk_widget_show(b);
     b = gtk_hbox_new(FALSE,3);
     gtk_box_pack_start(GTK_BOX(a),b,FALSE,FALSE,0);
     gtk_widget_show(b);
     c = gtk_button_new_with_label("OK");
     gtk_signal_connect(GTK_OBJECT(c),"clicked",
			GTK_SIGNAL_FUNC(record_dialog_ok),obj);
     gtk_box_pack_start(GTK_BOX(b),c,FALSE,FALSE,0);
     gtk_widget_show(c);
     c = gtk_button_new_with_label("Cancel");
     gtk_signal_connect_object(GTK_OBJECT(c),"clicked",
			       GTK_SIGNAL_FUNC(gtk_widget_destroy),
			       GTK_OBJECT(obj));
     gtk_box_pack_start(GTK_BOX(b),c,FALSE,FALSE,0);
     gtk_widget_show(c);
     c = gtk_button_new_with_label("Volume test");
     gtk_signal_connect(GTK_OBJECT(c),"clicked",
			GTK_SIGNAL_FUNC(record_dialog_volume_test),obj);
     gtk_box_pack_start(GTK_BOX(b),c,FALSE,FALSE,0);
     gtk_widget_show(c);
     c = gtk_button_new_with_label("Launch mixer");
     gtk_signal_connect(GTK_OBJECT(c),"clicked",GTK_SIGNAL_FUNC(launch_mixer),
			NULL);
     gtk_box_pack_start(GTK_BOX(b),c,FALSE,FALSE,0);
     gtk_widget_show(c);
}

GtkType record_dialog_get_type(void)
{
     static GtkType id = 0;
     if (!id) {
	  GtkTypeInfo info = {
	       "RecordDialog",
	       sizeof(RecordDialog),
	       sizeof(RecordDialogClass),
	       (GtkClassInitFunc) record_dialog_class_init,
	       (GtkObjectInitFunc) record_dialog_init
	  };
	  id = gtk_type_unique(gtk_window_get_type(),&info);
     }
     return id;
}

GtkWidget *record_dialog_new(void)
{
     return gtk_type_new(record_dialog_get_type());
}

Chunk *record_dialog_execute(RecordDialog *rd)
{
     gchar buf[1024]; 
     Ringbuf *r;
     guint32 i;
     gboolean stopflag=FALSE;
     guint32 bytecount = 0;
     ChunkTemp tmp;

     record_dialog_shutdown_flag = FALSE;
     record_dialog_ok_flag = FALSE;

     gtk_widget_show(GTK_WIDGET(rd));
     gtk_window_set_modal(GTK_WINDOW(rd),TRUE);
     
     while (!record_dialog_shutdown_flag)
	  mainloop();
     
     if (!record_dialog_ok_flag) return NULL;

     tmp = chunk_create_temp_init();

     r = ringbuf_new(65536);
     
     record_dialog_process_setup("Recording");
     
     input_select_format(&(rd_process.format));

     while (!stopflag) {
	  gtk_label_set_text(rd_process.status, 
			     get_time(rd_process.format.samplerate,
				      bytecount/rd_process.format.samplebytes,
				      NULL));
	  input_store(r);
	  while (ringbuf_isempty(r) && rd_process.wnd != NULL) {
	       while (gtk_events_pending())
		    gtk_main_iteration();
	       record_dialog_process_vu_update();
	       input_store(r);
	  }
	  if (rd_process.wnd == NULL) break;
	  bytecount += ringbuf_available(r);
	  while (!ringbuf_isempty(r)) {
	       // puts("Hey 3");
	       i = ringbuf_dequeue(r,buf,sizeof(buf));
	       record_dialog_process_add_vudata(buf,i);
	       if (chunk_create_temp_write(tmp,buf,i) ||
		   rd_process.wnd == NULL)
		    stopflag = TRUE;
	  }
     }

     /* Get and store all the remaining input... */
     while (1) {
	  input_store(r);
	  i = ringbuf_dequeue(r,buf,sizeof(buf));
	  if (!i || chunk_create_temp_write(tmp,buf,i)) break;
     }     

     input_stop();

     ringbuf_free(r);
     
     return chunk_new_from_datasource(
	  tempfile_finished(tmp,&(rd_process.format)));
}
