/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include "ringbuf.h"

Ringbuf *ringbuf_new(guint32 size)
{
     Ringbuf *r;

     r = g_malloc(sizeof(Ringbuf)+size-1);

     r->empty = TRUE;
     r->size = size;

#ifdef THREADSAFE
     r->mutex = g_mutex_new();
#endif

     return r;
}

void ringbuf_free(Ringbuf *bufptr)
{
#ifdef THREADSAFE
     g_mutex_free(bufptr->mutex);
#endif
     g_free(bufptr);
}

guint32 ringbuf_available(Ringbuf *buf)
{
     guint32 start,end;
     gboolean empty;

     g_mutex_lock(buf->mutex);
     start = buf->start;
     end = buf->end;
     empty = buf->empty;
     g_mutex_unlock(buf->mutex);

     if (empty) 
	  return 0;
     else if (end > start) 
	  return end - start;
     else 
	  return buf->size + end - start;
}

guint32 ringbuf_freespace(Ringbuf *buf)
{
     guint32 start,end;
     gboolean empty;

     g_mutex_lock(buf->mutex);
     start = buf->start;
     end = buf->end;
     empty = buf->empty;
     g_mutex_unlock(buf->mutex);

     if (empty) 
	  return buf->size;
     else if (start >= end)
	  return start - end;
     else
	  return buf->size + start - end;
}

void ringbuf_drain(Ringbuf *buf)
{
     g_mutex_lock(buf->mutex);
     buf->empty = TRUE;
     g_mutex_unlock(buf->mutex);
}

gboolean ringbuf_isempty(Ringbuf *buf)
{
     gboolean empty;
     g_mutex_lock(buf->mutex);
     empty = buf->empty;
     g_mutex_unlock(buf->mutex);
     return empty;
}

gboolean ringbuf_isfull(Ringbuf *buf)
{
     guint32 start,end;
     gboolean empty;
     g_mutex_lock(buf->mutex);
     start = buf->start;
     end = buf->end;
     empty = buf->empty;
     g_mutex_unlock(buf->mutex);
     return (!empty && start==end);
}

static guint32 do_ringbuf_enqueue(Ringbuf *buf, gpointer data, guint32 datasize)
{
     guint32 blksize;
     guint32 start,end;
     gboolean empty;
     
     start = buf->start;
     end = buf->end;
     empty = buf->empty;

     if (empty) {
	  start = 0;
	  end = 0;
	  blksize = MIN ( datasize, buf->size );
     } else if (end > start) 
	  blksize = MIN ( datasize, buf->size - end );
     else
	  blksize = MIN ( datasize, start - end );

     if (blksize == 0) return 0;

     memcpy( buf->data + end, data, blksize );
     
     end += blksize;
     if (end == buf->size) end = 0;
     empty = FALSE;

     buf->start = start;
     buf->end = end;
     buf->empty = empty;
     
     return blksize + do_ringbuf_enqueue( buf, G_STRUCT_MEMBER_P ( data, blksize ),
					  datasize - blksize );
}

guint32 ringbuf_enqueue(Ringbuf *buf, gpointer data, guint32 datasize)
{
     guint32 i;
     g_mutex_lock(buf->mutex);
     i = do_ringbuf_enqueue(buf,data,datasize);
     g_mutex_unlock(buf->mutex);
     return i;
}


guint32 ringbuf_enqueue_zeroes(Ringbuf *buf, guint32 count)
{
     gchar c[1024];
     guint32 startcount = count;
     memset(c,0,sizeof(c));
     g_mutex_lock(buf->mutex);
     while (count > 0 && (buf->empty || buf->start != buf->end) )
	  count -= do_ringbuf_enqueue(buf,c,MIN(sizeof(c),count));
     g_mutex_unlock(buf->mutex);
     return startcount - count;
}

guint32 do_ringbuf_dequeue(Ringbuf *buf, gpointer data, guint32 datasize)
{
     guint32 blksize,start,end;
     gboolean empty;
     
     start = buf->start;
     end = buf->end;
     empty = buf->empty;

     if (empty)
	  blksize = 0;
     else if (start >= end)
	  blksize = MIN ( datasize, buf->size - start );
     else
	  blksize = MIN ( datasize, end - start );

     if (blksize == 0) return 0;
     
     memcpy ( data, buf->data + start, blksize );

     start += blksize;
     if (start == buf->size) start = 0;
     if (start == end) empty = TRUE;

     buf->start = start;
     buf->end = end;
     buf->empty = empty;     

     return blksize + do_ringbuf_dequeue(buf, G_STRUCT_MEMBER_P(data, blksize ),
					 datasize - blksize );
}

guint32 ringbuf_dequeue(Ringbuf *buf, gpointer data, guint32 datasize)
{
     guint32 i;
     g_mutex_lock(buf->mutex);
     i = do_ringbuf_dequeue(buf,data,datasize);
     g_mutex_unlock(buf->mutex);
     return i;
}

guint32 do_ringbuf_transfer ( Ringbuf *source, Ringbuf *dest )
{
     guint32 blksize, trans, dest_start, dest_end;
     gboolean dest_empty;
     
     dest_start = dest->start;
     dest_end = dest->end;
     dest_empty = dest->empty;
     
     if (dest_empty) {
	  dest_start = 0;
	  dest_end = 0;
	  blksize = dest->size;
     } else if (dest_end > dest_start)
	  blksize = dest->size - dest_end ;
     else 
	  blksize = dest_end - dest_start;

     trans = do_ringbuf_dequeue ( source, dest->data + dest_end, blksize );

     if (trans == 0) return 0;

     dest_end += trans;
     if (dest_end == dest->size) dest_end = 0;
     dest_empty = FALSE;

     dest->start = dest_start;
     dest->end = dest_end;
     dest->empty = dest_empty;

     return trans + do_ringbuf_transfer ( source, dest );
}

guint32 ringbuf_transfer ( Ringbuf *source, Ringbuf *dest )
{
     guint32 i;
     g_mutex_lock(source->mutex);
     g_mutex_lock(dest->mutex);
     i = do_ringbuf_transfer(source,dest);
     g_mutex_unlock(dest->mutex);
     g_mutex_unlock(source->mutex);
     return i;
}
