/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */

#include "sound.h"
#include "inifile.h"
#include "main.h"

gboolean sound_lock_driver;
static gboolean sound_delayed_quit=FALSE;
static Dataformat playing_format;

#ifdef HAVE_OSS
  #include "sound-oss.c"
#endif

#if defined (HAVE_PORTAUDIO)
  #include "sound-portaudio.c"
#endif

#if defined (HAVE_SDL)
  #include "sound-sdl.c"
#endif

#include "sound-dummy.c"

struct sound_driver {
     gchar *name, *id;
     
     void (*preferences)(void);

     void (*init)(void);
     void (*quit)(void);

     gboolean (*output_supports_format)(Dataformat *format);
     void (*output_select_format)(Dataformat *format);
     gboolean (*output_want_data)(void);
     guint (*output_play)(gchar *buffer, guint bufsize);
     void (*output_stop)(void);
     gfloat (*output_play_delay_time)(void);
     guint32 (*output_played_bytes)(void);
     gboolean (*output_wait)(guint timeout);
     void (*output_clear_buffers)(void);

     gboolean (*input_supported)(void);
     gboolean (*input_supports_format)(Dataformat *format);
     void (*input_select_format)(Dataformat *format);
     void (*input_store)(Ringbuf *buffer);
     void (*input_stop)(void);   
};

static struct sound_driver drivers[] = {

#ifdef HAVE_OSS
     { "Open Sound System", "oss", oss_preferences, oss_init, oss_quit, 
       oss_output_supports_format, oss_output_select_format,
       oss_output_want_data, oss_output_play, oss_stop, 
       oss_output_play_delay_time, oss_output_played_bytes, oss_output_wait,
       oss_output_clear_buffers,
       oss_input_supported, oss_input_supports_format, oss_input_select_format,
       oss_input_store, oss_stop },
#endif

#if defined (HAVE_PORTAUDIO)
     { "PortAudio", "pa", NULL, portaudio_init, portaudio_quit, 
       portaudio_output_supports_format, portaudio_output_select_format,
       portaudio_output_want_data, portaudio_output_play, portaudio_output_stop,
       portaudio_output_play_delay_time, portaudio_output_played_bytes,
       NULL, portaudio_output_clear_buffers,
       portaudio_input_supported, portaudio_input_supports_format, 
       portaudio_input_select_format, portaudio_input_store, 
       portaudio_input_stop }, 
#endif

#if defined (HAVE_SDL)
     { "SDL (output only)", "sdl", NULL, sdl_init, sdl_quit, 
       sdl_output_supports_format, sdl_output_select_format, 
       sdl_output_want_data, sdl_output_play, sdl_output_stop, 
       sdl_output_play_delay_time, sdl_output_played_bytes, 
       NULL, sdl_output_clear_buffers,
       NULL, NULL, NULL, NULL, NULL },
#endif

     { "Dummy (no sound)", "dummy", NULL, dummy_init, dummy_quit,
       dummy_output_supports_format, dummy_output_select_format, 
       dummy_output_want_data, dummy_output_play, dummy_output_stop,
       dummy_output_play_delay_time, dummy_output_played_bytes, 
       NULL, dummy_output_clear_buffers, 
       dummy_input_supported, dummy_input_supports_format, 
       dummy_input_select_format, dummy_input_store, dummy_input_stop }

};

static guint current_driver = 0;

gchar *sound_driver_name(void)
{
     return drivers[current_driver].name;
}

gchar *sound_driver_id(void)
{
     return drivers[current_driver].id;
}

GList *sound_driver_valid_names(void)
{
    GList *l = NULL;
    int i;
    for (i=0; i<ARRAY_LENGTH(drivers); i++)
        l = g_list_append(l, drivers[i].name);
    return l;
}

gchar *sound_driver_id_from_name(gchar *name)
{
    int i;
    for (i=0; i<ARRAY_LENGTH(drivers); i++)
        if (!strcmp(drivers[i].name,name))
            return drivers[i].id;
    return NULL;
}

gchar *sound_driver_name_from_id(gchar *id)
{
    int i;
    for (i=0; i<ARRAY_LENGTH(drivers); i++)
        if (!strcmp(drivers[i].id,id))
            return drivers[i].name;
    return NULL;
}

gboolean sound_driver_has_preferences(gchar *id)
{
     int i;

     if (!id) return (drivers[current_driver].preferences != NULL);

     for (i=0; i<ARRAY_LENGTH(drivers); i++)
	  if (!strcmp(drivers[i].id,id))
	       return (drivers[i].preferences != NULL);
     g_assert_not_reached();
     return FALSE;
}

void sound_driver_show_preferences(gchar *id)
{
     int i;

     if ((!id) && sound_driver_has_preferences(id)) 
	  drivers[current_driver].preferences();

     for (i=0; i<ARRAY_LENGTH(drivers); i++)
	  if (!strcmp(drivers[i].id,id))
	       drivers[i].preferences();
}

void sound_init(void)
{
     gchar *c;
     int i;
     sound_lock_driver = inifile_get_gboolean("soundLock",FALSE);
     c = inifile_get(INI_SETTING_SOUNDDRIVER, drivers[0].id);
     for (i=0; i<ARRAY_LENGTH(drivers); i++) {
	  if (!strcmp(drivers[i].id,c)) {
	       current_driver = i;
	       break;
	  }
     }
     if (i == ARRAY_LENGTH(drivers)) {
	  g_printerr("Invalid driver name: %s\n",c);
	  current_driver = 0;
     }

     drivers[current_driver].init();
}

void sound_quit(void)
{
     drivers[current_driver].quit();
}

static void delayed_output_stop(void)
{
     drivers[current_driver].output_stop();
     sound_delayed_quit=FALSE;
}

gboolean output_supports_format(Dataformat *format)
{
     if (sound_delayed_quit) {
	  if (dataformat_equal(format,&playing_format)) return TRUE;
	  else delayed_output_stop();
     }
     return drivers[current_driver].output_supports_format(format);
}

void output_select_format(Dataformat *format)
{
     if (sound_delayed_quit) {
	  if (dataformat_equal(format,&playing_format)) {
	       sound_delayed_quit = FALSE;
	       return;
	  }
	  else delayed_output_stop();
     }
     drivers[current_driver].output_select_format(format);
     memcpy(&playing_format,format,sizeof(Dataformat));
}

gboolean input_supported(void)
{
     return drivers[current_driver].input_supported ? 
	  drivers[current_driver].input_supported() : 
	  FALSE;
}

gboolean input_supports_format(Dataformat *format)
{
     return drivers[current_driver].input_supports_format(format);
}

void input_select_format(Dataformat *format)
{
     drivers[current_driver].input_select_format(format);
}

void output_stop(void)
{
     if (sound_lock_driver) {
	  output_clear_buffers();
	  sound_delayed_quit=TRUE;
     } else
	  drivers[current_driver].output_stop();
}

gboolean output_want_data(void)
{
     return drivers[current_driver].output_want_data();
}

guint output_play(gchar *buffer, guint bufsize)
{
     return drivers[current_driver].output_play(buffer,bufsize);
}

gfloat output_play_delay_time(void)
{
     return drivers[current_driver].output_play_delay_time();
}

guint32 output_played_bytes(void)
{
     return drivers[current_driver].output_played_bytes();
}

void input_stop(void)
{
     drivers[current_driver].input_stop();
}

void input_store(Ringbuf *buffer)
{
     drivers[current_driver].input_store(buffer);
}

gboolean output_wait(guint timeout)
{
     if (drivers[current_driver].output_wait)
	  return drivers[current_driver].output_wait(timeout);
     else
	  return FALSE;
}

void output_clear_buffers(void)
{
     drivers[current_driver].output_clear_buffers();
}
