/*
 * Copyright (C) 2002 2003, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */

#include <string.h>
#include <math.h>
#include "viewcache.h"
#include "inifile.h"
#include "main.h"

/* How many columns to update (max) for each call to view_cache_update */
#define PIXELS_PER_UPDATE 10
#define GUINT32(x) ((guint32)x)
#define FLOAT(x) ((float)x)

ViewCache *view_cache_new(void)
{
     ViewCache *vc;
     vc = g_malloc(sizeof(*vc));
     vc->chunk = NULL;
     vc->handle = NULL;
     vc->values = NULL;
     vc->offsets = NULL;
     vc->calced = NULL;
     return vc;
}

static void view_cache_clear(ViewCache *vc)
{
     if (vc->chunk) {
	  if (vc->handle) chunk_close(vc->handle);
	  vc->handle = NULL;
	  gtk_object_unref(GTK_OBJECT(vc->chunk));
     }
     g_free(vc->values);
     vc->values = NULL;
     g_free(vc->offsets);
     vc->offsets = NULL;
     g_free(vc->calced);
     vc->calced = NULL;
}

void view_cache_free(ViewCache *vc)
{
     view_cache_clear(vc);
     g_free(vc);
}

gboolean view_cache_update(ViewCache *cache, Chunk *chunk, gint start_samp, 
			   gint end_samp, gint xres)
{
     static gchar *buf=NULL;
     static guint bufsize=0;
     static sample_t *sbuf=NULL;
     static guint sbufsize=0;
          
     sample_t *new_values;
     guint32 *new_offsets;
     gchar *new_calced;

     gboolean chunk_change;
     gboolean range_change;

     gfloat spp,f,g,h;
     guint i,j,k,l,m,n;
     gchar *c;
     guint channels = chunk?chunk->format.channels:0;

     sample_t x;

     if (xres > end_samp-start_samp) xres=end_samp-start_samp;

     chunk_change = (cache->chunk != chunk);
     range_change = (chunk != NULL) && (chunk_change ||
					cache->start != start_samp || 
					cache->end != end_samp || 
					cache->xres != xres);

     /* Handle changed chunk */
     if (chunk_change) {
	  view_cache_clear(cache);
	  cache->chunk = chunk;
	  if (chunk != NULL) 
	       gtk_object_ref(GTK_OBJECT(chunk));
     }

     /* Handle changed range */
     if (range_change) {
     
	  if (cache->handle == NULL) 
	       cache->handle = chunk_open(chunk);

	  spp = ((float)(end_samp - start_samp)) / ((float)xres);	       

	  /* Calculate desired number of samples/pixel */
	  cache->samples_per_pixel = 
	       inifile_get_guint32(INI_SETTING_VIEW_QUALITY,
				   INI_SETTING_VIEW_QUALITY_DEFAULT);
	  i = (gint)ceill(spp);
	  if (i <= cache->samples_per_pixel) cache->samples_per_pixel = 0;	       

	  /* Create new buffers */
	  new_values = g_malloc(xres * channels * 2 * sizeof(*new_values));
	  /* Make one extra entry in offsets with end point. (used below) */
	  new_offsets = g_malloc((xres+1) * sizeof(*new_offsets));
	  new_calced = g_malloc0(xres);

	  /* Calculate new_offsets */
	  new_offsets[0] = start_samp;
	  new_offsets[xres] = end_samp;
	  f = (float)start_samp + spp;
	  for (i=1; i<xres; i++,f+=spp) new_offsets[i] = (guint32)f;

	  /* Try to steal from the old buffer, even if it doesn't match exactly,
	     but not if we're in pixel<->sample mode and the old range was not.
	  */	  
	  /* Check for overlap */
	  if (!chunk_change &&
	      (xres < end_samp-start_samp || 
	       cache->xres==cache->end-cache->start) && (
	      (end_samp >= cache->start && start_samp < cache->end) 
	      || (start_samp <= cache->end && end_samp > cache->start))) {
	       if (end_samp-start_samp==cache->end-cache->start && xres==cache->xres) {
		    /* Special case: scrolling */
		    /* Set j=old location, i=new location, k=# of values to copy */
		    j=0;
		    i=0;
		    if (cache->offsets[0] < new_offsets[0]) 
			 j = GUINT32(FLOAT(new_offsets[0] - cache->offsets[0]) / spp);
		    else
			 i = GUINT32(FLOAT(cache->offsets[0] - new_offsets[0]) / spp);
		    k = xres - MAX(i,j);
		    memcpy(new_values+(2*i*channels), 
			   cache->values+(2*j*channels),
			   k*channels*2*sizeof(sample_t));
		    memcpy(new_offsets+i, cache->offsets+j, k*sizeof(new_offsets[0]));
		    memcpy(new_calced+i, cache->calced+j, k);
	       } else {
		    /* Common case */
		    j=0; /* Gammalt index */
		    i=0; /* Nytt index */
		    while (i<xres && j<cache->xres) {
			 /* Mjligheter:
			    1. gamla index j passar p nya index i. Kopiera, ka i,j
			    terfs: d i eller j kas.
			    2. j r fr tidigt, ka j. terfs: d i kas eller (2)
			    3. j r fr sent, ka i. terfs: d j kas eller (3) 
			    4. j r inte berknat, ka j. terfs: d j kas.
			 */
			 
			 if (cache->offsets[j] < new_offsets[i]) j++; /* 2 */
			 else if (!cache->calced[j]) j++;             /* 4 */
			 else if (cache->offsets[j] >= new_offsets[i+1]) i++; /*3*/
			 else {
			      /* Match! */
			      memcpy(new_values + i*channels*2, 
				     cache->values + j*channels*2, 
				     channels*2*sizeof(sample_t));
			      new_offsets[i] = cache->offsets[j];
			      new_calced[i] = 1;
			      
			      j++;
			      i++;
			 }
		    }
	       }
	  }

	  /* Update cache vars */
	  cache->start = start_samp;
	  cache->end = end_samp;
	  cache->xres = xres;
	  g_free(cache->values);
	  cache->values = new_values;
	  g_free(cache->offsets);
	  cache->offsets = new_offsets;
	  g_free(cache->calced);
	  cache->calced = new_calced;
	  return TRUE;
     }

     /* Regular update */

     if (chunk==NULL || cache->handle==NULL) return FALSE;

     /* Scan for uncalculated data */
     c = memchr(cache->calced, 0, xres);
     if (c == NULL) {
	  chunk_close(cache->handle);
	  cache->handle = NULL;
	  return TRUE;
     }
 
     i = c - cache->calced;
     
     /* Scan for end of uncalculated data */
     j = i+1;
     while (j<i+PIXELS_PER_UPDATE && j<xres && cache->calced[j]==0) j++;
    
     /* Fill in as calculated */
     memset(cache->calced+i, 1, j-i);

     if (cache->samples_per_pixel == 0) {
	  /* CONSECUTIVE MODE */	  
	  k = cache->offsets[j]-cache->offsets[i]; /* # of samples to read */
	  /* Read data */
	  l = k*chunk->format.samplebytes; /* # of raw bytes to read */
	  if (l>bufsize) {
	       g_free(buf);
	       bufsize = l;
	       buf = g_malloc(bufsize);
	  }
	  /* Ignore errors, the user will get error message anyway */
	  m=chunk_read_array(cache->handle,cache->offsets[i],l,buf);	  
	  g_assert(m==0 || m==l);
	  
	  /* Convert data */
	  m = k * channels * sizeof(sample_t); /* # of converted bytes */
	  if (m>sbufsize) {
	       g_free(sbuf);
	       sbufsize = m;
	       sbuf = g_malloc(sbufsize);
	  }
	  chunk_sample_get_array(buf, chunk, sbuf, k*channels);

	  /* Calculate max/min */
	  m = 0; /* Current full sample in sbuf */
	  n = i; /* Current pos in calced */
	  
	  while (n<j) {
	       /* Set "impossible" values */
	       for (l=0; l<channels; l++) {
		    cache->values[(n*channels+l)*2] = 1.0; /* Min */
		    cache->values[(n*channels+l)*2+1] = -1.0; /* Max */
	       }
	       /* Go through all necessary samples */
	       while (m+cache->offsets[i] < cache->offsets[n+1]) {
		    for (l=0; l<channels; l++) {
			 x = sbuf[m*channels+l];
			 if (x < cache->values[(n*channels+l)*2])
			      cache->values[(n*channels+l)*2] = x;
			 if (x > cache->values[(n*channels+l)*2+1])
			      cache->values[(n*channels+l)*2+1] = x;			 
		    }
		    m++;
	       }
	       n++;
	  }
     } else {
	  
	  /* REGULAR MODE */
	  while (i<j) {
	       k = cache->samples_per_pixel * chunk->format.samplebytes;
	       if (bufsize < k) {
		    g_free(buf);
		    bufsize = k;
		    buf = g_malloc(bufsize);
	       }
	       l=chunk_read_array(cache->handle, cache->offsets[i], k, buf);
	       g_assert(l==k || l==0);
	       
	       k = cache->samples_per_pixel * channels * sizeof(sample_t);
	       if (sbufsize < k) {
		    g_free(sbuf);
		    sbufsize = k;
		    sbuf = g_malloc(sbufsize);
	       }
	       chunk_sample_get_array(buf,chunk,sbuf,cache->samples_per_pixel*channels);
	       
	       for (k=0; k<channels; k++) {
		    g = 1.0;
		    h = -1.0;
		    for (l=0; l<cache->samples_per_pixel; l++) {
			 f = sbuf[l*channels+k];
			 if (f<g) g=f;
			 if (f>h) h=f;
		    }
		    cache->values[(i*channels+k)*2] = g;
		    cache->values[(i*channels+k)*2+1] = h;
	       }

	       i++;
	  }

     }
     
     return TRUE;
}

gboolean view_cache_uptodate(ViewCache *cache)
{
     return (cache->handle == NULL);
}

void view_cache_draw(ViewCache *cache, GdkDrawable *d, gint yres, 
		     GtkWidget *wid)
{
     gint i,j,min,max,lmin=0,lmax=0;
     gboolean haslast;
     float mid,size,f,g;
     GdkGC *gc;
     gchar *c;
     guint channels = (cache->chunk)?cache->chunk->format.channels:0;

     if (cache->chunk == NULL) return;

     for (j=0; j<channels; j++) {

	  mid = ((yres*j)/channels) + (yres/(2*channels));
	  size = yres/(2*channels) - 5;
	  gc = (j&1)?gc_red(wid):gc_blue(wid);

	  haslast = FALSE;

	  for (i=0; i<cache->xres; i++) {
	       if (cache->calced[i] == 0) {
		    c = memchr(cache->calced+i, 1, cache->xres-i);
		    if (c == NULL) break;
		    i = c-cache->calced;
		    haslast = FALSE;
	       }
	       f = cache->values[(i*channels+j)*2]*size+mid;
	       g = cache->values[(i*channels+j)*2+1]*size+mid;
	       min = (gint) f;
	       max = (gint) g;
	       if (haslast) {
		    if (min > lmax) min = lmax+1;
		    else if (max < lmin) max = lmin-1;
	       }
	       gdk_draw_line(d,gc,i,max,i,min);
	       lmin = min;
	       lmax = max;
	       haslast = TRUE;
	  }

     }
}
